package DbStorage;

import java.io.*;
import storage.primitiveStorage.*;
import storage.intermediateStorage.*;
import storage.advancedStorage.*;
import storage.tools.*;
import storage.*;

/**
 * This class works as a intermedite between the high level database data structures
 * and, the Storage and LOG modules. The database storage is compound of 5 segments:
 * dictionary segment, id segment, content segment and LOB segment .
 * This class synchonizes its methods to ensure
 * that the basic I/O database operatios will be athomic.  
 */
public final class DbStorageManager
{
    /** The number of the id segment. */
    static public final int ID_SEGMENT=0;
    
    /** The number of the content segment. */
    static public final int CONTENT_SEGMENT=1;
    
    /** The number of the LOB segment. */
    static public final int LOB_SEGMENT=2;
    
    /** The number of the parents' segment. */
    static public final int PARENTS_SEGMENT=3;

    /** The number of the dictionary segment. */
    static public final int DICTIONARY_SEGMENT=4;


    /** The size of the units in the id segment */
    static public final int ID_SEGMENT_UNIT_SIZE=96;

    /** The size of the units in the content segment */
    static public final int CONTENT_SEGMENT_UNIT_SIZE=200;
    
    /** The size of the units in the LOB segment */
    public int LOB_SEGMENT_UNIT_SIZE;

    /** The size of the units in the parents' segment */
    static public final int PARENTS_SEGMENT_UNIT_SIZE=73;

    /** The size of the units in the dictionary segment */
    static public final int DICTIONARY_SEGMENT_UNIT_SIZE=200;


    private AdvancedStorageManager asm;
    private UnitsStorageManager usm;
    private StorageToolsManager stm;


    /** 
     * Adds (Occups) a new unit in the segment.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param newUnitAddress Where the address of the new unit will be stored.
     */
    synchronized public void addUnit(long transactionNumber, int segmentNumber,  
				     DbUnitAddress newUnitAddress)
	throws StorageException, IOException
    {
	if (usm.findFreeUnit(segmentNumber,newUnitAddress))
	    {
		usm.addUnit(segmentNumber,newUnitAddress);
	    }
	else
	    newUnitAddress.setNull();
    }

    /** 
     * Tries to add (Occup) a new unit first in the same page, second in the same space and 
     * finally in the same segment of a specified unit.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment of the close unit.
     * @param unitAddress The address of the close unit in the segment. 
     * @param newUnitAddress Where the address of the new unit will be stored.
     */
    synchronized public void addUnitAsCloserAsPosibleOf(long transactionNumber, 
							int segmentNumber, DbUnitAddress unitAddress, 
							DbUnitAddress newUnitAddress)
	throws StorageException, IOException
    {
	if (usm.findFreeUnitAsCloserAsPosibleOf(segmentNumber,unitAddress,newUnitAddress))
	    {
		usm.addUnit(segmentNumber,newUnitAddress);
	    }
	else
	    newUnitAddress.setNull();
    }


    /** 
     * Adds (Occups) a specified unit in the segment.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit to ocup.
     */
    synchronized public void addSpecifiedUnit(long transactionNumber, 
					      int segmentNumber, DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	usm.addUnit(segmentNumber,unitAddress);
    }


    /**
     * Drops (Frees) the specified unit.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment. 
     */
    synchronized public void dropUnit(long transactionNumber, int segmentNumber, DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	usm.dropUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber());
    }
    
    /**
     * Reads fully a unit. This method does not consider anything about the free status of the unit.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param b The byte array which the content of the readed unit will be stored.
     * @param off The offset of <code>b</code> from which the readed content will be stored.
     */
    synchronized public void readUnit(long transactionNumber, 
				      int segmentNumber, DbUnitAddress unitAddress, byte[] b, int off)
	throws StorageException, IOException
    {
	usm.readUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber(),b,off);
    }
    
    /**
     * Reads partially a unit starting from an offset and reading a specific length.
     * This method does not consider anything about the free status of the unit.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param unitOffset The position of the unit from which the reading starts.
     * @param length The number of bytes to be readed.
     * @param b The byte array which the content of the readed unit will be stored.
     * @param off The offset of <code>b</code> from which the readed content will be stored.
     */
    synchronized public void readUnit(long transactionNumber,
				      int segmentNumber, DbUnitAddress unitAddress, 
				      int unitOffset, int length, byte[] b, int off)
	throws StorageException, IOException
    {
	usm.readUnit(segmentNumber,unitAddress.getSpaceNumber(),
		     unitAddress.getUnitNumber(),unitOffset,length,b,off);
    }
    
     /**
     * Writes fully a unit.
     * This method does not consider anything about the free status of the unit.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param b The byte array from which the units's content will be get.
     * @param off The offset of <code>b</code> from which the unit's content will be get.
     */
    synchronized public void writeUnit(long transactionNumber,
				       int segmentNumber, DbUnitAddress unitAddress, byte[] b, int off)
	throws StorageException, IOException
    {
	usm.writeUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber(),b,off);
    }

    /**
     * Writes partially a unit starting from an offset and writing a specific length of bytes.
     * This method does not consider anything about the free status of the unit.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param unitOffset The position of the unit from which the writing starts.
     * @param length The number of bytes to be wrote.
     * @param b The byte array from which the units's content will be get.
     * @param off The offset of <code>b</code> from which the unit's content will be get.
     */
    synchronized public void writeUnit(long transactionNumber,
				       int segmentNumber, DbUnitAddress unitAddress, 
				       int unitOffset, int length, byte[] b, int off)
	throws StorageException, IOException
    {
	usm.writeUnit(segmentNumber,unitAddress.getSpaceNumber(),
		      unitAddress.getUnitNumber(),unitOffset,length,b,off);
    }


    /**
     * Checks if a unit is free.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @return <code>true</code> if the specified unit is free, <code>false</code> otherwise.
     */
    synchronized public boolean isFreeUnit(long transactionNumber, 
					   int segmentNumber, DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	return usm.isFreeUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber());
    }

    /**
     * Checks if a unit is occupied.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @return <code>true</code> if the specified unit is occupied, <code>false</code> otherwise.
     */
    synchronized public boolean isOccupiedUnit(long transactionNumber, 
					       int segmentNumber, DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	return usm.isOccupiedUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber());
    }


    private byte[] dataAux;

    /**
     * Reads a <code>byte</code> inside a unit.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param unitOffset The position of the <code>byte</code> in the unit.
     * @return The <code>byte</code> in the specified page's position.
     */
    synchronized public byte readByte(long transactionNumber,
				      int segmentNumber, DbUnitAddress unitAddress, int unitOffset)
	throws StorageException, IOException
    {
	byte r;
	usm.readUnit(segmentNumber,unitAddress.getSpaceNumber(),
		     unitAddress.getUnitNumber(),unitOffset,1,dataAux,0);
	r=dataAux[0];
	return r;
    }

    /**
     * Reads a <code>char</code> inside a unit.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param unitOffset The position of the <code>char</code> in the unit.
     * @return The <code>char</code> in the specified page's position.
     */
    synchronized public char readChar(long transactionNumber,
				      int segmentNumber, DbUnitAddress unitAddress, int unitOffset)
	throws StorageException, IOException
    {
	char r=0;
	usm.readUnit(segmentNumber,unitAddress.getSpaceNumber(),
		     unitAddress.getUnitNumber(),unitOffset,2,dataAux,0);
	r |= (0xFF & dataAux[0]);
	r |= (0xFF & (dataAux[1]<<8));
	return r;
    }

    /**
     * Reads a <code>int</code> inside a unit.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param unitOffset The position of the <code>int</code> in the unit.
     * @return The <code>int</code> in the specified page's position.
     */
    synchronized public int readInt(long transactionNumber, 
				    int segmentNumber, DbUnitAddress unitAddress, int unitOffset)
	throws StorageException, IOException
    {
	int r=0;
	usm.readUnit(segmentNumber,unitAddress.getSpaceNumber(),
		     unitAddress.getUnitNumber(),unitOffset,4,dataAux,0);
	for (int i=0; i<4; i++)
	    r |= ((0xFF & dataAux[i]) << (i*8));
			
	return r;
    }


    /**
     * Reads a <code>long</code> inside a unit.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param unitOffset The position of the <code>long</code> in the unit.
     * @return The <code>long</code> in the specified page's position.
     */
    synchronized public long readLong(long transactionNumber,
				      int segmentNumber, DbUnitAddress unitAddress, int unitOffset)
	throws StorageException, IOException
    {
	long r=0;
	usm.readUnit(segmentNumber,unitAddress.getSpaceNumber(),
		     unitAddress.getUnitNumber(),unitOffset,8,dataAux,0);
	for (int i=0; i<8; i++)
	    r |= ((0xFFl & dataAux[i]) << (i*8));
	return r;
    }

    /**
     * Reads a <code>double</code> inside a unit.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param unitOffset The position of the <code>double</code> in the unit.
     * @return The <code>double</code> in the specified page's position.
     */
    synchronized public double readDouble(long transactionNumber,
					  int segmentNumber, DbUnitAddress unitAddress, int unitOffset)
	throws StorageException, IOException
    {
	long r=0;
	usm.readUnit(segmentNumber,unitAddress.getSpaceNumber(),
		     unitAddress.getUnitNumber(),unitOffset,8,dataAux,0);
	for (int i=0; i<8; i++)
	    r |= ((0xFFl & dataAux[i]) << (i*8));
	return Double.longBitsToDouble(r);
    }

    /**
     * Reads a <code>DbUnitAddress</code> inside a unit.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param unitOffset The position of the <code>DbUnitAddress</code> in the unit.
     * @param newUnitAddress The <code>DbUnitAddress</code> in the specified page's position.
     */
    synchronized public void readDbUnitAddress(long transactionNumber,
					       int segmentNumber, DbUnitAddress unitAddress, int unitOffset,
					       DbUnitAddress newUnitAddress)
	throws StorageException, IOException
    {
	usm.readUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber(),
		      unitOffset,DbUnitAddress.SIZE,newUnitAddress.getContent(),0);
    }



    /**
     * Writes a <code>byte</code> into a unit starting from an offset.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param unitOffset The position of the unit from which the writing starts.
     * @param b The <code>byte</code> to be wrote.
     */
    synchronized public void writeByte(long transactionNumber,
				       int segmentNumber, DbUnitAddress unitAddress, int unitOffset, byte b)
	throws StorageException, IOException
    {
	dataAux[0]=b;
	usm.writeUnit(segmentNumber,unitAddress.getSpaceNumber(),
		      unitAddress.getUnitNumber(),unitOffset,1,dataAux,0);
    }

    /**
     * Writes a <code>char</code> into a unit starting from an offset.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param unitOffset The position of the unit from which the writing starts.
     * @param c The <code>char</code> to be wrote.
     */
    synchronized public void writeChar(long transactionNumber,
				       int segmentNumber, DbUnitAddress unitAddress, int unitOffset, char c)
	throws StorageException, IOException
    {
	dataAux[0]=(byte)(0xFF & c);
	dataAux[1]=(byte)(0xFF & (c>>8));
	usm.writeUnit(segmentNumber,unitAddress.getSpaceNumber(),
		      unitAddress.getUnitNumber(),unitOffset,2,dataAux,0);
    }

    /**
     * Writes a <code>int</code> into a unit starting from an offset.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param unitOffset The position of the unit from which the writing starts.
     * @param i The <code>int</code> to be wrote.
     */
    synchronized public void writeInt(long transactionNumber,
				      int segmentNumber, DbUnitAddress unitAddress, int unitOffset, int i)
	throws StorageException, IOException
    {
	for (int j=0; j<4; j++)
	    dataAux[j]=(byte)(0xFF & (i>>(8*j)));
	usm.writeUnit(segmentNumber,unitAddress.getSpaceNumber(),
		      unitAddress.getUnitNumber(),unitOffset,4,dataAux,0);
    }

    /**
     * Writes a <code>long</code> into a unit starting from an offset.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param unitOffset The position of the unit from which the writing starts.
     * @param l The <code>long</code> to be wrote.
     */
    synchronized public void writeLong(long transactionNumber,
				       int segmentNumber, DbUnitAddress unitAddress, int unitOffset, long l)
	throws StorageException, IOException
    {
	for (int j=0; j<8; j++)
	    dataAux[j]=(byte)(0xFF & (l>>(8*j)));
	usm.writeUnit(segmentNumber,unitAddress.getSpaceNumber(),
		      unitAddress.getUnitNumber(),unitOffset,8,dataAux,0);
    }

    /**
     * Writes a <code>double</code> into a unit starting from an offset.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param unitOffset The position of the unit from which the writing starts.
     * @param d The <code>double</code> to be written.
     */
    synchronized public void writeDouble(long transactionNumber,
					 int segmentNumber, DbUnitAddress unitAddress, int unitOffset, double d)
	throws StorageException, IOException
    {
	long l=Double.doubleToRawLongBits(d);
	for (int j=0; j<8; j++)
	    dataAux[j]=(byte)(0xFF & (l>>(8*j)));
	usm.writeUnit(segmentNumber,unitAddress.getSpaceNumber(),
		      unitAddress.getUnitNumber(),unitOffset,8,dataAux,0);
    }


    /**
     * Writes a <code>DbUnitAddress</code> into a unit starting from an offset.
     * @param transactionNumber The number of transaction.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit in the segment.
     * @param unitOffset The position of the unit from which the writing starts.
     * @param newUnitAddress The <code>DbUnitAddress</code> to be wrote.
     */
    synchronized public void writeDbUnitAddress(long transactionNumber, 
						int segmentNumber, DbUnitAddress unitAddress, int unitOffset, 
						DbUnitAddress newUnitAddress)
	throws StorageException, IOException
    {
	usm.writeUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber(),
		      unitOffset,DbUnitAddress.SIZE,newUnitAddress.getContent(),0);
    }




    /**
     * Creates a default storage organization for a database.
     */
    synchronized public void createDefaultSettings(File dbDirectory)
	throws PrimitiveStorageException, IntermediateStorageException, AdvancedStorageException, 
	       IOException
    {
	PrimitiveStorageManager psm=getPrimitiveStorageManager();
	psm.loadFlexibleChunk(0,new File(dbDirectory,"IdSpa00.dat"),true);
	psm.loadFlexibleChunk(1,new File(dbDirectory,"ConSpa00.dat"),true);
	psm.loadFlexibleChunk(2,new File(dbDirectory,"LobSpa00.dat"),true);
	psm.loadFlexibleChunk(3,new File(dbDirectory,"ParSpa00.dat"),true);
	psm.loadFlexibleChunk(4,new File(dbDirectory,"DicSpa00.dat"),true);
	
	asm.getSegment(ID_SEGMENT).loadSpace(0,0,true);
	asm.getSegment(CONTENT_SEGMENT).loadSpace(0,1,true);
	asm.getSegment(LOB_SEGMENT).loadSpace(0,2,true);
	asm.getSegment(PARENTS_SEGMENT).loadSpace(0,3,true);
	asm.getSegment(DICTIONARY_SEGMENT).loadSpace(0,4,true);
	getStorageToolsManager().saveSettingsToFile(new File(dbDirectory,"storage.cfg"));
    }

    /**
     * Creates a new database storage manager. 
     * Sets the underlying advanced storage manager according to the database needs. 
     */
    public DbStorageManager(PrimitiveStorageManager psm)
	throws TooSmallPageSizeException
    {
	int pageSize=psm.getPageSize();
	if (pageSize<ID_SEGMENT_UNIT_SIZE || pageSize<CONTENT_SEGMENT_UNIT_SIZE ||
	    pageSize<PARENTS_SEGMENT_UNIT_SIZE || pageSize<DICTIONARY_SEGMENT_UNIT_SIZE)
	    throw new TooSmallPageSizeException();
	LOB_SEGMENT_UNIT_SIZE=psm.getPageSize();
	int[] unitSizes=new int[5];
	unitSizes[ID_SEGMENT]=ID_SEGMENT_UNIT_SIZE;
	unitSizes[CONTENT_SEGMENT]=CONTENT_SEGMENT_UNIT_SIZE;
	unitSizes[LOB_SEGMENT]=LOB_SEGMENT_UNIT_SIZE;
	unitSizes[PARENTS_SEGMENT]=PARENTS_SEGMENT_UNIT_SIZE;
	unitSizes[DICTIONARY_SEGMENT]=DICTIONARY_SEGMENT_UNIT_SIZE;
	try {
	    this.asm=new AdvancedStorageManager(new IntermediateStorageManager(psm),unitSizes);
	} catch(InvalidUnitSizeException e) {
	}
	this.usm=new UnitsStorageManager(asm);
	this.stm=new StorageToolsManager(asm);
	dataAux=new byte[8];
    }

    /**
     * Returns the underlying advanced storage manager.
     */
    public AdvancedStorageManager getAdvancedStorageManager()
    {
	return asm;
    }

    /**
     * Returns the underlying intermediate storage manager.
     */
    public IntermediateStorageManager getIntermediateStorageManager()
    {
	return asm.getIntermediateStorageManager();
    }

    /**
     * Returns the underlying primitive storage manager.
     */
    public PrimitiveStorageManager getPrimitiveStorageManager()
    {
	return asm.getPrimitiveStorageManager();
    }

    /**
     * Returns the underlying units' storage manager.
     */
    public UnitsStorageManager getUnitsStorageManager()
    {
	return usm;
    }

    /**
     * Returns the underlying storage tools' manager.
     */
    public StorageToolsManager getStorageToolsManager()
    {
	return stm;
    }

    /**
     * Returns the unit size of LOB segment.
     */
    public int getLobSegmentUnitSize()
    {
	return LOB_SEGMENT_UNIT_SIZE;
    }

    /**
     * Returns the unit size of Parents' segment.
     */
    public int getParentsSegmentUnitSize()
    {
	return PARENTS_SEGMENT_UNIT_SIZE;
    }

    /**
     * Returns the used page size.
     */
    public int getPageSize()
    {
	return getPrimitiveStorageManager().getPageSize();
    }
}
