package DbStorage;

import storage.UnitAddress;

/**
 * A <code>DbUnitAddress</code> represents a unit address
 * adapted for its storage and representation in semistructured database 
 * structures.
 */
public class DbUnitAddress extends UnitAddress
{
    /** The size in bytes of a <code>DbUnitAddress</code>. */
    static public final int SIZE=12;

    /** Represents a NULL (invalid) unit address. */
    static public final DbUnitAddress NULL=new DbUnitAddress();

    byte[] content=new byte[12];

    /** Constructs a NULL unit address. */
    public DbUnitAddress()
    {
	setNull();
    } 
            
    /** 
     * Constructs a unit address whith specified space and unit numbers. 
     * @param spaceNumber The space number of the unit address.
     * @param unitNumber The unit number of the unit address.
     */ 
    public DbUnitAddress(int spaceNumber, long unitNumber)
    {
	set(spaceNumber,unitNumber);
    }

    /** 
     * Contructs a unit address from its binary representation .
     * @param b The byte array from which the binary representation is got.
     * @param off The position of <code>b</code> from which the binary representation starts.
     */
    public DbUnitAddress(byte[] b, int off)
    {
	set(b,off);
    }

    /**
     * Returns the space number of this unit address.
     * @return The space number.
     */
    public int getSpaceNumber()
    {
	int r=0;
	r |= ((0x80&content[7])==0 ? (byte)0 : (byte)1);
	for (int i=8; i<12; i++)
	    if (i<11)
		r |= ((int)(content[i])&0xFF)<<(i*8+1);
	    else
		r |= ((int)(content[i])&0x3F)<<(i*8+1);
	return r;
    } 
       
    /**
     * Returns the unit number of this unit address.
     * @return The unit number.
     */
    public long getUnitNumber()
    {
	long r=0;
	for (int i=0; i<8; i++)
	    if (i<7)
		r |= ((long)(content[i])&0xFFl)<<(i*8);
	    else
		r |= ((long)(content[i])&0x7Fl)<<(i*8);
	return r;
    }
            
    /**
     * Sets the space and unit number of the unit address.
     * @param spaceNumber The space number.
     * @param unitNumber The unit number.
     */
    public void set(int spaceNumber, long unitNumber) 
    {
	if (spaceNumber<0 || unitNumber<0)
	    setNull();
	else
	    {
		for (int i=0; i<8; i++)
		    if (i<7)
			content[i]=(byte)(0xFF&(unitNumber>>(i*8)));
		    else
			content[i]=(byte)(0x7F&(unitNumber>>(i*8)));
		content[7] = (byte)(content[7] | ((0x1&spaceNumber)==0 ? 0 : 0x80));
		for (int i=0; i<4; i++)
		    if (i<4)
			content[8+i]=(byte)(0xFF&(spaceNumber>>(i*8+1)));
		    else
			content[8+i]=(byte)(0x3F&(spaceNumber>>(i*8+1)));
	    }
    }        

    /** 
     * Sets a unit address from its binary representation .
     * @param b The byte array from which the binary representation is got.
     * @param off The position of <code>b</code> from which the binary representation starts.
     */
    public void set(byte[] b, int off) 
    {
	for (int i=0; i<content.length; i++)
	    content[i]=b[i+off];
    }

    /**
     * Copies the content of a unit address in this one.
     * @param source The unit address to copy.
     */
    public void copyFrom(DbUnitAddress source) 
    {
	for (int i=0; i<content.length; i++)
	    content[i]=source.content[i];
    }

    /**
     * Sets this unit address as NULL.
     */
    public void setNull()
    {
	for (int i=0; i<content.length; i++)
	    content[i]=(byte)0xFF;
    }
    
    /**
     * Checks if a unit address has a NULL value.
     * @return <code>true</code> if this unit address is NULL, <code>false</code> otherwise.
     */
    public boolean isNull()
    {
	return (0x8&content[11]) != 0;
    }

    /**
     * Returns the binary representation of this unit address.
     * @return A byte array where the binary representation is. 
     */
    public byte[] getContent()
    {
	return content;
    }

    /**
     * Checks if this unit address is equals to other one.
     * @param addr The unit address to compare.
     */ 
    public boolean equals(DbUnitAddress addr)
    {
	if (!this.isNull() && !addr.isNull())
	    return this.getSpaceNumber()==addr.getSpaceNumber() && this.getUnitNumber()==addr.getUnitNumber();
	else
	    return (this.isNull() && addr.isNull());
    }
}
