package DbStorage.driver;

import DbStorage.*;
import storage.*;
import java.io.*;

/**
 * This class drives a content extension, this
 * is stored in the CONTENT segment. A content extension
 * is used when the datum's content can be fully stored
 * in one unit. 
 */
public final class ContentExtensionDriver
{
    static public final int CONTENT_EXTENSION_OFFSET=0;
    static public final int CONTENT_OFFSET=CONTENT_EXTENSION_OFFSET+DbUnitAddress.SIZE; //12
    
    /**
     * Creates a content extension.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param newUnitAddress The address of the created structure in the CONTENT segment.
     */
    static public void createContentExtension(DbStorageManager dbsm, long transactionNumber,
					      DbUnitAddress newUnitAddress)
	throws StorageException, IOException
    {
	dbsm.addUnit(transactionNumber,DbStorageManager.CONTENT_SEGMENT,newUnitAddress);
    }
    
    /**
     * Gets the content extension's address (in the CONTENT segment) of the specified extension.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content extension.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getContentExtension(DbStorageManager dbsm, long transactionNumber,
					   DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
			       unitAddress,CONTENT_EXTENSION_OFFSET,destiny);
    }
    
    /**
     * Sets the content extension's address (in the CONTENT segment) of the specified extension.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content extension.
     * @param source Where the unit address is stored.
     */
    static public void setContentExtension(DbStorageManager dbsm, long transactionNumber,
					   DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
				unitAddress,CONTENT_EXTENSION_OFFSET,source);
    }
    
    /**
     * Gets the extension's content.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content extension.
     * @param maxLength The maximun number of bytes to be got.
     * @param b Where the content will be stored.
     * @param off The begining position of b where the content will be stored.
     */
    static public int getContent(DbStorageManager dbsm, long transactionNumber,
				 DbUnitAddress unitAddress, int maxLength, byte[] b, int off)
	throws StorageException, IOException
    {
	int l=DbStorageManager.CONTENT_SEGMENT_UNIT_SIZE-CONTENT_OFFSET;
	if (l>maxLength)
	    l=maxLength;
	dbsm.readUnit(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
		      unitAddress,CONTENT_OFFSET,l,b,off);
	return l;
    }
    
    /**
     * Sets the extension's content.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content extension.
     * @param maxLength The maximun number of bytes to be setted.
     * @param b Where the content is stored.
     * @param off The begining position of b where the content is stored.
     */
    static public int setContent(DbStorageManager dbsm, long transactionNumber,
				 DbUnitAddress unitAddress, int maxLength, byte[] b, int off)
	throws StorageException, IOException
    {
	int l=DbStorageManager.CONTENT_SEGMENT_UNIT_SIZE-CONTENT_OFFSET;
	if (l>maxLength)
	    l=maxLength;
	dbsm.writeUnit(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
		       unitAddress,CONTENT_OFFSET,l,b,off);
	return l;
    }
}
