package DbStorage.driver;

import DbStorage.*;
import storage.*;
import java.io.*;

/**
 * This class drives a content node for building contents of non-primitive data, this
 * is stored in the CONTENT segment. A content node is used to keep 4 data structures
 * for the non-primitive data content, this structures are: A binary tree ordered by id,
 * a linked list ordered by id, a binary tree ordered by label and
 * a linked list ordered by label; those structures try to provide a high performance
 * for data management.   
 */
public final class ContentNodeDriver
{
    static public final int PREVIOUS_ID_NODE_OFFSET=0;
    static public final int NEXT_ID_NODE_OFFSET=PREVIOUS_ID_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int LEFT_ID_NODE_OFFSET=NEXT_ID_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int RIGHT_ID_NODE_OFFSET=LEFT_ID_NODE_OFFSET+DbUnitAddress.SIZE;	
    static public final int PARENT_ID_NODE_OFFSET=RIGHT_ID_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int ID_HEIGHT_OFFSET=PARENT_ID_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int PREVIOUS_LABEL_NODE_OFFSET=ID_HEIGHT_OFFSET+1;
    static public final int NEXT_LABEL_NODE_OFFSET=PREVIOUS_LABEL_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int LEFT_LABEL_NODE_OFFSET=NEXT_LABEL_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int RIGHT_LABEL_NODE_OFFSET=LEFT_LABEL_NODE_OFFSET+DbUnitAddress.SIZE;	
    static public final int PARENT_LABEL_NODE_OFFSET=RIGHT_LABEL_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int LABEL_HEIGHT_OFFSET=PARENT_LABEL_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int ID_OFFSET=LABEL_HEIGHT_OFFSET+1;
    static public final int LABEL_CONTENT_EXTENSION_OFFSET=ID_OFFSET+PhysicalID.SIZE;
    static public final int LABEL_SIZE_OFFSET=LABEL_CONTENT_EXTENSION_OFFSET+DbUnitAddress.SIZE;
    static public final int LABEL_CONTENT_OFFSET=LABEL_SIZE_OFFSET+4;	
    
       
    /**
     * Creates a content node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param newUnitAddress The address of the created structure in the CONTENT segment.
     */
    static public void createContentNode(DbStorageManager dbsm, long transactionNumber,
					 DbUnitAddress newUnitAddress)
	throws StorageException, IOException
    {
	dbsm.addUnit(transactionNumber,DbStorageManager.CONTENT_SEGMENT,newUnitAddress);
    }
	
    /**
     * Gets the previous node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getPreviousIdNode(DbStorageManager dbsm, long transactionNumber,
					 DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
			       unitAddress,PREVIOUS_ID_NODE_OFFSET,destiny);
    }
    
    /**
     * Sets the previous node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setPreviousIdNode(DbStorageManager dbsm, long transactionNumber,
					 DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
				unitAddress,PREVIOUS_ID_NODE_OFFSET,source);
    }
    
    /**
     * Gets the next node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getNextIdNode(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
			       unitAddress,NEXT_ID_NODE_OFFSET,destiny);
    }
    
    /**
     * Sets the next node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setNextIdNode(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
				unitAddress,NEXT_ID_NODE_OFFSET,source);
    }
    
    /**
     * Gets the left node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getLeftIdNode(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
			       unitAddress,LEFT_ID_NODE_OFFSET,destiny);
    }
    
    /**
     * Sets the left node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setLeftIdNode(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
				unitAddress,LEFT_ID_NODE_OFFSET,source);
    }
    
    /**
     * Gets the right node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getRightIdNode(DbStorageManager dbsm, long transactionNumber,
				      DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
			       unitAddress,RIGHT_ID_NODE_OFFSET,destiny);
    }
    
    /**
     * Sets the right node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setRightIdNode(DbStorageManager dbsm, long transactionNumber,
				      DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
				unitAddress,RIGHT_ID_NODE_OFFSET,source);
    }
    
    /**
     * Gets the parent node's address of the specified node, in the tree order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getParentIdNode(DbStorageManager dbsm, long transactionNumber,
				       DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
			       unitAddress,PARENT_ID_NODE_OFFSET,destiny);
    }
	
    /**
     * Sets the previous node's address of the specified node, in the tree order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setParentIdNode(DbStorageManager dbsm, long transactionNumber,
				       DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
				unitAddress,PARENT_ID_NODE_OFFSET,source);
    }
    
    /**
     * Gets the height of the specified node in the tree order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @return The height in the tree id order.
     */
    static public byte getIdHeight(DbStorageManager dbsm, long transactionNumber,
				   DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	return dbsm.readByte(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
			     unitAddress,ID_HEIGHT_OFFSET);
    }
    
    /**
     * Sets the height of the specified node in the tree order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source The height in the tree id order.
     */
    static public void setIdHeight(DbStorageManager dbsm, long transactionNumber,
				   DbUnitAddress unitAddress, byte source)
	throws StorageException, IOException
    {
	dbsm.writeByte(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
		       unitAddress,ID_HEIGHT_OFFSET,source);
    }
    
    /**
     * Gets the previous node's address of the specified node, in the order by label.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getPreviousLabelNode(DbStorageManager dbsm, long transactionNumber,
					    DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,unitAddress,
			       PREVIOUS_LABEL_NODE_OFFSET,destiny);
    }
    
    /**
     * Sets the previous node's address of the specified node, in the order by label.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setPreviousLabelNode(DbStorageManager dbsm, long transactionNumber,
					    DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,unitAddress,
				PREVIOUS_LABEL_NODE_OFFSET,source);
    }
    
    /**
     * Gets the next node's address of the specified node, in the order by label.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getNextLabelNode(DbStorageManager dbsm, long transactionNumber,
					DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
			       unitAddress,NEXT_LABEL_NODE_OFFSET,destiny);
    }
    
    /**
     * Sets the next node's address of the specified node, in the order by label.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setNextLabelNode(DbStorageManager dbsm, long transactionNumber,
					DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
				unitAddress,NEXT_LABEL_NODE_OFFSET,source);
    }
    
    /**
     * Gets the left node's address of the specified node, in the order by label.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getLeftLabelNode(DbStorageManager dbsm, long transactionNumber,
					DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
			       unitAddress,LEFT_LABEL_NODE_OFFSET,destiny);
    }
    
    /**
     * Sets the left node's address of the specified node, in the order by label.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setLeftLabelNode(DbStorageManager dbsm, long transactionNumber,
					DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
				unitAddress,LEFT_LABEL_NODE_OFFSET,source);
    }
    
    /**
     * Gets the right node's address of the specified node, in the order by label.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getRightLabelNode(DbStorageManager dbsm, long transactionNumber,
					 DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
			       unitAddress,RIGHT_LABEL_NODE_OFFSET,destiny);
    }
    
    /**
     * Sets the right node's address of the specified node, in the order by label.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setRightLabelNode(DbStorageManager dbsm, long transactionNumber,
					 DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
				unitAddress,RIGHT_LABEL_NODE_OFFSET,source);
    }
    
    /**
     * Gets the parent node's address of the specified node, in the tree order by label.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getParentLabelNode(DbStorageManager dbsm, long transactionNumber,
					  DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
			       unitAddress,PARENT_LABEL_NODE_OFFSET,destiny);
    }
    
    /**
     * Sets the parent node's address of the specified node, in the tree order by label.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setParentLabelNode(DbStorageManager dbsm, long transactionNumber,
					  DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
				unitAddress,PARENT_LABEL_NODE_OFFSET,source);
    }

    /**
     * Gets the height of the specified node in the tree order by label.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @return The height in the tree label order.
     */
    static public byte getLabelHeight(DbStorageManager dbsm, long transactionNumber,
				      DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	return dbsm.readByte(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
			     unitAddress,LABEL_HEIGHT_OFFSET);
    }
    
    /**
     * Sets the height of the specified node in the tree order by label.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source The height in the tree label order.
     */
    static public void setLabelHeight(DbStorageManager dbsm, long transactionNumber,
				      DbUnitAddress unitAddress, byte source)
	throws StorageException, IOException
    {
	dbsm.writeByte(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
		       unitAddress,LABEL_HEIGHT_OFFSET,source);
    }
    
    /**
     * Gets the physical id in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the id will be stored.
     */
    static public void getId(DbStorageManager dbsm, long transactionNumber,
			     DbUnitAddress unitAddress, PhysicalID destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
			       unitAddress,ID_OFFSET,destiny);
    }
    
    /**
     * Sets the physical id in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the id is stored.
     */
    static public void setId(DbStorageManager dbsm, long transactionNumber,
			     DbUnitAddress unitAddress, PhysicalID source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
				unitAddress,ID_OFFSET,source);
    }
    
    /**
     * Gets the label content extension's address (in the CONTENT segment) of the label
     * in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the label content extension.
     * @param destiny Where the address will be stored.
     */
    static public void getLabelContentExtension(DbStorageManager dbsm, long transactionNumber,
						DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,unitAddress,
			       LABEL_CONTENT_EXTENSION_OFFSET,destiny);
    }
    
    /**
     * Sets the label content extension's address (in the CONTENT segment) of the label
     * in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the label content extension.
     * @param source Where the address will is stored.
     */
    static public void setLabelContentExtension(DbStorageManager dbsm, long transactionNumber,
						DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.CONTENT_SEGMENT,unitAddress,
				LABEL_CONTENT_EXTENSION_OFFSET,source);
    }
    
    /**
     * Returns the size of the label in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @return The label's size.
     */
    static public int getLabelSize(DbStorageManager dbsm, long transactionNumber,
				   DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	return dbsm.readInt(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
			    unitAddress,LABEL_SIZE_OFFSET);
    }
    
    /**
     * Sets the size of the label in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source The label's size.
     */
    static public void setLabelSize(DbStorageManager dbsm, long transactionNumber,
				    DbUnitAddress unitAddress, int source)
	throws StorageException, IOException
    {
	dbsm.writeInt(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
		      unitAddress,LABEL_SIZE_OFFSET,source);
    }
    
    /**
     * Gets the part of label's content stored in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param maxLength The maximun number of bytes to be got.
     * @param b Where the content will be stored.
     * @param off The begining position of b where the label's content will be stored.
     */
    static public int getLabelContent(DbStorageManager dbsm, long transactionNumber,
				      DbUnitAddress unitAddress, int maxLength, byte[] b, int off)
	throws StorageException, IOException
    {
	int l=DbStorageManager.CONTENT_SEGMENT_UNIT_SIZE-LABEL_CONTENT_OFFSET;
	if (l>maxLength)
	    l=maxLength;
	dbsm.readUnit(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
		      unitAddress,LABEL_CONTENT_OFFSET,l,b,off);
	return l;
    }
    
    /**
     * Sets the part of label's content stored in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param maxLength The maximun number of bytes to be got.
     * @param b Where the content is stored.
     * @param off The begining position of b where the label's content is stored.
     */
    static public int setLabelContent(DbStorageManager dbsm, long transactionNumber,
				      DbUnitAddress unitAddress, int maxLength, byte[] b, int off)
	throws StorageException, IOException
    {
	int l=DbStorageManager.CONTENT_SEGMENT_UNIT_SIZE-LABEL_CONTENT_OFFSET;
	if (l>maxLength)
	    l=maxLength;
	dbsm.writeUnit(transactionNumber,DbStorageManager.CONTENT_SEGMENT,
		       unitAddress,LABEL_CONTENT_OFFSET,l,b,off);
	return l;
    }
}
