package DbStorage.driver;

import DbStorage.*;
import storage.*;
import java.io.*;

/**
 * This class drives a dictionary content node for building the dictionary content, this
 * is stored in the DICTIONARY segment. A dictionary contents node keeps 4 data structures
 * to storage the dictionary entries.
 */
public class DictionaryContentNodeDriver
{
    static public final int PREVIOUS_ID_NODE_OFFSET=0;
    static public final int NEXT_ID_NODE_OFFSET=PREVIOUS_ID_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int LEFT_ID_NODE_OFFSET=NEXT_ID_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int RIGHT_ID_NODE_OFFSET=LEFT_ID_NODE_OFFSET+DbUnitAddress.SIZE;	
    static public final int PARENT_ID_NODE_OFFSET=RIGHT_ID_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int ID_HEIGHT_OFFSET=PARENT_ID_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int PREVIOUS_NAME_NODE_OFFSET=ID_HEIGHT_OFFSET+1;
    static public final int NEXT_NAME_NODE_OFFSET=PREVIOUS_NAME_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int LEFT_NAME_NODE_OFFSET=NEXT_NAME_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int RIGHT_NAME_NODE_OFFSET=LEFT_NAME_NODE_OFFSET+DbUnitAddress.SIZE;	
    static public final int PARENT_NAME_NODE_OFFSET=RIGHT_NAME_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int NAME_HEIGHT_OFFSET=PARENT_NAME_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int ID_OFFSET=NAME_HEIGHT_OFFSET+1;
    static public final int NAME_CONTENT_EXTENSION_OFFSET=ID_OFFSET+PhysicalID.SIZE;
    static public final int NAME_SIZE_OFFSET=NAME_CONTENT_EXTENSION_OFFSET+DbUnitAddress.SIZE;
    static public final int NAME_CONTENT_OFFSET=NAME_SIZE_OFFSET+4;	
    
    
    /**
     * Creates a dictionary content node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param newUnitAddress The address of the created structure in the DICTIONARY segment.
     */
    static public void createDictionaryContentNode(DbStorageManager dbsm, long transactionNumber,
						   DbUnitAddress newUnitAddress)
	throws StorageException, IOException
    {
	dbsm.addUnit(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,newUnitAddress);
    }
    
    /**
     * Gets the previous node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getPreviousIdNode(DbStorageManager dbsm, long transactionNumber,
					 DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			       unitAddress,PREVIOUS_ID_NODE_OFFSET,destiny);
    }
	
    /**
     * Sets the previous node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param source Where the unit address is stored.
     */
    static public void setPreviousIdNode(DbStorageManager dbsm, long transactionNumber,
					 DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
				unitAddress,PREVIOUS_ID_NODE_OFFSET,source);
    }
    
    /**
     * Gets the next node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getNextIdNode(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			       unitAddress,NEXT_ID_NODE_OFFSET,destiny);
    }
    
    /**
     * Sets the next node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param source Where the unit address is stored.
     */
    static public void setNextIdNode(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
				unitAddress,NEXT_ID_NODE_OFFSET,source);
    }
	
    /**
     * Gets the left node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getLeftIdNode(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			       unitAddress,LEFT_ID_NODE_OFFSET,destiny);
    }
	
    /**
     * Sets the left node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param source Where the unit address is stored.
     */
    static public void setLeftIdNode(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
				unitAddress,LEFT_ID_NODE_OFFSET,source);
    }
    
    /**
     * Gets the right node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getRightIdNode(DbStorageManager dbsm, long transactionNumber,
				      DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			       unitAddress,RIGHT_ID_NODE_OFFSET,destiny);
    }
	
    /**
     * Sets the right node's address of the specified node, in the order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param source Where the unit address is stored.
     */
    static public void setRightIdNode(DbStorageManager dbsm, long transactionNumber,
				      DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
				unitAddress,RIGHT_ID_NODE_OFFSET,source);
    }

    /**
     * Gets the parent node's address of the specified node, in the tree order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getParentIdNode(DbStorageManager dbsm, long transactionNumber,
				       DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			       unitAddress,PARENT_ID_NODE_OFFSET,destiny);
    }
	
    /**
     * Sets the previous node's address of the specified node, in the tree order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param source Where the unit address is stored.
     */
    static public void setParentIdNode(DbStorageManager dbsm, long transactionNumber,
				       DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
				unitAddress,PARENT_ID_NODE_OFFSET,source);
    }

    /**
     * Gets the height of the specified node in the tree order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @return The height in the tree id order.
     */
    static public byte getIdHeight(DbStorageManager dbsm, long transactionNumber,
				   DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	return dbsm.readByte(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			     unitAddress,ID_HEIGHT_OFFSET);
    }
	
    /**
     * Sets the height of the specified node in the tree order by id.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param source The height in the tree id order.
     */
    static public void setIdHeight(DbStorageManager dbsm, long transactionNumber,
				   DbUnitAddress unitAddress, byte source)
	throws StorageException, IOException
    {
	dbsm.writeByte(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
		       unitAddress,ID_HEIGHT_OFFSET,source);
    }

    /**
     * Gets the previous node's address of the specified node, in the order by name.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getPreviousNameNode(DbStorageManager dbsm, long transactionNumber,
					   DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,unitAddress,
			       PREVIOUS_NAME_NODE_OFFSET,destiny);
    }
	
    /**
     * Sets the previous node's address of the specified node, in the order by name.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param source Where the unit address is stored.
     */
    static public void setPreviousNameNode(DbStorageManager dbsm, long transactionNumber,
					   DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,unitAddress,
				PREVIOUS_NAME_NODE_OFFSET,source);
    }

    /**
     * Gets the next node's address of the specified node, in the order by name.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getNextNameNode(DbStorageManager dbsm, long transactionNumber,
				       DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			       unitAddress,NEXT_NAME_NODE_OFFSET,destiny);
    }
	
    /**
     * Sets the next node's address of the specified node, in the order by name.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param source Where the unit address is stored.
     */
    static public void setNextNameNode(DbStorageManager dbsm, long transactionNumber,
				       DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
				unitAddress,NEXT_NAME_NODE_OFFSET,source);
    }
    
    /**
     * Gets the left node's address of the specified node, in the order by name.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getLeftNameNode(DbStorageManager dbsm, long transactionNumber,
				       DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			       unitAddress,LEFT_NAME_NODE_OFFSET,destiny);
    }
	
    /**
     * Sets the left node's address of the specified node, in the order by name.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param source Where the unit address is stored.
     */
    static public void setLeftNameNode(DbStorageManager dbsm, long transactionNumber,
				       DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
				unitAddress,LEFT_NAME_NODE_OFFSET,source);
    }

    /**
     * Gets the right node's address of the specified node, in the order by name.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getRightNameNode(DbStorageManager dbsm, long transactionNumber,
					DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			       unitAddress,RIGHT_NAME_NODE_OFFSET,destiny);
    }
	
    /**
     * Sets the right node's address of the specified node, in the order by name.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param source Where the unit address is stored.
     */
    static public void setRightNameNode(DbStorageManager dbsm, long transactionNumber,
					DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	    dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
				    unitAddress,RIGHT_NAME_NODE_OFFSET,source);
    }
    
    /**
     * Gets the parent node's address of the specified node, in the tree order by name.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getParentNameNode(DbStorageManager dbsm, long transactionNumber,
					 DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			       unitAddress,PARENT_NAME_NODE_OFFSET,destiny);
    }
	
    /**
     * Sets the parent node's address of the specified node, in the tree order by name.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param source Where the unit address is stored.
     */
    static public void setParentNameNode(DbStorageManager dbsm, long transactionNumber,
					 DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
				unitAddress,PARENT_NAME_NODE_OFFSET,source);
    }

    /**
     * Gets the height of the specified node in the tree order by name.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @return The height in the tree name order.
     */
    static public byte getNameHeight(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	return dbsm.readByte(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			     unitAddress,NAME_HEIGHT_OFFSET);
    }
	
    /**
     * Sets the height of the specified node in the tree order by name.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param source The height in the tree name order.
     */
    static public void setNameHeight(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, byte source)
	throws StorageException, IOException
    {
	dbsm.writeByte(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
		       unitAddress,NAME_HEIGHT_OFFSET,source);
    }
    
    /**
     * Gets the physical id in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param id Where the id will be stored.
     */
    static public void getId(DbStorageManager dbsm, long transactionNumber,
			     DbUnitAddress unitAddress, PhysicalID id)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			       unitAddress,ID_OFFSET,id);
    }
    
    /**
     * Sets the physical id in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param id Where the id is stored.
     */
    static public void setId(DbStorageManager dbsm, long transactionNumber,
			     DbUnitAddress unitAddress, PhysicalID id)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
				unitAddress,ID_OFFSET,id);
    }
    
    /**
     * Gets the name content extension's address (in the DICTIONARY segment) of the name
     * in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the name content extension.
     * @param destiny Where the address will be stored.
	 */
    static public void getNameContentExtension(DbStorageManager dbsm, long transactionNumber,
					       DbUnitAddress unitAddress, DbUnitAddress destiny)
	    throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,unitAddress,
			       NAME_CONTENT_EXTENSION_OFFSET,destiny);
    }
    
    /**
     * Sets the name content extension's address (in the DICTIONARY segment) of the name
     * in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the name content extension.
     * @param source Where the address will is stored.
     */
    static public void setNameContentExtension(DbStorageManager dbsm, long transactionNumber,
					       DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,unitAddress,
				NAME_CONTENT_EXTENSION_OFFSET,source);
    }
    
    /**
     * Returns the size of the name in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @return The name's size.
     */
    static public int getNameSize(DbStorageManager dbsm, long transactionNumber,
				  DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	return dbsm.readInt(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			    unitAddress,NAME_SIZE_OFFSET);
    }
    
    /**
     * Sets the size of the name in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param source The name's size.
     */
    static public void setNameSize(DbStorageManager dbsm, long transactionNumber,
				   DbUnitAddress unitAddress, int source)
	throws StorageException, IOException
    {
	dbsm.writeInt(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
		      unitAddress,NAME_SIZE_OFFSET,source);
    }
    
    /**
     * Gets the part of name's content stored in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param maxLength The maximun number of bytes to be got.
     * @param b Where the content will be stored.
     * @param off The begining position of b where the name's content will be stored.
     */
    static public int getNameContent(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, int maxLength, byte[] b, int off)
	throws StorageException, IOException
    {
	int l=DbStorageManager.DICTIONARY_SEGMENT_UNIT_SIZE-NAME_CONTENT_OFFSET;
	if (l>maxLength)
	    l=maxLength;
	dbsm.readUnit(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
		      unitAddress,NAME_CONTENT_OFFSET,l,b,off);
	return l;
    }
    
    /**
     * Sets the part of name's content stored in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary node.
     * @param maxLength The maximun number of bytes to be got.
     * @param b Where the content is stored.
     * @param off The begining position of b where the name's content is stored.
     */
    static public int setNameContent(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, int maxLength, byte[] b, int off)
	throws StorageException, IOException
    {
	int l=DbStorageManager.DICTIONARY_SEGMENT_UNIT_SIZE-NAME_CONTENT_OFFSET;
	if (l>maxLength)
	    l=maxLength;
	dbsm.writeUnit(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
		       unitAddress,NAME_CONTENT_OFFSET,l,b,off);
	return l;
    }
}
