package DbStorage.driver;

import DbStorage.*;
import storage.*;
import java.io.*;

/**
 * This class drives a dictionary header, this
 * is stored in the DICTIONARY segment. In a similar 
 * way of content' nodes the dictionary header keeps 4 data structures,
 * the header points tho these structures.  
 */
public final class DictionaryHeaderDriver
{
    static public final int IDS_ROOT_OFFSET=0; 
    static public final int IDS_HEAD_OFFSET=IDS_ROOT_OFFSET+DbUnitAddress.SIZE;
    static public final int NAMES_ROOT_OFFSET=IDS_HEAD_OFFSET+DbUnitAddress.SIZE;
    static public final int NAMES_HEAD_OFFSET=NAMES_ROOT_OFFSET+DbUnitAddress.SIZE;
    
    /**
     * Creates a dictionary header.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address where the structure will be stored in the DICTIONARY segment.
     */
    static public void createDictionaryHeader(DbStorageManager dbsm, long transactionNumber,
					      DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	dbsm.addSpecifiedUnit(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,unitAddress);
    }
    
    /**
     * Gets the root's address (in the DICTIONARY segment) of the id ordered tree collection 
     * of labeled SSD contained in the dictionary.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary header.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getIdsRoot(DbStorageManager dbsm, long transactionNumber,
				  DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			       unitAddress,IDS_ROOT_OFFSET,destiny);
    }
    
    /**
     * Sets the root's address (in the DICTIONARY segment) of the id ordered tree collection 
     * of labeled SSD contained in the dictionary.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary header.
     * @param source Where the unit address is stored.
     */
    static public void setIdsRoot(DbStorageManager dbsm, long transactionNumber,
				  DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
				unitAddress,IDS_ROOT_OFFSET,source);
    }
    
    /**
     * Gets the head's address  (in the DICTIONARY segment) of the id ordered list collection 
     * of labeled SSD contained in the dictionary.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary header.
     * @param destiny Where the unit address will be stored.
     */
    static public void getIdsHead(DbStorageManager dbsm, long transactionNumber,
				  DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			       unitAddress,IDS_HEAD_OFFSET,destiny);
    }
    
    /**
     * Sets the head's address (in the DICTIONARY segment) of the id ordered list collection 
     * of labeled SSD contained in the dictionary.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary header.
     * @param source Where the unit address is stored.
     */
    static public void setIdsHead(DbStorageManager dbsm, long transactionNumber,
				  DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
				unitAddress,IDS_HEAD_OFFSET,source);
    }
    
    /**
     * Gets the root's address (in the DICTIONARY segment) of the label ordered tree collection 
     * of labeled SSD contained in the dictionary.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary header.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getNamesRoot(DbStorageManager dbsm, long transactionNumber,
				    DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			       unitAddress,NAMES_ROOT_OFFSET,destiny);
    }
    
    /**
     * Sets the root's address (in the DICTIONARY segment) of the label ordered tree collection 
     * of labeled SSD contained in the dictionary.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary header.
     * @param source Where the unit address is stored.
     */
    static public void setNamesRoot(DbStorageManager dbsm, long transactionNumber,
				    DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
				unitAddress,NAMES_ROOT_OFFSET,source);
    }
    
    /**
     * Gets the head's address (in the DICTIONARY segment) of the label ordered tree collection 
     * of labeled SSD contained in the dictionary.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary header.
     * @param destiny Where the unit address will be stored.
     */
    static public void getNamesHead(DbStorageManager dbsm, long transactionNumber,
				    DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
			       unitAddress,NAMES_HEAD_OFFSET,destiny);
    }
    
    /**
     * Sets the head's address (in the DICTIONARY segment) of the label ordered tree collection 
     * of labeled SSD contained in the dictionary.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the dictionary header.
     * @param source Where the unit address is stored.
     */
    static public void setNamesHead(DbStorageManager dbsm, long transactionNumber,
				    DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,
				unitAddress,NAMES_HEAD_OFFSET,source);
    }
}
