package DbStorage.driver;

import DbStorage.*;
import storage.*;
import java.io.*;

/**
 * This class drives a semistructured datum descriptor for a non-primitive datum.
 * A non-primitive datum uses 4 data structures to storage its content, a
 * non-primitive datum descriptor poits to these structures.
 */
public final class NonPrimitiveSsdDescriptorDriver extends SsdDescriptorDriver
{
    static public final int IDS_ROOT_OFFSET=TYPE_OFFSET+4; // 48
    static public final int IDS_HEAD_OFFSET=IDS_ROOT_OFFSET+DbUnitAddress.SIZE; // 60
    static public final int LABELS_ROOT_OFFSET=IDS_HEAD_OFFSET+DbUnitAddress.SIZE; // 72
    static public final int LABELS_HEAD_OFFSET=LABELS_ROOT_OFFSET+DbUnitAddress.SIZE; // 84 
    
    /**
     * Gets the root's address (in the content segment) of the id ordered tree collection 
     * of labeled SSD contained in the specified datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getIdsRoot(DbStorageManager dbsm, long transactionNumber,
				  DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
			       unitAddress,IDS_ROOT_OFFSET,destiny);
    }
    
    /**
     * Sets the root's address (in the content segment) of the id ordered tree collection 
     * of labeled SSD contained in the specified datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param source Where the unit address is stored.
     */
    static public void setIdsRoot(DbStorageManager dbsm, long transactionNumber,
				  DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
				unitAddress,IDS_ROOT_OFFSET,source);
    }
    
    /**
     * Gets the head's address  (in the content segment) of the id ordered list collection 
     * of labeled SSD contained in the specified datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getIdsHead(DbStorageManager dbsm, long transactionNumber,
				  DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
			       unitAddress,IDS_HEAD_OFFSET,destiny);
    }
    
    /**
     * Sets the head's address (in the content segment) of the id ordered list collection 
     * of labeled SSD contained in the specified datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param source Where the unit address is stored.
     */
    static public void setIdsHead(DbStorageManager dbsm, long transactionNumber,
				  DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
				unitAddress,IDS_HEAD_OFFSET,source);
    }
    
    /**
     * Gets the root's address (in the content segment) of the label ordered tree collection 
     * of labeled SSD contained in the specified datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getLabelsRoot(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
			       unitAddress,LABELS_ROOT_OFFSET,destiny);
    }
    
    /**
     * Sets the root's address (in the content segment) of the label ordered tree collection 
     * of labeled SSD contained in the specified datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param source Where the unit address is stored.
     */
    static public void setLabelsRoot(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
				unitAddress,LABELS_ROOT_OFFSET,source);
    }
    
    /**
     * Gets the head's address (in the content segment) of the label ordered tree collection 
     * of labeled SSD contained in the specified datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getLabelsHead(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
			       unitAddress,LABELS_HEAD_OFFSET,destiny);
    }
    
    /**
     * Sets the head's address (in the content segment) of the label ordered tree collection 
     * of labeled SSD contained in the specified datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param source Where the unit address is stored.
     */
    static public void setLabelsHead(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
				unitAddress,LABELS_HEAD_OFFSET,source);
    }
}
