package DbStorage.driver;

import DbStorage.*;
import storage.*;
import java.io.*;

/**
 * This class drives a content node for building contents of id collections,
 * these correspond to parents' collections,
 * this is stored in the PARENTS segment. 
 * A node is used to keep 2 data structures
 * a linked list and an AA-tree.
 */
public final class ParentsNodeDriver
{
    static public final int PREVIOUS_NODE_OFFSET=0;
    static public final int NEXT_NODE_OFFSET=PREVIOUS_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int LEFT_NODE_OFFSET=NEXT_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int RIGHT_NODE_OFFSET=LEFT_NODE_OFFSET+DbUnitAddress.SIZE;	
    static public final int PARENT_NODE_OFFSET=RIGHT_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int HEIGHT_OFFSET=PARENT_NODE_OFFSET+DbUnitAddress.SIZE;
    static public final int ID_OFFSET=HEIGHT_OFFSET+1;

    /**
     * Creates a parents node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param newUnitAddress The address of the created structure in the PARENTS segment.
     */
    static public void createParentsNode(DbStorageManager dbsm, long transactionNumber,
					 DbUnitAddress newUnitAddress)
	throws StorageException, IOException
    {
	dbsm.addUnit(transactionNumber,DbStorageManager.PARENTS_SEGMENT,newUnitAddress);
    }


    /**
     * Gets the previous node's address of the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getPreviousNode(DbStorageManager dbsm, long transactionNumber,
					 DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.PARENTS_SEGMENT,
			       unitAddress,PREVIOUS_NODE_OFFSET,destiny);
    }
    
    /**
     * Sets the previous node's address of the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setPreviousNode(DbStorageManager dbsm, long transactionNumber,
					 DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.PARENTS_SEGMENT,
				unitAddress,PREVIOUS_NODE_OFFSET,source);
    }
    
    /**
     * Gets the next node's address of the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getNextNode(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.PARENTS_SEGMENT,
			       unitAddress,NEXT_NODE_OFFSET,destiny);
    }
    
    /**
     * Sets the next node's address of the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setNextNode(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.PARENTS_SEGMENT,
				unitAddress,NEXT_NODE_OFFSET,source);
    }
    
    /**
     * Gets the left node's address of the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getLeftNode(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.PARENTS_SEGMENT,
			       unitAddress,LEFT_NODE_OFFSET,destiny);
    }
    
    /**
     * Sets the left node's address of the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setLeftNode(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.PARENTS_SEGMENT,
				unitAddress,LEFT_NODE_OFFSET,source);
    }
    
    /**
     * Gets the right node's address of the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getRightNode(DbStorageManager dbsm, long transactionNumber,
				      DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.PARENTS_SEGMENT,
			       unitAddress,RIGHT_NODE_OFFSET,destiny);
    }
    
    /**
     * Sets the right node's address of the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setRightNode(DbStorageManager dbsm, long transactionNumber,
				      DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.PARENTS_SEGMENT,
				unitAddress,RIGHT_NODE_OFFSET,source);
    }
    
    /**
     * Gets the parent node's address of the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getParentNode(DbStorageManager dbsm, long transactionNumber,
				       DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.PARENTS_SEGMENT,
			       unitAddress,PARENT_NODE_OFFSET,destiny);
    }
	
    /**
     * Sets the previous node's address of the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the unit address is stored.
     */
    static public void setParentNode(DbStorageManager dbsm, long transactionNumber,
				       DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.PARENTS_SEGMENT,
				unitAddress,PARENT_NODE_OFFSET,source);
    }
    
    /**
     * Gets the height of the specified node in the tree.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @return The height in the tree.
     */
    static public byte getHeight(DbStorageManager dbsm, long transactionNumber,
				   DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	return dbsm.readByte(transactionNumber,DbStorageManager.PARENTS_SEGMENT,
			     unitAddress,HEIGHT_OFFSET);
    }
    
    /**
     * Sets the height of the specified node in the tree.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source The height in the tree.
     */
    static public void setHeight(DbStorageManager dbsm, long transactionNumber,
				   DbUnitAddress unitAddress, byte source)
	throws StorageException, IOException
    {
	dbsm.writeByte(transactionNumber,DbStorageManager.PARENTS_SEGMENT,
		       unitAddress,HEIGHT_OFFSET,source);
    }

    /**
     * Gets the physical id in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param destiny Where the id will be stored.
     */
    static public void getId(DbStorageManager dbsm, long transactionNumber,
			     DbUnitAddress unitAddress, PhysicalID destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.PARENTS_SEGMENT,
			       unitAddress,ID_OFFSET,destiny);
    }
    
    /**
     * Sets the physical id in the specified node.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the content node.
     * @param source Where the id is stored.
     */
    static public void setId(DbStorageManager dbsm, long transactionNumber,
			     DbUnitAddress unitAddress, PhysicalID source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.PARENTS_SEGMENT,
				unitAddress,ID_OFFSET,source);
    }
}
