package DbStorage.driver;

import DbStorage.*;
import storage.*;
import java.io.*;

/**
 * This class drives a semistructured datum descriptor for a simple primitive datum.
 */
public final class SimplePrimitiveSsdDescriptorDriver extends SsdDescriptorDriver
{
    static public final int CONTENT_EXTENSION_OFFSET=TYPE_OFFSET+4; // 48
    static public final int CONTENT_SIZE_OFFSET=CONTENT_EXTENSION_OFFSET+DbUnitAddress.SIZE; // 60
    static public final int CONTENT_OFFSET=CONTENT_SIZE_OFFSET+4; // 64
    
    /**
     * Gets the content extension's address (in the CONTENT segment for simple data or 
     * LOB segment for LOB data) of the specified primitive datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getContentExtension(DbStorageManager dbsm, long transactionNumber,
				    DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
			       unitAddress,CONTENT_EXTENSION_OFFSET,destiny);
    }
    
    /**
     * Sets the Content Extension's address (in the CONTENT segment for simple data or 
     * LOB segment for LOB data) of the specified primitive datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param source Where the unit address is stored.
     */
    static public void setContentExtension(DbStorageManager dbsm, long transactionNumber,
				    DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
				unitAddress,CONTENT_EXTENSION_OFFSET,source);
    }
    
    /**
     * Returns the content's size of the descripted simple primitive datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @return The size.
     */
    static public int getContentSize(DbStorageManager dbsm, long transactionNumber, 
			      DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	return dbsm.readInt(transactionNumber,DbStorageManager.ID_SEGMENT,
			    unitAddress,CONTENT_SIZE_OFFSET);
    }
    
    /**
     * Sets the content's size of the descripted simple primitive datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param source The size.
     */
    static public void setContentSize(DbStorageManager dbsm, long transactionNumber,
			       DbUnitAddress unitAddress, int source)
	throws StorageException, IOException
    {
	dbsm.writeInt(transactionNumber,DbStorageManager.ID_SEGMENT,
		      unitAddress,CONTENT_SIZE_OFFSET,source);
    }

    /**
     * Gets the part of datum's content stored in the specified descriptor.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param maxLength The maximun number of bytes to be got.
     * @param b Where the content will be stored.
     * @param off The begining position of b where the content will be stored.
     * @return The number of recovered bytes.
     */
    static public int getContent(DbStorageManager dbsm, long transactionNumber,
			  DbUnitAddress unitAddress, int maxLength, byte[] b, int off)
	throws StorageException, IOException
    {
	int l=DbStorageManager.ID_SEGMENT_UNIT_SIZE-CONTENT_OFFSET;
	if (l>maxLength)
	    l=maxLength;
	dbsm.readUnit(transactionNumber,DbStorageManager.ID_SEGMENT,
		      unitAddress,CONTENT_OFFSET,l,b,off);
	return l;
    }
    
    /**
     * Sets the part of datum's content stored in the specified descriptor.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param maxLength The maximun number of bytes to be setted.
     * @param b Where the content is stored.
     * @param off The begining position of b where the content is stored.
     * @return The number of wroten bytes.
     */
    static public int setContent(DbStorageManager dbsm, long transactionNumber,
			  DbUnitAddress unitAddress, int maxLength, byte[] b, int off)
	throws StorageException, IOException
    {
	int l=DbStorageManager.ID_SEGMENT_UNIT_SIZE-CONTENT_OFFSET;
	if (l>maxLength)
	    l=maxLength;
	dbsm.writeUnit(transactionNumber,DbStorageManager.ID_SEGMENT,
		       unitAddress,CONTENT_OFFSET,l,b,off);
	return l;
    }
}
