package DbStorage.driver;

import DbStorage.*;
import storage.*;
import java.io.*;

/**
 * This class drives a semistructured datum descriptor, this
 * is stored in the ID segment.
 */
public class SsdDescriptorDriver
{
    static public final int DICTIONARY_REFERENCES_OFFSET=0; // 0
    static public final int FIRST_PARENT_OFFSET=8; // 8
    static public final int ROOT_PARENTS_COLLECTION_OFFSET=FIRST_PARENT_OFFSET+PhysicalID.SIZE; // 20
    static public final int HEAD_PARENTS_COLLECTION_OFFSET=ROOT_PARENTS_COLLECTION_OFFSET+PhysicalID.SIZE; // 32
    static public final int TYPE_OFFSET=HEAD_PARENTS_COLLECTION_OFFSET+DbUnitAddress.SIZE; // 44
    
    
    /**
     * Creates a SSD descriptor.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param newUnitAddress The address of the created structure in the ID segment.
     */
    static public void createSsdDescriptor(DbStorageManager dbsm, long transactionNumber,
					   DbUnitAddress newUnitAddress)
	throws StorageException, IOException
    {
	dbsm.addUnit(transactionNumber,DbStorageManager.ID_SEGMENT,newUnitAddress);
    }
    
    /**
     * Returns the number of times that the dictionary references the descripted datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @return The number of datum's references in the dictionary.
     */
    static public long getDictionaryReferences(DbStorageManager dbsm, long transactionNumber, 
					       DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	return dbsm.readLong(transactionNumber,DbStorageManager.ID_SEGMENT,
			     unitAddress,DICTIONARY_REFERENCES_OFFSET);
    }
    
    /**
     * Sets the number of times that the dictionary references the descripted datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param source The number of datum's references in the dictionary.
     */
    static public void setDictionaryReferences(DbStorageManager dbsm, long transactionNumber, 
					       DbUnitAddress unitAddress, long source)
	throws StorageException, IOException
    {
	dbsm.writeLong(transactionNumber,DbStorageManager.ID_SEGMENT,
		       unitAddress,DICTIONARY_REFERENCES_OFFSET,source);
    }
    
    /**
     * Gets the physical id of the first parent.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param destiny Where the retrieved Id will be stored.
     */
    static public void getFirstParent(DbStorageManager dbsm, long transactionNumber, 
				      DbUnitAddress unitAddress, PhysicalID destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
			       unitAddress,FIRST_PARENT_OFFSET,destiny);
    }
    
    /**
     * Sets the physical id of the first parent.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param source Where the Id is stored.
     */
    static public void setFirstParent(DbStorageManager dbsm, long transactionNumber, 
				      DbUnitAddress unitAddress, PhysicalID source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
				unitAddress,FIRST_PARENT_OFFSET,source);
    }
    
    /**
     * Gets the unit addres in the parent's segment where the parents' collection is.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getRootParentsCollection(DbStorageManager dbsm, long transactionNumber, 
						DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
			       unitAddress,ROOT_PARENTS_COLLECTION_OFFSET,destiny);
    }
    
    /**
     * Sets the unit addres in the parents segment where the parents' collection is.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param source Where the unit address is stored.
     */
    static public void setRootParentsCollection(DbStorageManager dbsm, long transactionNumber, 
						DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
				unitAddress,ROOT_PARENTS_COLLECTION_OFFSET,source);
    }
    
    /**
     * Gets the unit addres in the parent's segment where the parents' collection is.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param destiny Where the retrieved unit address will be stored.
     */
    static public void getHeadParentsCollection(DbStorageManager dbsm, long transactionNumber, 
						DbUnitAddress unitAddress, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	dbsm.readDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
			       unitAddress,HEAD_PARENTS_COLLECTION_OFFSET,destiny);
    }
    
    /**
     * Sets the unit addres in the parents segment where the parents' collection is.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param source Where the unit address is stored.
     */
    static public void setHeadParentsCollection(DbStorageManager dbsm, long transactionNumber, 
						DbUnitAddress unitAddress, DbUnitAddress source)
	throws StorageException, IOException
    {
	dbsm.writeDbUnitAddress(transactionNumber,DbStorageManager.ID_SEGMENT,
				unitAddress,HEAD_PARENTS_COLLECTION_OFFSET,source);
    }

    /**
     * Returns the type of the descripted datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @return The type.
     */
    static public int getType(DbStorageManager dbsm, long transactionNumber, 
			      DbUnitAddress unitAddress)
	throws StorageException, IOException
    {
	return dbsm.readInt(transactionNumber,DbStorageManager.ID_SEGMENT,unitAddress,TYPE_OFFSET);
    }
    
    /**
     * Sets the type of the descripted datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param unitAddress The address of the SSD descriptor.
     * @param source The type.
     */
    static public void setType(DbStorageManager dbsm, long transactionNumber, 
			       DbUnitAddress unitAddress, int source)
	throws StorageException, IOException
    {
	dbsm.writeInt(transactionNumber,DbStorageManager.ID_SEGMENT,unitAddress,TYPE_OFFSET,source);
    }
}
