package DbStorage.manager;

import DbStorage.*;
import DbStorage.driver.*;
import storage.*;
import java.io.*;

/**
 * This class manages the content of the dictionary.
 * The dictionary collection is implemented using 4 data structures combined in one, 
 * 2 AA-trees and 2 double linked list are used, for ordering the collection
 * by identificator and ssd-table name.
 */
public final class DictionaryManager
{
    /**
     * Sets the initial values for the dictionary header.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     */
    static public void formatHeader(DbStorageManager dbsm, long transactionNumber,
				    DbUnitAddress header)
	throws StorageException, IOException
    {
	DictionaryHeaderDriver.setIdsRoot(dbsm,transactionNumber,header,DbUnitAddress.NULL);
	DictionaryHeaderDriver.setIdsHead(dbsm,transactionNumber,header,DbUnitAddress.NULL);
	DictionaryHeaderDriver.setNamesRoot(dbsm,transactionNumber,header,DbUnitAddress.NULL);
	DictionaryHeaderDriver.setNamesHead(dbsm,transactionNumber,header,DbUnitAddress.NULL);
    }
    
    static private void createNode(DbStorageManager dbsm, long transactionNumber,
				   String name, PhysicalID id,
				   DbUnitAddress newUnitAddress)
	throws StorageException, IOException, UnavailableUnitException
    {
	newUnitAddress.setNull();
	DictionaryContentNodeDriver.createDictionaryContentNode(dbsm,transactionNumber,newUnitAddress);
	if (newUnitAddress.isNull())
	    throw new UnavailableUnitException();
	DictionaryContentNodeDriver.setPreviousIdNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setNextIdNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setRightIdNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setParentIdNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setIdHeight(dbsm,transactionNumber,newUnitAddress,(byte)1);
	DictionaryContentNodeDriver.setPreviousNameNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setNextNameNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setLeftNameNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setRightNameNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setParentNameNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setNameHeight(dbsm,transactionNumber,newUnitAddress,(byte)1);
	DictionaryContentNodeDriver.setId(dbsm,transactionNumber,newUnitAddress,id);
	
	byte[] nameContent=DataConvertion.stringToBytes(name);
	DictionaryContentNodeDriver.setNameSize(dbsm,transactionNumber,newUnitAddress,nameContent.length);
	
	int max=nameContent.length;
	int ini=0;
	int len=DictionaryContentNodeDriver.setNameContent(dbsm,transactionNumber,newUnitAddress,max,nameContent,ini);
	
	ini=ini+len;
	max=max-len;
	if (max<=0)
	    DictionaryContentNodeDriver.setNameContentExtension(dbsm,transactionNumber,
						  newUnitAddress,DbUnitAddress.NULL);
	else //if (max>0)
	    {
		DbUnitAddress u=new DbUnitAddress();
		u.copyFrom(newUnitAddress);
		DbUnitAddress v=new DbUnitAddress();
		boolean first=true;
		while (max>0)
		    {
			DictionaryContentExtensionDriver.
			    createDictionaryContentExtension(dbsm,transactionNumber,v);
			if (v.isNull())
			    throw new UnavailableUnitException();
			len=DictionaryContentExtensionDriver.setContent(dbsm,transactionNumber,v,max,nameContent,ini);
			if (first)
			    DictionaryContentNodeDriver.setNameContentExtension(dbsm,transactionNumber,u,v);
			else
			    DictionaryContentExtensionDriver.
				setDictionaryContentExtension(dbsm,transactionNumber,u,v);
			ini=ini+len;
			max=max-len;
			u.copyFrom(v);
			first=false;
		    }
		if (first)
		    DictionaryContentNodeDriver.setNameContentExtension(dbsm,transactionNumber,
									   u,DbUnitAddress.NULL);
		else
		    DictionaryContentExtensionDriver.
			setDictionaryContentExtension(dbsm,transactionNumber,u,DbUnitAddress.NULL);
	    }
    }

    // Pre: lssd is valid.
    static private void createNode(DbStorageManager dbsm, long transactionNumber,
				   DictionaryEntry entry,
				   DbUnitAddress newUnitAddress)
	throws StorageException, IOException, UnavailableUnitException
    {
	createNode(dbsm,transactionNumber,entry.getName(),entry.getID(),newUnitAddress);
    }

    /**
     * Returns the name of a specified node of the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param nodeAddress The addres of the node in the DICTIONARY segment.
     * @return The name in a <code>String</code> object.
     */
    static public String getNameFromNode(DbStorageManager dbsm, long transactionNumber,
					 DbUnitAddress nodeAddress)
	throws StorageException, IOException
    {
	int siz=DictionaryContentNodeDriver.getNameSize(dbsm,transactionNumber,nodeAddress);
	byte[] dataAux=new byte[siz];
	int ini=0;
	int len=DictionaryContentNodeDriver.getNameContent(dbsm,transactionNumber,nodeAddress,siz,dataAux,ini);
	DbUnitAddress u=new DbUnitAddress();
	DictionaryContentNodeDriver.getNameContentExtension(dbsm,transactionNumber,nodeAddress,u);
	while (!u.isNull())
	    {
		siz-=len;
		ini+=len;
		len=DictionaryContentExtensionDriver.getContent(dbsm,transactionNumber,u,siz,dataAux,ini);
		DictionaryContentExtensionDriver.
		    getDictionaryContentExtension(dbsm,transactionNumber,u,u);
	    }
	return DataConvertion.bytesToString(dataAux);
    }

    /**
     * Returns the dictionary entry of a specified node of the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param nodeAddress The addres of the node in the DICTIONARY segment.
     * @return The dictionary entry in a <code>DictionaryEntry<code> onject.
     */
    static public DictionaryEntry getDictionaryEntryFromNode(DbStorageManager dbsm, long transactionNumber,
							     DbUnitAddress nodeAddress)
	throws StorageException, IOException
    {
	PhysicalID aux=new PhysicalID();
	DictionaryContentNodeDriver.getId(dbsm,transactionNumber,nodeAddress,aux);
	return new DictionaryEntry(getNameFromNode(dbsm,transactionNumber,nodeAddress),aux);
    }


    // Methods for Ids' Collections

    static private void insertAfterInIdList(DbStorageManager dbsm, long transactionNumber, DbUnitAddress header,
					    DbUnitAddress current, DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress targetNext=new DbUnitAddress();
	DictionaryContentNodeDriver.getNextIdNode(dbsm,transactionNumber,target,targetNext);
	DictionaryContentNodeDriver.setNextIdNode(dbsm,transactionNumber,current,targetNext);
	DictionaryContentNodeDriver.setPreviousIdNode(dbsm,transactionNumber,current,target);
	if (!targetNext.isNull())
	    DictionaryContentNodeDriver.setPreviousIdNode(dbsm,transactionNumber,targetNext,current);
	DictionaryContentNodeDriver.setNextIdNode(dbsm,transactionNumber,target,current);
    }

    static private void insertBeforeInIdList(DbStorageManager dbsm, long transactionNumber, DbUnitAddress header,
					     DbUnitAddress current, DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress idHead=new DbUnitAddress();
	DictionaryHeaderDriver.getIdsHead(dbsm,transactionNumber,header,idHead);
	if (idHead.equals(target))
	    DictionaryHeaderDriver.setIdsHead(dbsm,transactionNumber,header,current);
	
	DbUnitAddress targetPrevious=new DbUnitAddress();
	DictionaryContentNodeDriver.getPreviousIdNode(dbsm,transactionNumber,target,targetPrevious);
	DictionaryContentNodeDriver.setPreviousIdNode(dbsm,transactionNumber,current,targetPrevious);
	DictionaryContentNodeDriver.setNextIdNode(dbsm,transactionNumber,current,target);
	if (!targetPrevious.isNull())
	    DictionaryContentNodeDriver.setNextIdNode(dbsm,transactionNumber,targetPrevious,current);
	DictionaryContentNodeDriver.setPreviousIdNode(dbsm,transactionNumber,target,current);
    }

    static private void removeFromIdList(DbStorageManager dbsm, long transactionNumber, DbUnitAddress header,
					 DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress idHead=new DbUnitAddress();
	DictionaryHeaderDriver.getIdsHead(dbsm,transactionNumber,header,idHead);
	DbUnitAddress targetNext=new DbUnitAddress();
	DictionaryContentNodeDriver.getNextIdNode(dbsm,transactionNumber,target,targetNext);
	DbUnitAddress targetPrevious=new DbUnitAddress();
	DictionaryContentNodeDriver.getPreviousIdNode(dbsm,transactionNumber,target,targetPrevious);

	if (idHead.equals(target))
	    DictionaryHeaderDriver.setIdsHead(dbsm,transactionNumber,header,targetNext);
	if (!targetPrevious.isNull())
	    DictionaryContentNodeDriver.setNextIdNode(dbsm,transactionNumber,targetPrevious,targetNext);
	if (!targetNext.isNull())
	    DictionaryContentNodeDriver.setPreviousIdNode(dbsm,transactionNumber,targetNext,targetPrevious);
	
	DictionaryContentNodeDriver.setPreviousIdNode(dbsm,transactionNumber,target,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setNextIdNode(dbsm,transactionNumber,target,DbUnitAddress.NULL);
    }

    // Pre: node!=null && node.left!=null
    static private DbUnitAddress rotateWithLeftChildInIdTree(DbStorageManager dbsm, long transactionNumber,
							     DbUnitAddress header, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress m=new DbUnitAddress();
	DictionaryContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,node,m);
	
	DbUnitAddress aux=new DbUnitAddress();
	DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,m,aux);
	DictionaryContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    DictionaryContentNodeDriver.setParentIdNode(dbsm,transactionNumber,aux,node);
	DictionaryContentNodeDriver.setRightIdNode(dbsm,transactionNumber,m,node);
	
	DictionaryContentNodeDriver.getParentIdNode(dbsm,transactionNumber,node,aux);
	if (aux.isNull())
	    DictionaryHeaderDriver.setIdsRoot(dbsm,transactionNumber,header,m);
	else
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		DictionaryContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(node))
		    DictionaryContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,aux,m);
		else
		    DictionaryContentNodeDriver.setRightIdNode(dbsm,transactionNumber,aux,m);
	    }
	DictionaryContentNodeDriver.setParentIdNode(dbsm,transactionNumber,m,aux);
	DictionaryContentNodeDriver.setParentIdNode(dbsm,transactionNumber,node,m);
	return m;
    }

    // Pre: node!=null && node.right!=null
     static private DbUnitAddress rotateWithRightChildInIdTree(DbStorageManager dbsm, long transactionNumber,
							       DbUnitAddress header, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress m=new DbUnitAddress();
	DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,node,m);
	
	DbUnitAddress aux=new DbUnitAddress();
	DictionaryContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,m,aux);
	DictionaryContentNodeDriver.setRightIdNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    DictionaryContentNodeDriver.setParentIdNode(dbsm,transactionNumber,aux,node);
	DictionaryContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,m,node);
	
	DictionaryContentNodeDriver.getParentIdNode(dbsm,transactionNumber,node,aux);
	if (aux.isNull())
	    DictionaryHeaderDriver.setIdsRoot(dbsm,transactionNumber,header,m);
	else
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		DictionaryContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(node))
		    DictionaryContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,aux,m);
		else
		    DictionaryContentNodeDriver.setRightIdNode(dbsm,transactionNumber,aux,m);
	    }
	DictionaryContentNodeDriver.setParentIdNode(dbsm,transactionNumber,m,aux);
	DictionaryContentNodeDriver.setParentIdNode(dbsm,transactionNumber,node,m);
	return m;
    }

    // Pre: node!=null
    static private DbUnitAddress skewInIdTree(DbStorageManager dbsm, long transactionNumber,
					      DbUnitAddress header, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress aux=new DbUnitAddress();
	DictionaryContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,node,aux);
	
	if (!aux.isNull() && 
	    (DictionaryContentNodeDriver.getIdHeight(dbsm,transactionNumber,node)==
	     DictionaryContentNodeDriver.getIdHeight(dbsm,transactionNumber,aux)))
	    return rotateWithLeftChildInIdTree(dbsm,transactionNumber,header,node);
	else
	    return node;
    }

    // Pre: node!=null
    static private DbUnitAddress splitInIdTree(DbStorageManager dbsm, long transactionNumber,
					       DbUnitAddress header, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress aux=new DbUnitAddress();
	DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,aux,aux2);
		if (!aux2.isNull() &&
		    (DictionaryContentNodeDriver.getIdHeight(dbsm,transactionNumber,node)==
		     DictionaryContentNodeDriver.getIdHeight(dbsm,transactionNumber,aux2)))
		    {
			DbUnitAddress aux3=rotateWithRightChildInIdTree(dbsm,transactionNumber,header,node);
			DictionaryContentNodeDriver.setIdHeight(dbsm,transactionNumber,aux3,
						      (byte)(DictionaryContentNodeDriver
							     .getIdHeight(dbsm,transactionNumber,aux3)+1));
			return aux3;
		    }
	    }
	return node;
    }

    static private DbUnitAddress findNearestInIdTree(DbStorageManager dbsm, long transactionNumber,
						     DbUnitAddress header, DictionaryEntry entry)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	current.setNull();
	DbUnitAddress target=new DbUnitAddress();
	DictionaryHeaderDriver.getIdsRoot(dbsm,transactionNumber,header,target);
	DictionaryEntry aux;
	while (!target.isNull())
	    {
		current.copyFrom(target);
		aux=getDictionaryEntryFromNode(dbsm,transactionNumber,current);
		int r=DictionaryEntry.ByIDComparator.compare(entry,aux);
		if (r<0)
		    DictionaryContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,current,target);
		else if (r>0)
		    DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,current,target);
		else
		    return current;
	    }
	return current;
    }

    static private void adjustIdTreeAfterInsertion(DbStorageManager dbsm, long transactionNumber,
						   DbUnitAddress header, DbUnitAddress node)
	throws StorageException, IOException
    {
	while (!node.isNull())
	    {
		node=skewInIdTree(dbsm,transactionNumber,header,node);
		node=splitInIdTree(dbsm,transactionNumber,header,node);
		DictionaryContentNodeDriver.getParentIdNode(dbsm,transactionNumber,node,node);
	    }
    }

    static private void reinsertNodeInIdTree(DbStorageManager dbsm, long transactionNumber,
					     DbUnitAddress header, DbUnitAddress isolated, DbUnitAddress deleted)
	throws StorageException, IOException
    {
	if (!isolated.equals(deleted))
	    {
		DictionaryContentNodeDriver.setIdHeight(dbsm,transactionNumber,isolated,
							DictionaryContentNodeDriver.
							getIdHeight(dbsm,transactionNumber,deleted));
		DbUnitAddress aux=new DbUnitAddress();
		DictionaryContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,deleted,aux);
		DictionaryContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,isolated,aux);
		DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,deleted,aux);
		DictionaryContentNodeDriver.setRightIdNode(dbsm,transactionNumber,isolated,aux);
		DictionaryContentNodeDriver.getParentIdNode(dbsm,transactionNumber,deleted,aux);
		DictionaryContentNodeDriver.setParentIdNode(dbsm,transactionNumber,isolated,aux);
		
		if (aux.isNull())
		    DictionaryHeaderDriver.setIdsRoot(dbsm,transactionNumber,header,isolated);
		else
		    {
			DbUnitAddress aux2=new DbUnitAddress();
			DictionaryContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,aux,aux2);
			if (aux2.equals(deleted))
			    DictionaryContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,aux,isolated);
			else
			    DictionaryContentNodeDriver.setRightIdNode(dbsm,transactionNumber,aux,isolated);
		    }

		DictionaryContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,deleted,aux);
		if (!aux.isNull())
		    DictionaryContentNodeDriver.setParentIdNode(dbsm,transactionNumber,aux,isolated);
		DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,deleted,aux);
		if (!aux.isNull())
		    DictionaryContentNodeDriver.setParentIdNode(dbsm,transactionNumber,aux,isolated);	
	    }
	
	DictionaryContentNodeDriver.setIdHeight(dbsm,transactionNumber,deleted,(byte)0);
	DictionaryContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setRightIdNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setParentIdNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
    }


    static private void removeFromIdTree(DbStorageManager dbsm, long transactionNumber,
					 DbUnitAddress header, DbUnitAddress deleted)
	throws StorageException, IOException
    {
	DbUnitAddress victim=new DbUnitAddress();
	DbUnitAddress adjusted=new DbUnitAddress();
	DbUnitAddress aux=new DbUnitAddress();
	DbUnitAddress aux2=new DbUnitAddress();

	DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,deleted,aux);
	if (!aux.isNull())
	    DictionaryContentNodeDriver.getNextIdNode(dbsm,transactionNumber,deleted,victim);
	else
	    victim.copyFrom(deleted);
	     
	DictionaryContentNodeDriver.getParentIdNode(dbsm,transactionNumber,victim,aux);
	if (aux.equals(deleted))
	    adjusted.copyFrom(victim);
	else
	    DictionaryContentNodeDriver.getParentIdNode(dbsm,transactionNumber,victim,adjusted);
	if (aux.isNull())
	    DictionaryHeaderDriver.setIdsRoot(dbsm,transactionNumber,header,DbUnitAddress.NULL);
	else
	    {
		DictionaryContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(victim))
		    {
			DbUnitAddress aux3=new DbUnitAddress();
			DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,victim,aux3);
			DictionaryContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,aux,aux3);
			DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,victim,aux2);
			if (!aux2.isNull())
			    DictionaryContentNodeDriver.setParentIdNode(dbsm,transactionNumber,aux2,aux);
		    }
		else
		    {
			DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,victim,aux2);
			if (!aux2.isNull())
			    DictionaryContentNodeDriver.setParentIdNode(dbsm,transactionNumber,aux2,aux);
			DictionaryContentNodeDriver.setRightIdNode(dbsm,transactionNumber,aux,aux2);
		    }
	    }
	reinsertNodeInIdTree(dbsm,transactionNumber,header,victim,deleted);

	while (!adjusted.isNull())
	    {
		DictionaryContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,adjusted,aux);
		DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,adjusted,aux2);
		byte adjustedHeight=DictionaryContentNodeDriver.getIdHeight(dbsm,transactionNumber,adjusted);
		if (((!aux.isNull() && 
		      DictionaryContentNodeDriver.getIdHeight(dbsm,transactionNumber,aux)<adjustedHeight-1) ||
		     (aux.isNull() && 0<adjustedHeight-1)) ||
		    ((!aux2.isNull() && 
		      DictionaryContentNodeDriver.getIdHeight(dbsm,transactionNumber,aux2)<adjustedHeight-1) ||
		     (aux2.isNull() && 0<adjustedHeight-1)))
		    {
			DictionaryContentNodeDriver.setIdHeight(dbsm,transactionNumber,adjusted,
						      (byte)(DictionaryContentNodeDriver.getIdHeight(dbsm,
											   transactionNumber,
											   adjusted)-1));
			adjustedHeight--;
			if (!aux2.isNull() && 
			    DictionaryContentNodeDriver.getIdHeight(dbsm,transactionNumber,aux2)>adjustedHeight)
			    DictionaryContentNodeDriver.setIdHeight(dbsm,transactionNumber,aux2,adjustedHeight);
			adjusted=skewInIdTree(dbsm,transactionNumber,header,adjusted);
			DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,adjusted,aux2);
			if (!aux2.isNull())
			    {
				DictionaryContentNodeDriver.setRightIdNode(dbsm,transactionNumber,adjusted,
								 skewInIdTree(dbsm,transactionNumber,header,aux2));
				DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,adjusted,aux2);
				DbUnitAddress aux3=new DbUnitAddress();
				DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,aux2,aux3);
				if (!aux3.isNull())
				    DictionaryContentNodeDriver.setRightIdNode(dbsm,transactionNumber,aux2,
								     skewInIdTree(dbsm,transactionNumber,header,aux3));
			    }
			adjusted=splitInIdTree(dbsm,transactionNumber,header,adjusted);
			DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,adjusted,aux2);
			if (!aux2.isNull())
			    DictionaryContentNodeDriver.setRightIdNode(dbsm,transactionNumber,adjusted,
							     splitInIdTree(dbsm,transactionNumber,header,aux2));
		    }
		DictionaryContentNodeDriver.getParentIdNode(dbsm,transactionNumber,adjusted,adjusted);
	    }	
    }

    // Pre: node is isolated
    static private void insertById(DbStorageManager dbsm, long transactionNumber,
				   DbUnitAddress header, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress aux=new DbUnitAddress();
	DictionaryHeaderDriver.getIdsRoot(dbsm,transactionNumber,header,aux);
	if (aux.isNull())
	    {
		DictionaryHeaderDriver.setIdsHead(dbsm,transactionNumber,header,node);
		DictionaryHeaderDriver.setIdsRoot(dbsm,transactionNumber,header,node);
	    }
	else
	    {
		DictionaryEntry entry=getDictionaryEntryFromNode(dbsm,transactionNumber,node);
		aux=findNearestInIdTree(dbsm,transactionNumber,header,entry);
		DictionaryEntry entryAux=getDictionaryEntryFromNode(dbsm,transactionNumber,aux);
		int r=DictionaryEntry.ByIDComparator.compare(entry,entryAux);
		if (r!=0)
		    {
			if (r<0)
			    {
				DictionaryContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,aux,node);
				insertBeforeInIdList(dbsm,transactionNumber,header,node,aux);
			    }
			else
			    {
				DictionaryContentNodeDriver.setRightIdNode(dbsm,transactionNumber,aux,node);
				insertAfterInIdList(dbsm,transactionNumber,header,node,aux);
			    }
			DictionaryContentNodeDriver.setParentIdNode(dbsm,transactionNumber,node,aux);
			adjustIdTreeAfterInsertion(dbsm,transactionNumber,header,aux);
		    }
	    }
    }



    // Methods for Names' Collections

    static private void insertAfterInNameList(DbStorageManager dbsm, long transactionNumber, DbUnitAddress header,
					      DbUnitAddress current, DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress targetNext=new DbUnitAddress();
	DictionaryContentNodeDriver.getNextNameNode(dbsm,transactionNumber,target,targetNext);
	DictionaryContentNodeDriver.setNextNameNode(dbsm,transactionNumber,current,targetNext);
	DictionaryContentNodeDriver.setPreviousNameNode(dbsm,transactionNumber,current,target);
	if (!targetNext.isNull())
	    DictionaryContentNodeDriver.setPreviousNameNode(dbsm,transactionNumber,targetNext,current);
	DictionaryContentNodeDriver.setNextNameNode(dbsm,transactionNumber,target,current);
    }

    static private void insertBeforeInNameList(DbStorageManager dbsm, long transactionNumber, DbUnitAddress header,
					       DbUnitAddress current, DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress nameHead=new DbUnitAddress();
	DictionaryHeaderDriver.getNamesHead(dbsm,transactionNumber,header,nameHead);
	if (nameHead.equals(target))
	    DictionaryHeaderDriver.setNamesHead(dbsm,transactionNumber,header,current);
	
	DbUnitAddress targetPrevious=new DbUnitAddress();
	DictionaryContentNodeDriver.getPreviousNameNode(dbsm,transactionNumber,target,targetPrevious);
	DictionaryContentNodeDriver.setPreviousNameNode(dbsm,transactionNumber,current,targetPrevious);
	DictionaryContentNodeDriver.setNextNameNode(dbsm,transactionNumber,current,target);
	if (!targetPrevious.isNull())
	    DictionaryContentNodeDriver.setNextNameNode(dbsm,transactionNumber,targetPrevious,current);
	DictionaryContentNodeDriver.setPreviousNameNode(dbsm,transactionNumber,target,current);
    }

    static private void removeFromNameList(DbStorageManager dbsm, long transactionNumber, DbUnitAddress header,
					   DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress nameHead=new DbUnitAddress();
	DictionaryHeaderDriver.getNamesHead(dbsm,transactionNumber,header,nameHead);
	DbUnitAddress targetNext=new DbUnitAddress();
	DictionaryContentNodeDriver.getNextNameNode(dbsm,transactionNumber,target,targetNext);
	DbUnitAddress targetPrevious=new DbUnitAddress();
	DictionaryContentNodeDriver.getPreviousNameNode(dbsm,transactionNumber,target,targetPrevious);

	if (nameHead.equals(target))
	    DictionaryHeaderDriver.setNamesHead(dbsm,transactionNumber,header,targetNext);
	if (!targetPrevious.isNull())
	    DictionaryContentNodeDriver.setNextNameNode(dbsm,transactionNumber,targetPrevious,targetNext);
	if (!targetNext.isNull())
	    DictionaryContentNodeDriver.setPreviousNameNode(dbsm,transactionNumber,targetNext,targetPrevious);
	
	DictionaryContentNodeDriver.setPreviousNameNode(dbsm,transactionNumber,target,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setNextNameNode(dbsm,transactionNumber,target,DbUnitAddress.NULL);
    }

    // Pre: node!=null && node.left!=null
    static private DbUnitAddress rotateWithLeftChildInNameTree(DbStorageManager dbsm, long transactionNumber,
							       DbUnitAddress header, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress m=new DbUnitAddress();
	DictionaryContentNodeDriver.getLeftNameNode(dbsm,transactionNumber,node,m);
	
	DbUnitAddress aux=new DbUnitAddress();
	DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,m,aux);
	DictionaryContentNodeDriver.setLeftNameNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    DictionaryContentNodeDriver.setParentNameNode(dbsm,transactionNumber,aux,node);
	DictionaryContentNodeDriver.setRightNameNode(dbsm,transactionNumber,m,node);
	
	DictionaryContentNodeDriver.getParentNameNode(dbsm,transactionNumber,node,aux);
	if (aux.isNull())
	    DictionaryHeaderDriver.setNamesRoot(dbsm,transactionNumber,header,m);
	else
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		DictionaryContentNodeDriver.getLeftNameNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(node))
		    DictionaryContentNodeDriver.setLeftNameNode(dbsm,transactionNumber,aux,m);
		else
		    DictionaryContentNodeDriver.setRightNameNode(dbsm,transactionNumber,aux,m);
	    }
	DictionaryContentNodeDriver.setParentNameNode(dbsm,transactionNumber,m,aux);
	DictionaryContentNodeDriver.setParentNameNode(dbsm,transactionNumber,node,m);
	return m;
    }

    // Pre: node!=null && node.right!=null
     static private DbUnitAddress rotateWithRightChildInNameTree(DbStorageManager dbsm, long transactionNumber,
								 DbUnitAddress header, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress m=new DbUnitAddress();
	DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,node,m);
	
	DbUnitAddress aux=new DbUnitAddress();
	DictionaryContentNodeDriver.getLeftNameNode(dbsm,transactionNumber,m,aux);
	DictionaryContentNodeDriver.setRightNameNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    DictionaryContentNodeDriver.setParentNameNode(dbsm,transactionNumber,aux,node);
	DictionaryContentNodeDriver.setLeftNameNode(dbsm,transactionNumber,m,node);
	
	DictionaryContentNodeDriver.getParentNameNode(dbsm,transactionNumber,node,aux);
	if (aux.isNull())
	    DictionaryHeaderDriver.setNamesRoot(dbsm,transactionNumber,header,m);
	else
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		DictionaryContentNodeDriver.getLeftNameNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(node))
		    DictionaryContentNodeDriver.setLeftNameNode(dbsm,transactionNumber,aux,m);
		else
		    DictionaryContentNodeDriver.setRightNameNode(dbsm,transactionNumber,aux,m);
	    }
	DictionaryContentNodeDriver.setParentNameNode(dbsm,transactionNumber,m,aux);
	DictionaryContentNodeDriver.setParentNameNode(dbsm,transactionNumber,node,m);
	return m;
    }

    // Pre: node!=null
    static private DbUnitAddress skewInNameTree(DbStorageManager dbsm, long transactionNumber,
						DbUnitAddress header, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress aux=new DbUnitAddress();
	DictionaryContentNodeDriver.getLeftNameNode(dbsm,transactionNumber,node,aux);
	
	if (!aux.isNull() && 
	    (DictionaryContentNodeDriver.getNameHeight(dbsm,transactionNumber,node)==
	     DictionaryContentNodeDriver.getNameHeight(dbsm,transactionNumber,aux)))
	    return rotateWithLeftChildInNameTree(dbsm,transactionNumber,header,node);
	else
	    return node;
    }

    // Pre: node!=null
    static private DbUnitAddress splitInNameTree(DbStorageManager dbsm, long transactionNumber,
						 DbUnitAddress header, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress aux=new DbUnitAddress();
	DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,aux,aux2);
		if (!aux2.isNull() &&
		    (DictionaryContentNodeDriver.getNameHeight(dbsm,transactionNumber,node)==
		     DictionaryContentNodeDriver.getNameHeight(dbsm,transactionNumber,aux2)))
		    {
			DbUnitAddress aux3=rotateWithRightChildInNameTree(dbsm,transactionNumber,header,node);
			DictionaryContentNodeDriver.setNameHeight(dbsm,transactionNumber,aux3,
						      (byte)(DictionaryContentNodeDriver
							     .getNameHeight(dbsm,transactionNumber,aux3)+1));
			return aux3;
		    }
	    }
	return node;
    }

    static private DbUnitAddress findNearestInNameTree(DbStorageManager dbsm, long transactionNumber,
						       DbUnitAddress header, DictionaryEntry entry)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	current.setNull();
	DbUnitAddress target=new DbUnitAddress();
	DictionaryHeaderDriver.getNamesRoot(dbsm,transactionNumber,header,target);
	String aux;
	while (!target.isNull())
	    {
		current.copyFrom(target);
		aux=getNameFromNode(dbsm,transactionNumber,current);
		int r=entry.getName().compareTo(aux);
		if (r<0)
		    DictionaryContentNodeDriver.getLeftNameNode(dbsm,transactionNumber,current,target);
		else if (r>0)
		    DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,current,target);
		else
		    return current;
	    }
	return current;
    }

    static private void adjustNameTreeAfterInsertion(DbStorageManager dbsm, long transactionNumber,
						     DbUnitAddress header, DbUnitAddress node)
	throws StorageException, IOException
    {
	while (!node.isNull())
	    {
		node=skewInNameTree(dbsm,transactionNumber,header,node);
		node=splitInNameTree(dbsm,transactionNumber,header,node);
		DictionaryContentNodeDriver.getParentNameNode(dbsm,transactionNumber,node,node);
	    }
    }

    static private void reinsertNodeInNameTree(DbStorageManager dbsm, long transactionNumber,
						DbUnitAddress header, DbUnitAddress isolated, DbUnitAddress deleted)
	throws StorageException, IOException
    {
	if (!isolated.equals(deleted))
	    {
		DictionaryContentNodeDriver.setNameHeight(dbsm,transactionNumber,isolated,
					      DictionaryContentNodeDriver.getNameHeight(dbsm,
											transactionNumber,
											deleted));
		DbUnitAddress aux=new DbUnitAddress();
		DictionaryContentNodeDriver.getLeftNameNode(dbsm,transactionNumber,deleted,aux);
		DictionaryContentNodeDriver.setLeftNameNode(dbsm,transactionNumber,isolated,aux);
		DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,deleted,aux);
		DictionaryContentNodeDriver.setRightNameNode(dbsm,transactionNumber,isolated,aux);
		DictionaryContentNodeDriver.getParentNameNode(dbsm,transactionNumber,deleted,aux);
		DictionaryContentNodeDriver.setParentNameNode(dbsm,transactionNumber,isolated,aux);
		
		if (aux.isNull())
		    DictionaryHeaderDriver.setNamesRoot(dbsm,transactionNumber,header,isolated);
		else
		    {
			DbUnitAddress aux2=new DbUnitAddress();
			DictionaryContentNodeDriver.getLeftNameNode(dbsm,transactionNumber,aux,aux2);
			if (aux2.equals(deleted))
			    DictionaryContentNodeDriver.setLeftNameNode(dbsm,transactionNumber,aux,isolated);
			else
			    DictionaryContentNodeDriver.setRightNameNode(dbsm,transactionNumber,aux,isolated);
		    }

		DictionaryContentNodeDriver.getLeftNameNode(dbsm,transactionNumber,deleted,aux);
		if (!aux.isNull())
		    DictionaryContentNodeDriver.setParentNameNode(dbsm,transactionNumber,aux,isolated);
		DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,deleted,aux);
		if (!aux.isNull())
		    DictionaryContentNodeDriver.setParentNameNode(dbsm,transactionNumber,aux,isolated);	
	    }
	
	DictionaryContentNodeDriver.setNameHeight(dbsm,transactionNumber,deleted,(byte)0);
	DictionaryContentNodeDriver.setLeftNameNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setRightNameNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
	DictionaryContentNodeDriver.setParentNameNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
    }


    static private void removeFromNameTree(DbStorageManager dbsm, long transactionNumber,
					   DbUnitAddress header, DbUnitAddress deleted)
	throws StorageException, IOException
    {
	DbUnitAddress victim=new DbUnitAddress();
	DbUnitAddress adjusted=new DbUnitAddress();
	DbUnitAddress aux=new DbUnitAddress();
	DbUnitAddress aux2=new DbUnitAddress();

	DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,deleted,aux);
	if (!aux.isNull())
	    DictionaryContentNodeDriver.getNextNameNode(dbsm,transactionNumber,deleted,victim);
	else
	    victim.copyFrom(deleted);
	     
	DictionaryContentNodeDriver.getParentNameNode(dbsm,transactionNumber,victim,aux);
	if (aux.equals(deleted))
	    adjusted.copyFrom(victim);
	else
	    DictionaryContentNodeDriver.getParentNameNode(dbsm,transactionNumber,victim,adjusted);
	if (aux.isNull())
	    DictionaryHeaderDriver.setNamesRoot(dbsm,transactionNumber,header,DbUnitAddress.NULL);
	else
	    {
		DictionaryContentNodeDriver.getLeftNameNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(victim))
		    {
			DbUnitAddress aux3=new DbUnitAddress();
			DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,victim,aux3);
			DictionaryContentNodeDriver.setLeftNameNode(dbsm,transactionNumber,aux,aux3);
			DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,victim,aux2);
			if (!aux2.isNull())
			    DictionaryContentNodeDriver.setParentNameNode(dbsm,transactionNumber,aux2,aux);
		    }
		else
		    {
			DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,victim,aux2);
			if (!aux2.isNull())
			    DictionaryContentNodeDriver.setParentNameNode(dbsm,transactionNumber,aux2,aux);
			DictionaryContentNodeDriver.setRightNameNode(dbsm,transactionNumber,aux,aux2);
		    }
	    }
	reinsertNodeInNameTree(dbsm,transactionNumber,header,victim,deleted);

	while (!adjusted.isNull())
	    {
		DictionaryContentNodeDriver.getLeftNameNode(dbsm,transactionNumber,adjusted,aux);
		DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,adjusted,aux2);
		byte adjustedHeight=DictionaryContentNodeDriver.getNameHeight(dbsm,transactionNumber,adjusted);
		if (((!aux.isNull() && 
		      DictionaryContentNodeDriver.getNameHeight(dbsm,transactionNumber,aux)<adjustedHeight-1) ||
		     (aux.isNull() && 0<adjustedHeight-1)) ||
		    ((!aux2.isNull() && 
		      DictionaryContentNodeDriver.getNameHeight(dbsm,transactionNumber,aux2)<adjustedHeight-1) ||
		     (aux2.isNull() && 0<adjustedHeight-1)))
		    {
			DictionaryContentNodeDriver.setNameHeight(dbsm,transactionNumber,adjusted,
						      (byte)(DictionaryContentNodeDriver.getNameHeight(dbsm,
											   transactionNumber,
											   adjusted)-1));
			adjustedHeight--;
			if (!aux2.isNull() && 
			    DictionaryContentNodeDriver.getNameHeight(dbsm,transactionNumber,aux2)>adjustedHeight)
			    DictionaryContentNodeDriver.setNameHeight(dbsm,transactionNumber,aux2,adjustedHeight);
			adjusted=skewInNameTree(dbsm,transactionNumber,header,adjusted);
			DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,adjusted,aux2);
			if (!aux2.isNull())
			    {
				DictionaryContentNodeDriver.setRightNameNode(dbsm,transactionNumber,adjusted,
								 skewInNameTree(dbsm,transactionNumber,header,aux2));
				DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,adjusted,aux2);
				DbUnitAddress aux3=new DbUnitAddress();
				DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,aux2,aux3);
				if (!aux3.isNull())
				    DictionaryContentNodeDriver.setRightNameNode(dbsm,transactionNumber,aux2,
								     skewInNameTree(dbsm,transactionNumber,header,aux3));
			    }
			adjusted=splitInNameTree(dbsm,transactionNumber,header,adjusted);
			DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,adjusted,aux2);
			if (!aux2.isNull())
			    DictionaryContentNodeDriver.setRightNameNode(dbsm,transactionNumber,adjusted,
							     splitInNameTree(dbsm,transactionNumber,header,aux2));
		    }
		DictionaryContentNodeDriver.getParentNameNode(dbsm,transactionNumber,adjusted,adjusted);
	    }	
    }

    // Pre: entry is valid.
    static private DbUnitAddress insertByName(DbStorageManager dbsm, long transactionNumber,
					      DbUnitAddress header, DictionaryEntry entry)
	throws StorageException, IOException, UnavailableUnitException
    {
		DbUnitAddress aux=new DbUnitAddress();
	DictionaryHeaderDriver.getNamesRoot(dbsm,transactionNumber,header,aux);
	if (aux.isNull())
	    {
		createNode(dbsm,transactionNumber,entry,aux);
		DictionaryHeaderDriver.setNamesHead(dbsm,transactionNumber,header,aux);
		DictionaryHeaderDriver.setNamesRoot(dbsm,transactionNumber,header,aux);
		return aux;
	    }
	else
	    {
		aux=findNearestInNameTree(dbsm,transactionNumber,header,entry);
		String nameAux=getNameFromNode(dbsm,transactionNumber,aux);
		int r=entry.getName().compareTo(nameAux);
		DbUnitAddress aux2=new DbUnitAddress();
		if (r!=0)
		    {
			createNode(dbsm,transactionNumber,entry,aux2);
			if (r<0)
			    {
				DictionaryContentNodeDriver.setLeftNameNode(dbsm,transactionNumber,aux,aux2);
				insertBeforeInNameList(dbsm,transactionNumber,header,aux2,aux);
			    }
			else
			    {
				DictionaryContentNodeDriver.setRightNameNode(dbsm,transactionNumber,aux,aux2);
				insertAfterInNameList(dbsm,transactionNumber,header,aux2,aux);
			    }
			DictionaryContentNodeDriver.setParentNameNode(dbsm,transactionNumber,aux2,aux);
			adjustNameTreeAfterInsertion(dbsm,transactionNumber,header,aux);
			return aux2;
		    }
		else
		    {
			aux2.setNull();
			return aux2;
		    }
	    }
    }

    /**
     * Removes a specified node from the dictionary collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     * @param node The node to remove.
     */
    static public void remove(DbStorageManager dbsm, long transactionNumber,
			      DbUnitAddress header, DbUnitAddress node)
	throws StorageException, IOException
    {
	removeFromIdTree(dbsm,transactionNumber,header,node);
	removeFromIdList(dbsm,transactionNumber,header,node);
	removeFromNameTree(dbsm,transactionNumber,header,node);
	removeFromNameList(dbsm,transactionNumber,header,node);
	dbsm.dropUnit(transactionNumber,DbStorageManager.DICTIONARY_SEGMENT,node);
    }


    // Public Methods

    /**
     * Checks if a specified dictionary entry belongs to the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     * @param entry The dictionary entry.
     * @return <code>true</code> if the dictionary entry is in the collection,
     * <code>false</code> otherwise.
     */
    static public boolean contains(DbStorageManager dbsm, long transactionNumber,
				   DbUnitAddress header, DictionaryEntry entry)
	throws StorageException, IOException
    {
	DbUnitAddress aux=new DbUnitAddress();
	aux=findNearestInNameTree(dbsm,transactionNumber,header,entry);
	DictionaryEntry entryAux=getDictionaryEntryFromNode(dbsm,transactionNumber,aux);
	int r=DictionaryEntry.ByNameComparator.compare(entry,entryAux);
	if (r==0)
	    return true;
	else
	    return false;
    }

    /**
     * Checks if a specified dictionary entry belongs to the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     * @param entryName The name of the dictionary entry.
     * @param entryID The identifier of the dictionary entry.
     * @return <code>true</code> if the dictionary entry is in the collection,
     * <code>false</code> otherwise.
     */
    static public boolean contains(DbStorageManager dbsm, long transactionNumber,
				   DbUnitAddress header, String entryName, PhysicalID entryID)
	throws StorageException, IOException
    {
	return contains(dbsm,transactionNumber,header,new DictionaryEntry(entryName,entryID));
    }

    /**
     * Inserts a specified dictionary entry to the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     * @param entry The dictionary entry.
     */
    static public void insert(DbStorageManager dbsm, long transactionNumber,
			      DbUnitAddress header, DictionaryEntry entry)
	throws StorageException, IOException, UnavailableUnitException,
	       AlreadyExistingException
    {
	DbUnitAddress aux=insertByName(dbsm,transactionNumber,header,entry);
	if (!aux.isNull())
	    insertById(dbsm,transactionNumber,header,aux);
	else
	    throw new AlreadyExistingException();
    }

    /**
     * Inserts a specified dictionary entry to the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     * @param entryName The name of the dictionary entry.
     * @param entryID The identifier of the dictionary entry.
     */
    static public void insert(DbStorageManager dbsm, long transactionNumber,
			      DbUnitAddress header, String entryName, PhysicalID entryID)
	throws StorageException, IOException, UnavailableUnitException,
	       AlreadyExistingException
    {
	insert(dbsm,transactionNumber,header,new DictionaryEntry(entryName,entryID));
    }

    /**
     * Finds the first node in the collection with the specified identifier.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     * @param matchID The identifier to find.
     * @param destiny Where the node's address is stored, it will be a null address if the
     * identifier is not found.
     */
    static public void getFirstWithID(DbStorageManager dbsm, long transactionNumber,
				      DbUnitAddress header, PhysicalID matchID, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	current.setNull();
	DbUnitAddress target=new DbUnitAddress();
	DictionaryHeaderDriver.getIdsRoot(dbsm,transactionNumber,header,target);
	DictionaryEntry aux=null;
	while (!target.isNull())
	    {
		current.copyFrom(target);
		aux=getDictionaryEntryFromNode(dbsm,transactionNumber,current);
		int r=matchID.compareTo(aux.getID());
		if (r<=0)
		    DictionaryContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,current,target);
		else
		    DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,current,target);
	    }
	if (current.isNull())
	    destiny.setNull();
	else
	    {
		if (matchID.compareTo(aux.getID())==0)
		    destiny.copyFrom(current);
		else
		    {
			DictionaryContentNodeDriver.getNextIdNode(dbsm,transactionNumber,current,current);
			if (!current.isNull())
			    {
				aux=getDictionaryEntryFromNode(dbsm,transactionNumber,current);
				if (matchID.compareTo(aux.getID())==0)
				    destiny.copyFrom(current);
				else
				    destiny.setNull();
			    }
			else
			     destiny.setNull();
		    }
	    }
    }

    /**
     * Checks if a specified identifier is in some dictionary entry in the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     * @param matchID The identifier to find.
     * @return <code>true</code> if the identifier is in the collection,
     * <code>false</code> otherwise.
     */
    static public boolean containsID(DbStorageManager dbsm, long transactionNumber,
				     DbUnitAddress header, PhysicalID matchID)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	current.setNull();
	DbUnitAddress target=new DbUnitAddress();
	DictionaryHeaderDriver.getIdsRoot(dbsm,transactionNumber,header,target);
	DictionaryEntry aux=null;
	while (!target.isNull())
	    {
		current.copyFrom(target);
		aux=getDictionaryEntryFromNode(dbsm,transactionNumber,current);
		int r=matchID.compareTo(aux.getID());
		if (r==0)
		    return true;
		else if (r<0)
		    DictionaryContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,current,target);
		else
		    DictionaryContentNodeDriver.getRightIdNode(dbsm,transactionNumber,current,target);
	    }
	return false;
    }

    /**
     * Finds the first node in the collection with the specified name.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     * @param matchName The name to find.
     * @param destiny Where the node's address is stored, it will be a null address if the
     * identifier is not found.
     */
    static private void getNodeWithName(DbStorageManager dbsm, long transactionNumber,
					DbUnitAddress header, String matchName, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	destiny.setNull();
	DbUnitAddress current=new DbUnitAddress();
	current.setNull();
	DbUnitAddress target=new DbUnitAddress();
	DictionaryHeaderDriver.getNamesRoot(dbsm,transactionNumber,header,target);
	DictionaryEntry aux=null;
	while (!target.isNull())
	    {
		current.copyFrom(target);
		aux=getDictionaryEntryFromNode(dbsm,transactionNumber,current);
		int r=matchName.compareTo(aux.getName());
		if (r<0)
		    DictionaryContentNodeDriver.getLeftNameNode(dbsm,transactionNumber,current,target);
		else if (r>0)
		    DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,current,target);
		else
		    {
			destiny.copyFrom(current);
			return;
		    }
	    }
    }

    /**
     * Finds the first identifier in the collection with the specified name.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     * @param matchName The name to find.
     * @param id Where the identificator is stored.
     */
    static public void getID(DbStorageManager dbsm, long transactionNumber,
			      DbUnitAddress header, String matchName, PhysicalID id)
	throws StorageException, IOException, NotExistingException
    {
	id.setNull();
	DbUnitAddress current=new DbUnitAddress();
	getNodeWithName(dbsm,transactionNumber,header,matchName,current);
	if (!current.isNull())
	    DictionaryContentNodeDriver.getId(dbsm,transactionNumber,current,id);
	else
	    throw new NotExistingException();
    }

    /**
     * Checks if a specified nameed is in some dictionary entry in the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     * @param matchName The labe to find.
     * @return <code>true</code> if the name is in the collection,
     * <code>false</code> otherwise.
     */
    static public boolean containsName(DbStorageManager dbsm, long transactionNumber,
				       DbUnitAddress header, String matchName)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	current.setNull();
	DbUnitAddress target=new DbUnitAddress();
	DictionaryHeaderDriver.getNamesRoot(dbsm,transactionNumber,header,target);
	DictionaryEntry aux=null;
	while (!target.isNull())
	    {
		current.copyFrom(target);
		aux=getDictionaryEntryFromNode(dbsm,transactionNumber,current);
		int r=matchName.compareTo(aux.getName());
		if (r==0)
		    return true;
		else if (r<0)
		    DictionaryContentNodeDriver.getLeftNameNode(dbsm,transactionNumber,current,target);
		else
		    DictionaryContentNodeDriver.getRightNameNode(dbsm,transactionNumber,current,target);
	    }
	return false;
    }

    /**
     * Deletes a specified dictionary entry from the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     * @param entry The dictionary entry.
     */
    static public void delete(DbStorageManager dbsm, long transactionNumber,
			      DbUnitAddress header, DictionaryEntry entry)
	throws StorageException, IOException, NotExistingException
    {
	DbUnitAddress aux=new DbUnitAddress();
	aux=findNearestInNameTree(dbsm,transactionNumber,header,entry);
	DictionaryEntry entryAux=getDictionaryEntryFromNode(dbsm,transactionNumber,aux);
	int r=DictionaryEntry.ByNameComparator.compare(entry,entryAux);
	if (r==0)
	    remove(dbsm,transactionNumber,header,aux);
	else
	    throw new NotExistingException();
    }

    /**
     * Deletes a specified dictionary entry from the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     * @param entryName The name of the dictionary entry.
     * @param entryID The identifier of the dictionary entry.
     */
    static public void delete(DbStorageManager dbsm, long transactionNumber,
			      DbUnitAddress header, String entryName, PhysicalID entryID)
	throws StorageException, IOException, NotExistingException
    {
	delete(dbsm,transactionNumber,header,new DictionaryEntry(entryName,entryID));
    }

    /**
     * Deletes a specified dictionary entry from the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     * @param deletedID The identifier of the dictionary entries to delete.
     */
    static public void deleteID(DbStorageManager dbsm, long transactionNumber,
				DbUnitAddress header, PhysicalID deletedID)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	DbUnitAddress next=new DbUnitAddress();
	PhysicalID aux=new PhysicalID();
	getFirstWithID(dbsm,transactionNumber,header,deletedID,current);
	boolean end=false;
	while (!end && !current.isNull())
	    {
		DictionaryContentNodeDriver.getId(dbsm,transactionNumber,current,aux);
		if (aux.compareTo(deletedID)!=0)
		    end=true;
		else
		    {
			DictionaryContentNodeDriver.getNextIdNode(dbsm,transactionNumber,current,next);
			remove(dbsm,transactionNumber,header,current);
			current.copyFrom(next);
		    }
	    }
    }

    /**
     * Deletes a specified dictionary entry from the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param header The address of the dictionary header.
     * @param deletedName The name of the dictionary entry to delete.
     */
    static public void deleteName(DbStorageManager dbsm, long transactionNumber,
				  DbUnitAddress header, String deletedName)
	throws StorageException, IOException, NotExistingException
    {
	DbUnitAddress current=new DbUnitAddress();
	getNodeWithName(dbsm,transactionNumber,header,deletedName,current);
	if (!current.isNull())
	    remove(dbsm,transactionNumber,header,current);
	else
	    throw new NotExistingException();
    }
}
