package DbStorage.manager;

import DbStorage.*;
import DbStorage.driver.*;
import storage.*;
import ssdbms.ssd.*;
import java.io.*;

/**
 * This class is used to create new semistructured data.
 */
public final class IdManager
{
    static private void createIsolated(DbStorageManager dbsm, long transactionNumber, 
				       int type, PhysicalID newID)
	throws StorageException, IOException, UnavailableUnitException
    {
	newID.setNull();
	SsdDescriptorDriver.createSsdDescriptor(dbsm,transactionNumber,newID);
	if (newID.isNull())
	    throw new UnavailableUnitException();
	SsdDescriptorDriver.setDictionaryReferences(dbsm,transactionNumber,newID,0);
	SsdDescriptorDriver.setFirstParent(dbsm,transactionNumber,newID,PhysicalID.NULL);
	SsdDescriptorDriver.setRootParentsCollection(dbsm,transactionNumber,newID,DbUnitAddress.NULL);
	SsdDescriptorDriver.setHeadParentsCollection(dbsm,transactionNumber,newID,DbUnitAddress.NULL);
	SsdDescriptorDriver.setType(dbsm,transactionNumber,newID,type);
    }
    
    static private void createSimplePrimitive(DbStorageManager dbsm, long transactionNumber, 
					     int type, byte[] value, PhysicalID newID)
	throws StorageException, IOException, UnavailableUnitException
    {
	createIsolated(dbsm,transactionNumber,type,newID);
	SimplePrimitiveSsdDescriptorDriver.setContentSize(dbsm,transactionNumber,newID,value.length);
	
	int max=value.length;
	int ini=0;
	int len=SimplePrimitiveSsdDescriptorDriver.setContent(dbsm,transactionNumber,newID,max,value,ini);
	
	ini=ini+len;
	max=max-len;
	if (max<=0)
	    SimplePrimitiveSsdDescriptorDriver.setContentExtension(dbsm,transactionNumber,
								   newID,DbUnitAddress.NULL);
	else //if (max>0)
	    {
		DbUnitAddress u=new DbUnitAddress();
		u.copyFrom(newID);
		DbUnitAddress v=new DbUnitAddress();
		boolean first=true;
		while (max>0)
		    {
			ContentExtensionDriver.createContentExtension(dbsm,transactionNumber,v);
			if (v.isNull())
			    throw new UnavailableUnitException();
			len=ContentExtensionDriver.setContent(dbsm,transactionNumber,v,max,value,ini);
			if (first)
			    SimplePrimitiveSsdDescriptorDriver.setContentExtension(dbsm,transactionNumber,u,v);
			else
			    ContentExtensionDriver.setContentExtension(dbsm,transactionNumber,u,v);
			ini=ini+len;
			max=max-len;
			u.copyFrom(v);
			first=false;
		    }
		if (first)
		    SimplePrimitiveSsdDescriptorDriver.setContentExtension(dbsm,transactionNumber,
									   u,DbUnitAddress.NULL);
		else
		    ContentExtensionDriver.setContentExtension(dbsm,transactionNumber,u,DbUnitAddress.NULL);
	    }
	
    }

    /**
     * Creates a non primitive datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param newID Where the identifier of the new datum is stored.
     */
    static public void createNonPrimitive(DbStorageManager dbsm, long transactionNumber, PhysicalID newID)
	throws StorageException, IOException, UnavailableUnitException
    {
	createIsolated(dbsm,transactionNumber,SSDValue.NON_PRIMITIVE,newID);
	NonPrimitiveSsdDescriptorDriver.setIdsRoot(dbsm,transactionNumber,newID,DbUnitAddress.NULL);
	NonPrimitiveSsdDescriptorDriver.setIdsHead(dbsm,transactionNumber,newID,DbUnitAddress.NULL);
	NonPrimitiveSsdDescriptorDriver.setLabelsRoot(dbsm,transactionNumber,newID,DbUnitAddress.NULL);
	NonPrimitiveSsdDescriptorDriver.setLabelsHead(dbsm,transactionNumber,newID,DbUnitAddress.NULL);
    }

    static private void createLOB(DbStorageManager dbsm, long transactionNumber, int type, PhysicalID newID)
	throws StorageException, IOException, UnavailableUnitException
    {
	createIsolated(dbsm,transactionNumber,type,newID);
	LobPrimitiveSsdDescriptorDriver.setContentExtension(dbsm,transactionNumber,newID,DbUnitAddress.NULL);
	LobPrimitiveSsdDescriptorDriver.setContentSize(dbsm,transactionNumber,newID,0);
    }

    /**
     * Creates a primitive BLOB datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param newID Where the identifier of the new datum is stored.
     */
    static public void createBLOB(DbStorageManager dbsm, long transactionNumber, PhysicalID newID)
	throws StorageException, IOException, UnavailableUnitException
    {
	createLOB(dbsm,transactionNumber,SSDValue.BLOB,newID);
    }

    /**
     * Creates a primitive CLOB datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param newID Where the identifier of the new datum is stored.
     */
    static public void createCLOB(DbStorageManager dbsm, long transactionNumber, PhysicalID newID)
	throws StorageException, IOException, UnavailableUnitException
    {
	createLOB(dbsm,transactionNumber,SSDValue.CLOB,newID);
    }

    /**
     * Returns the byte representation of the content of a simple primitive datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the simple primitive datum.
     * @return A byte array with the content of the datum.
     */
    static public byte[] getSimpleContent(DbStorageManager dbsm, long transactionNumber, PhysicalID id)
	throws StorageException, IOException
    {
	int siz=SimplePrimitiveSsdDescriptorDriver.getContentSize(dbsm,transactionNumber,id);
	byte[] dataAux=new byte[siz];
	int ini=0;
	int len=SimplePrimitiveSsdDescriptorDriver.getContent(dbsm,transactionNumber,id,siz,dataAux,ini);
	DbUnitAddress u=new DbUnitAddress();
	SimplePrimitiveSsdDescriptorDriver.getContentExtension(dbsm,transactionNumber,id,u);
	while (!u.isNull())
	    {
		siz-=len;
		ini+=len;
		len=ContentExtensionDriver.getContent(dbsm,transactionNumber,u,siz,dataAux,ini);
		ContentExtensionDriver.getContentExtension(dbsm,transactionNumber,u,u);
	    }
	return dataAux;
    }


    /**
     * Creates a integer primitive datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param value The value of the datum.
     * @param newID Where the identifier of the new datum is stored.
     */
    static public void createInteger(DbStorageManager dbsm, long transactionNumber, long value, PhysicalID newID)
	throws StorageException, IOException, UnavailableUnitException
    {
	createSimplePrimitive(dbsm,transactionNumber,SSDValue.INTEGER,DataConvertion.longToBytes(value),newID);
    }

    /**
     * Creates a real primitive datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param value The value of the datum.
     * @param newID Where the identifier of the new datum is stored.
     */
    static public void createReal(DbStorageManager dbsm, long transactionNumber, double value, PhysicalID newID)
	throws StorageException, IOException, UnavailableUnitException
    {
	createSimplePrimitive(dbsm,transactionNumber,SSDValue.REAL,DataConvertion.doubleToBytes(value),newID);
    }

    /**
     * Creates a strin primitive datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param value The value of the datum.
     * @param newID Where the identifier of the new datum is stored.
     */
    static public void createString(DbStorageManager dbsm, long transactionNumber, String value, PhysicalID newID)
	throws StorageException, IOException, UnavailableUnitException
    {
	createSimplePrimitive(dbsm,transactionNumber,SSDValue.STRING,DataConvertion.stringToBytes(value),newID);
    }

    /**
     * Creates a byte sequence primitive datum.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param value The value of the datum.
     * @param newID Where the identifier of the new datum is stored.
     */
    static public void createByteSequence(DbStorageManager dbsm, long transactionNumber, byte[] value, 
					  PhysicalID newID)
	throws StorageException, IOException, UnavailableUnitException
    {
	createSimplePrimitive(dbsm,transactionNumber,SSDValue.BYTE_SEQUENCE,value,newID);
    }


}
