package DbStorage.manager;

import DbStorage.*;
import DbStorage.driver.*;
import storage.*;
import java.io.*;

/**
 * This class manages the content of a collection of labeled semistructured data.
 * This collection is implemented using 4 data structures combined in one, 
 * 2 AA-trees and 2 double linked list are used, for ordering the collection
 * by identificator and label.
 */
public final class LabeledSsdCollectionManager
{
    static private void createNode(DbStorageManager dbsm, long transactionNumber,
				  String label, PhysicalID id,
				  DbUnitAddress newUnitAddress)
	throws StorageException, IOException, UnavailableUnitException
    {
	newUnitAddress.setNull();
	ContentNodeDriver.createContentNode(dbsm,transactionNumber,newUnitAddress);
	if (newUnitAddress.isNull())
	    throw new UnavailableUnitException();
	ContentNodeDriver.setPreviousIdNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ContentNodeDriver.setNextIdNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ContentNodeDriver.setRightIdNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ContentNodeDriver.setParentIdNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ContentNodeDriver.setIdHeight(dbsm,transactionNumber,newUnitAddress,(byte)1);
	ContentNodeDriver.setPreviousLabelNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ContentNodeDriver.setNextLabelNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ContentNodeDriver.setLeftLabelNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ContentNodeDriver.setRightLabelNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ContentNodeDriver.setParentLabelNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ContentNodeDriver.setLabelHeight(dbsm,transactionNumber,newUnitAddress,(byte)1);
	ContentNodeDriver.setId(dbsm,transactionNumber,newUnitAddress,id);
	
	byte[] labelContent=DataConvertion.stringToBytes(label);
	ContentNodeDriver.setLabelSize(dbsm,transactionNumber,newUnitAddress,labelContent.length);
	
	int max=labelContent.length;
	int ini=0;
	int len=ContentNodeDriver.setLabelContent(dbsm,transactionNumber,newUnitAddress,max,labelContent,ini);
	
	ini=ini+len;
	max=max-len;
	if (max<=0)
	    ContentNodeDriver.setLabelContentExtension(dbsm,transactionNumber,
						  newUnitAddress,DbUnitAddress.NULL);
	else //if (max>0)
	    {
		DbUnitAddress u=new DbUnitAddress();
		u.copyFrom(newUnitAddress);
		DbUnitAddress v=new DbUnitAddress();
		boolean first=true;
		while (max>0)
		    {
			ContentExtensionDriver.createContentExtension(dbsm,transactionNumber,v);
			if (v.isNull())
			    throw new UnavailableUnitException();
			len=ContentExtensionDriver.setContent(dbsm,transactionNumber,v,max,labelContent,ini);
			if (first)
			    ContentNodeDriver.setLabelContentExtension(dbsm,transactionNumber,u,v);
			else
			    ContentExtensionDriver.setContentExtension(dbsm,transactionNumber,u,v);
			ini=ini+len;
			max=max-len;
			u.copyFrom(v);
			first=false;
		    }
		if (first)
		    ContentNodeDriver.setLabelContentExtension(dbsm,transactionNumber,
									   u,DbUnitAddress.NULL);
		else
		    ContentExtensionDriver.setContentExtension(dbsm,transactionNumber,u,DbUnitAddress.NULL);
	    }
    }

    // Pre: lssd is valid.
    static private void createNode(DbStorageManager dbsm, long transactionNumber,
				  PhysicalLabeledSSD lssd,
				  DbUnitAddress newUnitAddress)
	throws StorageException, IOException, UnavailableUnitException
    {
	createNode(dbsm,transactionNumber,lssd.getLabel(),lssd.getID(),newUnitAddress);
    }

    /**
     * Returns the label of a specified node of the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param nodeAddress The addres of the node in the CONTENT segment.
     * @return The label in a <code>String</code> object.
     */
    static public String getLabelFromNode(DbStorageManager dbsm, long transactionNumber,
					  DbUnitAddress nodeAddress)
	throws StorageException, IOException
    {
	int siz=ContentNodeDriver.getLabelSize(dbsm,transactionNumber,nodeAddress);
	byte[] dataAux=new byte[siz];
	int ini=0;
	int len=ContentNodeDriver.getLabelContent(dbsm,transactionNumber,nodeAddress,siz,dataAux,ini);
	DbUnitAddress u=new DbUnitAddress();
	ContentNodeDriver.getLabelContentExtension(dbsm,transactionNumber,nodeAddress,u);
	while (!u.isNull())
	    {
		siz-=len;
		ini+=len;
		len=ContentExtensionDriver.getContent(dbsm,transactionNumber,u,siz,dataAux,ini);
		ContentExtensionDriver.getContentExtension(dbsm,transactionNumber,u,u);
	    }
	return DataConvertion.bytesToString(dataAux);
    }

    /**
     * Returns the labeled semistructured datum of a specified node of the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param nodeAddress The addres of the node in the CONTENT segment.
     * @return The labeled semistructured datum in a <code>PhysicalLabeledSSD<code> onject.
     */
    static public PhysicalLabeledSSD getPhysicalLabeledSsdFromNode(DbStorageManager dbsm, long transactionNumber,
								   DbUnitAddress nodeAddress)
	throws StorageException, IOException
    {
	PhysicalID aux=new PhysicalID();
	ContentNodeDriver.getId(dbsm,transactionNumber,nodeAddress,aux);
	return new PhysicalLabeledSSD(getLabelFromNode(dbsm,transactionNumber,nodeAddress),aux);
    }


    // Methods for Ids' Collections

    static private void insertAfterInIdList(DbStorageManager dbsm, long transactionNumber, PhysicalID id,
				    DbUnitAddress current, DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress targetNext=new DbUnitAddress();
	ContentNodeDriver.getNextIdNode(dbsm,transactionNumber,target,targetNext);
	ContentNodeDriver.setNextIdNode(dbsm,transactionNumber,current,targetNext);
	ContentNodeDriver.setPreviousIdNode(dbsm,transactionNumber,current,target);
	if (!targetNext.isNull())
	    ContentNodeDriver.setPreviousIdNode(dbsm,transactionNumber,targetNext,current);
	ContentNodeDriver.setNextIdNode(dbsm,transactionNumber,target,current);
    }

    static private void insertBeforeInIdList(DbStorageManager dbsm, long transactionNumber, PhysicalID id,
				    DbUnitAddress current, DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress idHead=new DbUnitAddress();
	NonPrimitiveSsdDescriptorDriver.getIdsHead(dbsm,transactionNumber,id,idHead);
	if (idHead.equals(target))
	    NonPrimitiveSsdDescriptorDriver.setIdsHead(dbsm,transactionNumber,id,current);
	
	DbUnitAddress targetPrevious=new DbUnitAddress();
	ContentNodeDriver.getPreviousIdNode(dbsm,transactionNumber,target,targetPrevious);
	ContentNodeDriver.setPreviousIdNode(dbsm,transactionNumber,current,targetPrevious);
	ContentNodeDriver.setNextIdNode(dbsm,transactionNumber,current,target);
	if (!targetPrevious.isNull())
	    ContentNodeDriver.setNextIdNode(dbsm,transactionNumber,targetPrevious,current);
	ContentNodeDriver.setPreviousIdNode(dbsm,transactionNumber,target,current);
    }

    static private void removeFromIdList(DbStorageManager dbsm, long transactionNumber, PhysicalID id,
				 DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress idHead=new DbUnitAddress();
	NonPrimitiveSsdDescriptorDriver.getIdsHead(dbsm,transactionNumber,id,idHead);
	DbUnitAddress targetNext=new DbUnitAddress();
	ContentNodeDriver.getNextIdNode(dbsm,transactionNumber,target,targetNext);
	DbUnitAddress targetPrevious=new DbUnitAddress();
	ContentNodeDriver.getPreviousIdNode(dbsm,transactionNumber,target,targetPrevious);

	if (idHead.equals(target))
	    NonPrimitiveSsdDescriptorDriver.setIdsHead(dbsm,transactionNumber,id,targetNext);
	if (!targetPrevious.isNull())
	    ContentNodeDriver.setNextIdNode(dbsm,transactionNumber,targetPrevious,targetNext);
	if (!targetNext.isNull())
	    ContentNodeDriver.setPreviousIdNode(dbsm,transactionNumber,targetNext,targetPrevious);
	
	ContentNodeDriver.setPreviousIdNode(dbsm,transactionNumber,target,DbUnitAddress.NULL);
	ContentNodeDriver.setNextIdNode(dbsm,transactionNumber,target,DbUnitAddress.NULL);
    }

    // Pre: node!=null && node.left!=null
    static private DbUnitAddress rotateWithLeftChildInIdTree(DbStorageManager dbsm, long transactionNumber,
						     PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress m=new DbUnitAddress();
	ContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,node,m);
	
	DbUnitAddress aux=new DbUnitAddress();
	ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,m,aux);
	ContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    ContentNodeDriver.setParentIdNode(dbsm,transactionNumber,aux,node);
	ContentNodeDriver.setRightIdNode(dbsm,transactionNumber,m,node);
	
	ContentNodeDriver.getParentIdNode(dbsm,transactionNumber,node,aux);
	if (aux.isNull())
	    NonPrimitiveSsdDescriptorDriver.setIdsRoot(dbsm,transactionNumber,id,m);
	else
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		ContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(node))
		    ContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,aux,m);
		else
		    ContentNodeDriver.setRightIdNode(dbsm,transactionNumber,aux,m);
	    }
	ContentNodeDriver.setParentIdNode(dbsm,transactionNumber,m,aux);
	ContentNodeDriver.setParentIdNode(dbsm,transactionNumber,node,m);
	return m;
    }

    // Pre: node!=null && node.right!=null
     static private DbUnitAddress rotateWithRightChildInIdTree(DbStorageManager dbsm, long transactionNumber,
						     PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress m=new DbUnitAddress();
	ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,node,m);
	
	DbUnitAddress aux=new DbUnitAddress();
	ContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,m,aux);
	ContentNodeDriver.setRightIdNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    ContentNodeDriver.setParentIdNode(dbsm,transactionNumber,aux,node);
	ContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,m,node);
	
	ContentNodeDriver.getParentIdNode(dbsm,transactionNumber,node,aux);
	if (aux.isNull())
	    NonPrimitiveSsdDescriptorDriver.setIdsRoot(dbsm,transactionNumber,id,m);
	else
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		ContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(node))
		    ContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,aux,m);
		else
		    ContentNodeDriver.setRightIdNode(dbsm,transactionNumber,aux,m);
	    }
	ContentNodeDriver.setParentIdNode(dbsm,transactionNumber,m,aux);
	ContentNodeDriver.setParentIdNode(dbsm,transactionNumber,node,m);
	return m;
    }

    // Pre: node!=null
    static private DbUnitAddress skewInIdTree(DbStorageManager dbsm, long transactionNumber,
					     PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress aux=new DbUnitAddress();
	ContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,node,aux);
	
	if (!aux.isNull() && 
	    (ContentNodeDriver.getIdHeight(dbsm,transactionNumber,node)==
	     ContentNodeDriver.getIdHeight(dbsm,transactionNumber,aux)))
	    return rotateWithLeftChildInIdTree(dbsm,transactionNumber,id,node);
	else
	    return node;
    }

    // Pre: node!=null
    static private DbUnitAddress splitInIdTree(DbStorageManager dbsm, long transactionNumber,
					      PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress aux=new DbUnitAddress();
	ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,aux,aux2);
		if (!aux2.isNull() &&
		    (ContentNodeDriver.getIdHeight(dbsm,transactionNumber,node)==
		     ContentNodeDriver.getIdHeight(dbsm,transactionNumber,aux2)))
		    {
			DbUnitAddress aux3=rotateWithRightChildInIdTree(dbsm,transactionNumber,id,node);
			ContentNodeDriver.setIdHeight(dbsm,transactionNumber,aux3,
						      (byte)(ContentNodeDriver
							     .getIdHeight(dbsm,transactionNumber,aux3)+1));
			return aux3;
		    }
	    }
	return node;
    }

    static private DbUnitAddress findNearestInIdTree(DbStorageManager dbsm, long transactionNumber,
						  PhysicalID id, PhysicalLabeledSSD lssd)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	current.setNull();
	DbUnitAddress target=new DbUnitAddress();
	NonPrimitiveSsdDescriptorDriver.getIdsRoot(dbsm,transactionNumber,id,target);
	PhysicalLabeledSSD aux;
	while (!target.isNull())
	    {
		current.copyFrom(target);
		aux=getPhysicalLabeledSsdFromNode(dbsm,transactionNumber,current);
		int r=PhysicalLabeledSSD.ByIDComparator.compare(lssd,aux);
		if (r<0)
		    ContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,current,target);
		else if (r>0)
		    ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,current,target);
		else
		    return current;
	    }
	return current;
    }

    static private void adjustIdTreeAfterInsertion(DbStorageManager dbsm, long transactionNumber,
						  PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	while (!node.isNull())
	    {
		node=skewInIdTree(dbsm,transactionNumber,id,node);
		node=splitInIdTree(dbsm,transactionNumber,id,node);
		ContentNodeDriver.getParentIdNode(dbsm,transactionNumber,node,node);
	    }
    }

    static private void reinsertNodeInIdTree(DbStorageManager dbsm, long transactionNumber,
					    PhysicalID id, DbUnitAddress isolated, DbUnitAddress deleted)
	throws StorageException, IOException
    {
	if (!isolated.equals(deleted))
	    {
		ContentNodeDriver.setIdHeight(dbsm,transactionNumber,isolated,
					      ContentNodeDriver.getIdHeight(dbsm,transactionNumber,deleted));
		DbUnitAddress aux=new DbUnitAddress();
		ContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,deleted,aux);
		ContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,isolated,aux);
		ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,deleted,aux);
		ContentNodeDriver.setRightIdNode(dbsm,transactionNumber,isolated,aux);
		ContentNodeDriver.getParentIdNode(dbsm,transactionNumber,deleted,aux);
		ContentNodeDriver.setParentIdNode(dbsm,transactionNumber,isolated,aux);
		
		if (aux.isNull())
		    NonPrimitiveSsdDescriptorDriver.setIdsRoot(dbsm,transactionNumber,id,isolated);
		else
		    {
			DbUnitAddress aux2=new DbUnitAddress();
			ContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,aux,aux2);
			if (aux2.equals(deleted))
			    ContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,aux,isolated);
			else
			    ContentNodeDriver.setRightIdNode(dbsm,transactionNumber,aux,isolated);
		    }

		ContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,deleted,aux);
		if (!aux.isNull())
		    ContentNodeDriver.setParentIdNode(dbsm,transactionNumber,aux,isolated);
		ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,deleted,aux);
		if (!aux.isNull())
		    ContentNodeDriver.setParentIdNode(dbsm,transactionNumber,aux,isolated);	
	    }
	
	ContentNodeDriver.setIdHeight(dbsm,transactionNumber,deleted,(byte)0);
	ContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
	ContentNodeDriver.setRightIdNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
	ContentNodeDriver.setParentIdNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
    }


    static private void removeFromIdTree(DbStorageManager dbsm, long transactionNumber,
					PhysicalID id, DbUnitAddress deleted)
	throws StorageException, IOException
    {
	DbUnitAddress victim=new DbUnitAddress();
	DbUnitAddress adjusted=new DbUnitAddress();
	DbUnitAddress aux=new DbUnitAddress();
	DbUnitAddress aux2=new DbUnitAddress();

	ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,deleted,aux);
	if (!aux.isNull())
	    ContentNodeDriver.getNextIdNode(dbsm,transactionNumber,deleted,victim);
	else
	    victim.copyFrom(deleted);
	     
	ContentNodeDriver.getParentIdNode(dbsm,transactionNumber,victim,aux);
	if (aux.equals(deleted))
	    adjusted.copyFrom(victim);
	else
	    ContentNodeDriver.getParentIdNode(dbsm,transactionNumber,victim,adjusted);
	if (aux.isNull())
	    NonPrimitiveSsdDescriptorDriver.setIdsRoot(dbsm,transactionNumber,id,DbUnitAddress.NULL);
	else
	    {
		ContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(victim))
		    {
			DbUnitAddress aux3=new DbUnitAddress();
			ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,victim,aux3);
			ContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,aux,aux3);
			ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,victim,aux2);
			if (!aux2.isNull())
			    ContentNodeDriver.setParentIdNode(dbsm,transactionNumber,aux2,aux);
		    }
		else
		    {
			ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,victim,aux2);
			if (!aux2.isNull())
			    ContentNodeDriver.setParentIdNode(dbsm,transactionNumber,aux2,aux);
			ContentNodeDriver.setRightIdNode(dbsm,transactionNumber,aux,aux2);
		    }
	    }
	reinsertNodeInIdTree(dbsm,transactionNumber,id,victim,deleted);

	while (!adjusted.isNull())
	    {
		ContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,adjusted,aux);
		ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,adjusted,aux2);
		byte adjustedHeight=ContentNodeDriver.getIdHeight(dbsm,transactionNumber,adjusted);
		if (((!aux.isNull() && 
		      ContentNodeDriver.getIdHeight(dbsm,transactionNumber,aux)<adjustedHeight-1) ||
		     (aux.isNull() && 0<adjustedHeight-1)) ||
		    ((!aux2.isNull() && 
		      ContentNodeDriver.getIdHeight(dbsm,transactionNumber,aux2)<adjustedHeight-1) ||
		     (aux2.isNull() && 0<adjustedHeight-1)))
		    {
			ContentNodeDriver.setIdHeight(dbsm,transactionNumber,adjusted,
						      (byte)(ContentNodeDriver.getIdHeight(dbsm,
											   transactionNumber,
											   adjusted)-1));
			adjustedHeight--;
			if (!aux2.isNull() && 
			    ContentNodeDriver.getIdHeight(dbsm,transactionNumber,aux2)>adjustedHeight)
			    ContentNodeDriver.setIdHeight(dbsm,transactionNumber,aux2,adjustedHeight);
			adjusted=skewInIdTree(dbsm,transactionNumber,id,adjusted);
			ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,adjusted,aux2);
			if (!aux2.isNull())
			    {
				ContentNodeDriver.setRightIdNode(dbsm,transactionNumber,adjusted,
								 skewInIdTree(dbsm,transactionNumber,id,aux2));
				ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,adjusted,aux2);
				DbUnitAddress aux3=new DbUnitAddress();
				ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,aux2,aux3);
				if (!aux3.isNull())
				    ContentNodeDriver.setRightIdNode(dbsm,transactionNumber,aux2,
								     skewInIdTree(dbsm,transactionNumber,id,aux3));
			    }
			adjusted=splitInIdTree(dbsm,transactionNumber,id,adjusted);
			ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,adjusted,aux2);
			if (!aux2.isNull())
			    ContentNodeDriver.setRightIdNode(dbsm,transactionNumber,adjusted,
							     splitInIdTree(dbsm,transactionNumber,id,aux2));
		    }
		ContentNodeDriver.getParentIdNode(dbsm,transactionNumber,adjusted,adjusted);
	    }	
    }


    // Pre: lssd is valid.
    static private DbUnitAddress insertById(DbStorageManager dbsm, long transactionNumber,
					   PhysicalID id, PhysicalLabeledSSD lssd)
	throws StorageException, IOException, UnavailableUnitException
    {
	DbUnitAddress aux=new DbUnitAddress();
	NonPrimitiveSsdDescriptorDriver.getIdsRoot(dbsm,transactionNumber,id,aux);
	if (aux.isNull())
	    {
		createNode(dbsm,transactionNumber,lssd,aux);
		NonPrimitiveSsdDescriptorDriver.setIdsHead(dbsm,transactionNumber,id,aux);
		NonPrimitiveSsdDescriptorDriver.setIdsRoot(dbsm,transactionNumber,id,aux);
		return aux;
	    }
	else
	    {
		aux=findNearestInIdTree(dbsm,transactionNumber,id,lssd);
		PhysicalLabeledSSD lssdAux=getPhysicalLabeledSsdFromNode(dbsm,transactionNumber,aux);
		int r=PhysicalLabeledSSD.ByIDComparator.compare(lssd,lssdAux);
		DbUnitAddress aux2=new DbUnitAddress();
		if (r!=0)
		    {
			createNode(dbsm,transactionNumber,lssd,aux2);
			if (r<0)
			    {
				ContentNodeDriver.setLeftIdNode(dbsm,transactionNumber,aux,aux2);
				insertBeforeInIdList(dbsm,transactionNumber,id,aux2,aux);
			    }
			else
			    {
				ContentNodeDriver.setRightIdNode(dbsm,transactionNumber,aux,aux2);
				insertAfterInIdList(dbsm,transactionNumber,id,aux2,aux);
			    }
			ContentNodeDriver.setParentIdNode(dbsm,transactionNumber,aux2,aux);
			adjustIdTreeAfterInsertion(dbsm,transactionNumber,id,aux);
			return aux2;
		    }
		else
		    {
			aux2.setNull();
			return aux2;
		    }
	    }
    }


    // Methods for Labels' Collections

    static private void insertAfterInLabelList(DbStorageManager dbsm, long transactionNumber, PhysicalID id,
					       DbUnitAddress current, DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress targetNext=new DbUnitAddress();
	ContentNodeDriver.getNextLabelNode(dbsm,transactionNumber,target,targetNext);
	ContentNodeDriver.setNextLabelNode(dbsm,transactionNumber,current,targetNext);
	ContentNodeDriver.setPreviousLabelNode(dbsm,transactionNumber,current,target);
	if (!targetNext.isNull())
	    ContentNodeDriver.setPreviousLabelNode(dbsm,transactionNumber,targetNext,current);
	ContentNodeDriver.setNextLabelNode(dbsm,transactionNumber,target,current);
    }

    static private void insertBeforeInLabelList(DbStorageManager dbsm, long transactionNumber, PhysicalID id,
						DbUnitAddress current, DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress labelHead=new DbUnitAddress();
	NonPrimitiveSsdDescriptorDriver.getLabelsHead(dbsm,transactionNumber,id,labelHead);
	if (labelHead.equals(target))
	    NonPrimitiveSsdDescriptorDriver.setLabelsHead(dbsm,transactionNumber,id,current);
	
	DbUnitAddress targetPrevious=new DbUnitAddress();
	ContentNodeDriver.getPreviousLabelNode(dbsm,transactionNumber,target,targetPrevious);
	ContentNodeDriver.setPreviousLabelNode(dbsm,transactionNumber,current,targetPrevious);
	ContentNodeDriver.setNextLabelNode(dbsm,transactionNumber,current,target);
	if (!targetPrevious.isNull())
	    ContentNodeDriver.setNextLabelNode(dbsm,transactionNumber,targetPrevious,current);
	ContentNodeDriver.setPreviousLabelNode(dbsm,transactionNumber,target,current);
    }

    static private void removeFromLabelList(DbStorageManager dbsm, long transactionNumber, PhysicalID id,
					    DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress labelHead=new DbUnitAddress();
	NonPrimitiveSsdDescriptorDriver.getLabelsHead(dbsm,transactionNumber,id,labelHead);
	DbUnitAddress targetNext=new DbUnitAddress();
	ContentNodeDriver.getNextLabelNode(dbsm,transactionNumber,target,targetNext);
	DbUnitAddress targetPrevious=new DbUnitAddress();
	ContentNodeDriver.getPreviousLabelNode(dbsm,transactionNumber,target,targetPrevious);

	if (labelHead.equals(target))
	    NonPrimitiveSsdDescriptorDriver.setLabelsHead(dbsm,transactionNumber,id,targetNext);
	if (!targetPrevious.isNull())
	    ContentNodeDriver.setNextLabelNode(dbsm,transactionNumber,targetPrevious,targetNext);
	if (!targetNext.isNull())
	    ContentNodeDriver.setPreviousLabelNode(dbsm,transactionNumber,targetNext,targetPrevious);
	
	ContentNodeDriver.setPreviousLabelNode(dbsm,transactionNumber,target,DbUnitAddress.NULL);
	ContentNodeDriver.setNextLabelNode(dbsm,transactionNumber,target,DbUnitAddress.NULL);
    }

    // Pre: node!=null && node.left!=null
    static private DbUnitAddress rotateWithLeftChildInLabelTree(DbStorageManager dbsm, long transactionNumber,
								PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress m=new DbUnitAddress();
	ContentNodeDriver.getLeftLabelNode(dbsm,transactionNumber,node,m);
	
	DbUnitAddress aux=new DbUnitAddress();
	ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,m,aux);
	ContentNodeDriver.setLeftLabelNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    ContentNodeDriver.setParentLabelNode(dbsm,transactionNumber,aux,node);
	ContentNodeDriver.setRightLabelNode(dbsm,transactionNumber,m,node);
	
	ContentNodeDriver.getParentLabelNode(dbsm,transactionNumber,node,aux);
	if (aux.isNull())
	    NonPrimitiveSsdDescriptorDriver.setLabelsRoot(dbsm,transactionNumber,id,m);
	else
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		ContentNodeDriver.getLeftLabelNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(node))
		    ContentNodeDriver.setLeftLabelNode(dbsm,transactionNumber,aux,m);
		else
		    ContentNodeDriver.setRightLabelNode(dbsm,transactionNumber,aux,m);
	    }
	ContentNodeDriver.setParentLabelNode(dbsm,transactionNumber,m,aux);
	ContentNodeDriver.setParentLabelNode(dbsm,transactionNumber,node,m);
	return m;
    }

    // Pre: node!=null && node.right!=null
     static private DbUnitAddress rotateWithRightChildInLabelTree(DbStorageManager dbsm, long transactionNumber,
								  PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress m=new DbUnitAddress();
	ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,node,m);
	
	DbUnitAddress aux=new DbUnitAddress();
	ContentNodeDriver.getLeftLabelNode(dbsm,transactionNumber,m,aux);
	ContentNodeDriver.setRightLabelNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    ContentNodeDriver.setParentLabelNode(dbsm,transactionNumber,aux,node);
	ContentNodeDriver.setLeftLabelNode(dbsm,transactionNumber,m,node);
	
	ContentNodeDriver.getParentLabelNode(dbsm,transactionNumber,node,aux);
	if (aux.isNull())
	    NonPrimitiveSsdDescriptorDriver.setLabelsRoot(dbsm,transactionNumber,id,m);
	else
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		ContentNodeDriver.getLeftLabelNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(node))
		    ContentNodeDriver.setLeftLabelNode(dbsm,transactionNumber,aux,m);
		else
		    ContentNodeDriver.setRightLabelNode(dbsm,transactionNumber,aux,m);
	    }
	ContentNodeDriver.setParentLabelNode(dbsm,transactionNumber,m,aux);
	ContentNodeDriver.setParentLabelNode(dbsm,transactionNumber,node,m);
	return m;
    }

    // Pre: node!=null
    static private DbUnitAddress skewInLabelTree(DbStorageManager dbsm, long transactionNumber,
						 PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress aux=new DbUnitAddress();
	ContentNodeDriver.getLeftLabelNode(dbsm,transactionNumber,node,aux);
	
	if (!aux.isNull() && 
	    (ContentNodeDriver.getLabelHeight(dbsm,transactionNumber,node)==
	     ContentNodeDriver.getLabelHeight(dbsm,transactionNumber,aux)))
	    return rotateWithLeftChildInLabelTree(dbsm,transactionNumber,id,node);
	else
	    return node;
    }

    // Pre: node!=null
    static private DbUnitAddress splitInLabelTree(DbStorageManager dbsm, long transactionNumber,
						  PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress aux=new DbUnitAddress();
	ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,aux,aux2);
		if (!aux2.isNull() &&
		    (ContentNodeDriver.getLabelHeight(dbsm,transactionNumber,node)==
		     ContentNodeDriver.getLabelHeight(dbsm,transactionNumber,aux2)))
		    {
			DbUnitAddress aux3=rotateWithRightChildInLabelTree(dbsm,transactionNumber,id,node);
			ContentNodeDriver.setLabelHeight(dbsm,transactionNumber,aux3,
						      (byte)(ContentNodeDriver
							     .getLabelHeight(dbsm,transactionNumber,aux3)+1));
			return aux3;
		    }
	    }
	return node;
    }

    static private DbUnitAddress findNearestInLabelTree(DbStorageManager dbsm, long transactionNumber,
							PhysicalID id, PhysicalLabeledSSD lssd)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	current.setNull();
	DbUnitAddress target=new DbUnitAddress();
	NonPrimitiveSsdDescriptorDriver.getLabelsRoot(dbsm,transactionNumber,id,target);
	PhysicalLabeledSSD aux;
	while (!target.isNull())
	    {
		current.copyFrom(target);
		aux=getPhysicalLabeledSsdFromNode(dbsm,transactionNumber,current);
		int r=PhysicalLabeledSSD.ByLabelComparator.compare(lssd,aux);
		if (r<0)
		    ContentNodeDriver.getLeftLabelNode(dbsm,transactionNumber,current,target);
		else if (r>0)
		    ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,current,target);
		else
		    return current;
	    }
	return current;
    }

    static private void adjustLabelTreeAfterInsertion(DbStorageManager dbsm, long transactionNumber,
						      PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	while (!node.isNull())
	    {
		node=skewInLabelTree(dbsm,transactionNumber,id,node);
		node=splitInLabelTree(dbsm,transactionNumber,id,node);
		ContentNodeDriver.getParentLabelNode(dbsm,transactionNumber,node,node);
	    }
    }

    static private void reinsertNodeInLabelTree(DbStorageManager dbsm, long transactionNumber,
						PhysicalID id, DbUnitAddress isolated, DbUnitAddress deleted)
	throws StorageException, IOException
    {
	if (!isolated.equals(deleted))
	    {
		ContentNodeDriver.setLabelHeight(dbsm,transactionNumber,isolated,
					      ContentNodeDriver.getLabelHeight(dbsm,transactionNumber,deleted));
		DbUnitAddress aux=new DbUnitAddress();
		ContentNodeDriver.getLeftLabelNode(dbsm,transactionNumber,deleted,aux);
		ContentNodeDriver.setLeftLabelNode(dbsm,transactionNumber,isolated,aux);
		ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,deleted,aux);
		ContentNodeDriver.setRightLabelNode(dbsm,transactionNumber,isolated,aux);
		ContentNodeDriver.getParentLabelNode(dbsm,transactionNumber,deleted,aux);
		ContentNodeDriver.setParentLabelNode(dbsm,transactionNumber,isolated,aux);
		
		if (aux.isNull())
		    NonPrimitiveSsdDescriptorDriver.setLabelsRoot(dbsm,transactionNumber,id,isolated);
		else
		    {
			DbUnitAddress aux2=new DbUnitAddress();
			ContentNodeDriver.getLeftLabelNode(dbsm,transactionNumber,aux,aux2);
			if (aux2.equals(deleted))
			    ContentNodeDriver.setLeftLabelNode(dbsm,transactionNumber,aux,isolated);
			else
			    ContentNodeDriver.setRightLabelNode(dbsm,transactionNumber,aux,isolated);
		    }

		ContentNodeDriver.getLeftLabelNode(dbsm,transactionNumber,deleted,aux);
		if (!aux.isNull())
		    ContentNodeDriver.setParentLabelNode(dbsm,transactionNumber,aux,isolated);
		ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,deleted,aux);
		if (!aux.isNull())
		    ContentNodeDriver.setParentLabelNode(dbsm,transactionNumber,aux,isolated);	
	    }
	
	ContentNodeDriver.setLabelHeight(dbsm,transactionNumber,deleted,(byte)0);
	ContentNodeDriver.setLeftLabelNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
	ContentNodeDriver.setRightLabelNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
	ContentNodeDriver.setParentLabelNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
    }


    static private void removeFromLabelTree(DbStorageManager dbsm, long transactionNumber,
					    PhysicalID id, DbUnitAddress deleted)
	throws StorageException, IOException
    {
	DbUnitAddress victim=new DbUnitAddress();
	DbUnitAddress adjusted=new DbUnitAddress();
	DbUnitAddress aux=new DbUnitAddress();
	DbUnitAddress aux2=new DbUnitAddress();

	ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,deleted,aux);
	if (!aux.isNull())
	    ContentNodeDriver.getNextLabelNode(dbsm,transactionNumber,deleted,victim);
	else
	    victim.copyFrom(deleted);
	     
	ContentNodeDriver.getParentLabelNode(dbsm,transactionNumber,victim,aux);
	if (aux.equals(deleted))
	    adjusted.copyFrom(victim);
	else
	    ContentNodeDriver.getParentLabelNode(dbsm,transactionNumber,victim,adjusted);
	if (aux.isNull())
	    NonPrimitiveSsdDescriptorDriver.setLabelsRoot(dbsm,transactionNumber,id,DbUnitAddress.NULL);
	else
	    {
		ContentNodeDriver.getLeftLabelNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(victim))
		    {
			DbUnitAddress aux3=new DbUnitAddress();
			ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,victim,aux3);
			ContentNodeDriver.setLeftLabelNode(dbsm,transactionNumber,aux,aux3);
			ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,victim,aux2);
			if (!aux2.isNull())
			    ContentNodeDriver.setParentLabelNode(dbsm,transactionNumber,aux2,aux);
		    }
		else
		    {
			ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,victim,aux2);
			if (!aux2.isNull())
			    ContentNodeDriver.setParentLabelNode(dbsm,transactionNumber,aux2,aux);
			ContentNodeDriver.setRightLabelNode(dbsm,transactionNumber,aux,aux2);
		    }
	    }
	reinsertNodeInLabelTree(dbsm,transactionNumber,id,victim,deleted);

	while (!adjusted.isNull())
	    {
		ContentNodeDriver.getLeftLabelNode(dbsm,transactionNumber,adjusted,aux);
		ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,adjusted,aux2);
		byte adjustedHeight=ContentNodeDriver.getLabelHeight(dbsm,transactionNumber,adjusted);
		if (((!aux.isNull() && 
		      ContentNodeDriver.getLabelHeight(dbsm,transactionNumber,aux)<adjustedHeight-1) ||
		     (aux.isNull() && 0<adjustedHeight-1)) ||
		    ((!aux2.isNull() && 
		      ContentNodeDriver.getLabelHeight(dbsm,transactionNumber,aux2)<adjustedHeight-1) ||
		     (aux2.isNull() && 0<adjustedHeight-1)))
		    {
			ContentNodeDriver.setLabelHeight(dbsm,transactionNumber,adjusted,
						      (byte)(ContentNodeDriver.getLabelHeight(dbsm,
											   transactionNumber,
											   adjusted)-1));
			adjustedHeight--;
			if (!aux2.isNull() && 
			    ContentNodeDriver.getLabelHeight(dbsm,transactionNumber,aux2)>adjustedHeight)
			    ContentNodeDriver.setLabelHeight(dbsm,transactionNumber,aux2,adjustedHeight);
			adjusted=skewInLabelTree(dbsm,transactionNumber,id,adjusted);
			ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,adjusted,aux2);
			if (!aux2.isNull())
			    {
				ContentNodeDriver.setRightLabelNode(dbsm,transactionNumber,adjusted,
								 skewInLabelTree(dbsm,transactionNumber,id,aux2));
				ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,adjusted,aux2);
				DbUnitAddress aux3=new DbUnitAddress();
				ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,aux2,aux3);
				if (!aux3.isNull())
				    ContentNodeDriver.setRightLabelNode(dbsm,transactionNumber,aux2,
								     skewInLabelTree(dbsm,transactionNumber,id,aux3));
			    }
			adjusted=splitInLabelTree(dbsm,transactionNumber,id,adjusted);
			ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,adjusted,aux2);
			if (!aux2.isNull())
			    ContentNodeDriver.setRightLabelNode(dbsm,transactionNumber,adjusted,
							     splitInLabelTree(dbsm,transactionNumber,id,aux2));
		    }
		ContentNodeDriver.getParentLabelNode(dbsm,transactionNumber,adjusted,adjusted);
	    }	
    }

    // Pre: node is isolated
    static private void insertByLabel(DbStorageManager dbsm, long transactionNumber,
				      PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress aux=new DbUnitAddress();
	NonPrimitiveSsdDescriptorDriver.getLabelsRoot(dbsm,transactionNumber,id,aux);
	if (aux.isNull())
	    {
		NonPrimitiveSsdDescriptorDriver.setLabelsHead(dbsm,transactionNumber,id,node);
		NonPrimitiveSsdDescriptorDriver.setLabelsRoot(dbsm,transactionNumber,id,node);
	    }
	else
	    {
		PhysicalLabeledSSD lssd=getPhysicalLabeledSsdFromNode(dbsm,transactionNumber,node);
		aux=findNearestInLabelTree(dbsm,transactionNumber,id,lssd);
		PhysicalLabeledSSD lssdAux=getPhysicalLabeledSsdFromNode(dbsm,transactionNumber,aux);
		int r=PhysicalLabeledSSD.ByLabelComparator.compare(lssd,lssdAux);
		if (r!=0)
		    {
			if (r<0)
			    {
				ContentNodeDriver.setLeftLabelNode(dbsm,transactionNumber,aux,node);
				insertBeforeInLabelList(dbsm,transactionNumber,id,node,aux);
			    }
			else
			    {
				ContentNodeDriver.setRightLabelNode(dbsm,transactionNumber,aux,node);
				insertAfterInLabelList(dbsm,transactionNumber,id,node,aux);
			    }
			ContentNodeDriver.setParentLabelNode(dbsm,transactionNumber,node,aux);
			adjustLabelTreeAfterInsertion(dbsm,transactionNumber,id,aux);
		    }
	    }
    }

    /**
     * Deletes a specified node from the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the non primitive datum that represents the collection.
     * @param node The node to delete.
     */
    static public void remove(DbStorageManager dbsm, long transactionNumber,
			      PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	removeFromIdTree(dbsm,transactionNumber,id,node);
	removeFromIdList(dbsm,transactionNumber,id,node);
	removeFromLabelTree(dbsm,transactionNumber,id,node);
	removeFromLabelList(dbsm,transactionNumber,id,node);
	dbsm.dropUnit(transactionNumber,DbStorageManager.CONTENT_SEGMENT,node);
    }


    // Public Methods

    /**
     * Checks if a specified labeled semistructured datum belongs to the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the non primitive datum that represents the collection.
     * @param lssd The labeled semistructured datum.
     * @return <code>true</code> if the labeled semistructured datum is in the collection,
     * <code>false</code> otherwise.
     */
    static public boolean contains(DbStorageManager dbsm, long transactionNumber,
				   PhysicalID id, PhysicalLabeledSSD lssd)
	throws StorageException, IOException
    {
	DbUnitAddress aux=new DbUnitAddress();
	aux=findNearestInIdTree(dbsm,transactionNumber,id,lssd);
	PhysicalLabeledSSD lssdAux=getPhysicalLabeledSsdFromNode(dbsm,transactionNumber,aux);
	int r=PhysicalLabeledSSD.ByIDComparator.compare(lssd,lssdAux);
	if (r==0)
	    return true;
	else
	    return false;
    }

    /**
     * Checks if a specified labeled semistructured datum belongs to the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the non primitive datum that represents the collection.
     * @param lssdLabel The label of the labeled semistructured datum.
     * @param lssdID The identifier of the labeled semistructured datum.
     * @return <code>true</code> if the labeled semistructured datum is in the collection,
     * <code>false</code> otherwise.
     */
    static public boolean contains(DbStorageManager dbsm, long transactionNumber,
				   PhysicalID id, String lssdLabel, PhysicalID lssdID)
	throws StorageException, IOException
    {
	return contains(dbsm,transactionNumber,id,new PhysicalLabeledSSD(lssdLabel,lssdID));
    }

    /**
     * Inserts a specified labeled semistructured datum to the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the non primitive datum that represents the collection.
     * @param lssd The labeled semistructured datum.
     */
    static public void insert(DbStorageManager dbsm, long transactionNumber,
			      PhysicalID id, PhysicalLabeledSSD lssd)
	throws StorageException, IOException, UnavailableUnitException,
	       AlreadyExistingException
    {
	DbUnitAddress aux=insertById(dbsm,transactionNumber,id,lssd);
	if (!aux.isNull())
	    insertByLabel(dbsm,transactionNumber,id,aux);
	else
	    throw new AlreadyExistingException();
    }

    /**
     * Inserts a specified labeled semistructured datum to the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the non primitive datum that represents the collection.
     * @param lssdLabel The label of the labeled semistructured datum.
     * @param lssdID The identifier of the labeled semistructured datum.
     */
    static public void insert(DbStorageManager dbsm, long transactionNumber,
			      PhysicalID id, String lssdLabel, PhysicalID lssdID)
	throws StorageException, IOException, UnavailableUnitException,
	       AlreadyExistingException
    {
	insert(dbsm,transactionNumber,id,new PhysicalLabeledSSD(lssdLabel,lssdID));
    }

    /**
     * Finds the first node in the collection with the specified identifier.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the non primitive datum that represents the collection.
     * @param matchID The identifier to find.
     * @param destiny Where the node's address is stored, it will be a null address if the
     * identifier is not found.
     */
    static public void getFirstWithID(DbStorageManager dbsm, long transactionNumber,
				      PhysicalID id, PhysicalID matchID, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	current.setNull();
	DbUnitAddress target=new DbUnitAddress();
	NonPrimitiveSsdDescriptorDriver.getIdsRoot(dbsm,transactionNumber,id,target);
	PhysicalLabeledSSD aux=null;
	while (!target.isNull())
	    {
		current.copyFrom(target);
		aux=getPhysicalLabeledSsdFromNode(dbsm,transactionNumber,current);
		int r=matchID.compareTo(aux.getID());
		if (r<=0)
		    ContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,current,target);
		else
		    ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,current,target);
	    }
	if (current.isNull())
	    destiny.setNull();
	else
	    {
		if (matchID.compareTo(aux.getID())==0)
		    destiny.copyFrom(current);
		else
		    {
			ContentNodeDriver.getNextIdNode(dbsm,transactionNumber,current,current);
			if (!current.isNull())
			    {
				aux=getPhysicalLabeledSsdFromNode(dbsm,transactionNumber,current);
				if (matchID.compareTo(aux.getID())==0)
				    destiny.copyFrom(current);
				else
				    destiny.setNull();
			    }
			else
			    destiny.setNull();
		    }
	    }
    }

    /**
     * Checks if a specified identifier is in some labeled semistructured datum in the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the non primitive datum that represents the collection.
     * @param matchID The identifier to find.
     * @return <code>true</code> if the identifier is in the collection,
     * <code>false</code> otherwise.
     */
    static public boolean containsID(DbStorageManager dbsm, long transactionNumber,
				     PhysicalID id, PhysicalID matchID)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	current.setNull();
	DbUnitAddress target=new DbUnitAddress();
	NonPrimitiveSsdDescriptorDriver.getIdsRoot(dbsm,transactionNumber,id,target);
	PhysicalLabeledSSD aux=null;
	while (!target.isNull())
	    {
		current.copyFrom(target);
		aux=getPhysicalLabeledSsdFromNode(dbsm,transactionNumber,current);
		int r=matchID.compareTo(aux.getID());
		if (r==0)
		    return true;
		else if (r<0)
		    ContentNodeDriver.getLeftIdNode(dbsm,transactionNumber,current,target);
		else
		    ContentNodeDriver.getRightIdNode(dbsm,transactionNumber,current,target);
	    }
	return false;
    }

    /**
     * Finds the first node in the collection with the specified label.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the non primitive datum that represents the collection.
     * @param matchLabel The label to find.
     * @param destiny Where the node's address is stored, it will be a null address if the
     * identifier is not found.
     */
    static public void getFirstWithLabel(DbStorageManager dbsm, long transactionNumber,
					 PhysicalID id, String matchLabel, DbUnitAddress destiny)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	current.setNull();
	DbUnitAddress target=new DbUnitAddress();
	NonPrimitiveSsdDescriptorDriver.getLabelsRoot(dbsm,transactionNumber,id,target);
	PhysicalLabeledSSD aux=null;
	while (!target.isNull())
	    {
		current.copyFrom(target);
		aux=getPhysicalLabeledSsdFromNode(dbsm,transactionNumber,current);
		int r=matchLabel.compareTo(aux.getLabel());
		if (r<=0)
		    ContentNodeDriver.getLeftLabelNode(dbsm,transactionNumber,current,target);
		else
		    ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,current,target);
	    }
	if (current.isNull())
	    destiny.setNull();
	else
	    {
		if (matchLabel.compareTo(aux.getLabel())==0)
		    destiny.copyFrom(current);
		else
		    {
			ContentNodeDriver.getNextLabelNode(dbsm,transactionNumber,current,current);
			if (!current.isNull())
			    {
				aux=getPhysicalLabeledSsdFromNode(dbsm,transactionNumber,current);
				if (matchLabel.compareTo(aux.getLabel())==0)
				    destiny.copyFrom(current);
				else
				    destiny.setNull();
			    }
			else
			    destiny.setNull();
		    }
	    }
    }

    /**
     * Checks if a specified labeled is in some labeled semistructured datum in the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the non primitive datum that represents the collection.
     * @param matchLabel The labe to find.
     * @return <code>true</code> if the label is in the collection,
     * <code>false</code> otherwise.
     */
    static public boolean containsLabel(DbStorageManager dbsm, long transactionNumber,
					PhysicalID id, String matchLabel)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	current.setNull();
	DbUnitAddress target=new DbUnitAddress();
	NonPrimitiveSsdDescriptorDriver.getLabelsRoot(dbsm,transactionNumber,id,target);
	PhysicalLabeledSSD aux=null;
	while (!target.isNull())
	    {
		current.copyFrom(target);
		aux=getPhysicalLabeledSsdFromNode(dbsm,transactionNumber,current);
		int r=matchLabel.compareTo(aux.getLabel());
		if (r==0)
		    return true;
		else if (r<0)
		    ContentNodeDriver.getLeftLabelNode(dbsm,transactionNumber,current,target);
		else
		    ContentNodeDriver.getRightLabelNode(dbsm,transactionNumber,current,target);
	    }
	return false;
    }

    /**
     * Deletes a specified labeled semistructured datum from the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the non primitive datum that represents the collection.
     * @param lssd The labeled semistructured datum.
     */
    static public void delete(DbStorageManager dbsm, long transactionNumber,
			      PhysicalID id, PhysicalLabeledSSD lssd)
	throws StorageException, IOException, NotExistingException
    {
	DbUnitAddress aux=new DbUnitAddress();
	aux=findNearestInIdTree(dbsm,transactionNumber,id,lssd);
	PhysicalLabeledSSD lssdAux=getPhysicalLabeledSsdFromNode(dbsm,transactionNumber,aux);
	int r=PhysicalLabeledSSD.ByIDComparator.compare(lssd,lssdAux);
	if (r==0)
	    remove(dbsm,transactionNumber,id,aux);
	else
	    throw new NotExistingException();
    }

    /**
     * Deletes a specified labeled semistructured datum from the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the non primitive datum that represents the collection.
     * @param lssdLabel The label of the labeled semistructured datum.
     * @param lssdID The identifier of the labeled semistructured datum.
     */
    static public void delete(DbStorageManager dbsm, long transactionNumber,
			      PhysicalID id, String lssdLabel, PhysicalID lssdID)
	throws StorageException, IOException, NotExistingException
    {
	delete(dbsm,transactionNumber,id,new PhysicalLabeledSSD(lssdLabel,lssdID));
    }

    /**
     * Deletes a specified labeled semistructured datum from the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the non primitive datum that represents the collection.
     * @param deletedID The identifier of the labeled semistructured data to delete.
     */
    static public void deleteID(DbStorageManager dbsm, long transactionNumber,
			      PhysicalID id, PhysicalID deletedID)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	DbUnitAddress next=new DbUnitAddress();
	PhysicalID aux=new PhysicalID();
	getFirstWithID(dbsm,transactionNumber,id,deletedID,current);
	boolean end=false;
	while (!end && !current.isNull())
	    {
		ContentNodeDriver.getId(dbsm,transactionNumber,current,aux);
		if (aux.compareTo(deletedID)!=0)
		    end=true;
		else
		    {
			ContentNodeDriver.getNextIdNode(dbsm,transactionNumber,current,next);
			remove(dbsm,transactionNumber,id,current);
			current.copyFrom(next);
		    }
	    }
    }

    /**
     * Deletes a specified labeled semistructured datum from the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the non primitive datum that represents the collection.
     * @param deletedLabel The label of the labeled semistructured data to delete.
     */
    static public void deleteLabel(DbStorageManager dbsm, long transactionNumber,
				   PhysicalID id, String deletedLabel)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	DbUnitAddress next=new DbUnitAddress();
	String aux;
	getFirstWithLabel(dbsm,transactionNumber,id,deletedLabel,current);
	boolean end=false;
	while (!end && !current.isNull())
	    {
		aux=getLabelFromNode(dbsm,transactionNumber,current);
		if (aux.compareTo(deletedLabel)!=0)
		    end=true;
		else
		    {
			ContentNodeDriver.getNextLabelNode(dbsm,transactionNumber,current,next);
			remove(dbsm,transactionNumber,id,current);
			current.copyFrom(next);
		    }
	    }
    }
}
