package DbStorage.manager;

import DbStorage.*;
import DbStorage.driver.*;
import storage.*;
import java.io.*;

/**
 * This class manages the content of a collection of identifiers.
 * This collection is implemented using 2 data structures combined in one, 
 * an AA-tree and a double linked list are used, for ordering the collection
 * by identifier.
 */
public final class ParentsCollectionManager
{
    static private void createNode(DbStorageManager dbsm, long transactionNumber,
				   PhysicalID id,
				   DbUnitAddress newUnitAddress)
	throws StorageException, IOException, UnavailableUnitException
    {
	newUnitAddress.setNull();
	ParentsNodeDriver.createParentsNode(dbsm,transactionNumber,newUnitAddress);
	if (newUnitAddress.isNull())
	    throw new UnavailableUnitException();
	ParentsNodeDriver.setPreviousNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ParentsNodeDriver.setNextNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ParentsNodeDriver.setLeftNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ParentsNodeDriver.setRightNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ParentsNodeDriver.setParentNode(dbsm,transactionNumber,newUnitAddress,DbUnitAddress.NULL);
	ParentsNodeDriver.setHeight(dbsm,transactionNumber,newUnitAddress,(byte)1);
	ParentsNodeDriver.setId(dbsm,transactionNumber,newUnitAddress,id);
    }


    // Methods for the Tree-List Collection

    static private void insertAfterInList(DbStorageManager dbsm, long transactionNumber, PhysicalID id,
					  DbUnitAddress current, DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress targetNext=new DbUnitAddress();
	ParentsNodeDriver.getNextNode(dbsm,transactionNumber,target,targetNext);
	ParentsNodeDriver.setNextNode(dbsm,transactionNumber,current,targetNext);
	ParentsNodeDriver.setPreviousNode(dbsm,transactionNumber,current,target);
	if (!targetNext.isNull())
	    ParentsNodeDriver.setPreviousNode(dbsm,transactionNumber,targetNext,current);
	ParentsNodeDriver.setNextNode(dbsm,transactionNumber,target,current);
    }

    static private void insertBeforeInList(DbStorageManager dbsm, long transactionNumber, PhysicalID id,
					   DbUnitAddress current, DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress head=new DbUnitAddress();
	SsdDescriptorDriver.getHeadParentsCollection(dbsm,transactionNumber,id,head);
	if (head.equals(target))
	    SsdDescriptorDriver.setHeadParentsCollection(dbsm,transactionNumber,id,current);
	
	DbUnitAddress targetPrevious=new DbUnitAddress();
	ParentsNodeDriver.getPreviousNode(dbsm,transactionNumber,target,targetPrevious);
	ParentsNodeDriver.setPreviousNode(dbsm,transactionNumber,current,targetPrevious);
	ParentsNodeDriver.setNextNode(dbsm,transactionNumber,current,target);
	if (!targetPrevious.isNull())
	    ParentsNodeDriver.setNextNode(dbsm,transactionNumber,targetPrevious,current);
	ParentsNodeDriver.setPreviousNode(dbsm,transactionNumber,target,current);
    }

    static private void removeFromList(DbStorageManager dbsm, long transactionNumber, PhysicalID id,
				       DbUnitAddress target)
	throws StorageException, IOException
    {
	DbUnitAddress head=new DbUnitAddress();
	SsdDescriptorDriver.getHeadParentsCollection(dbsm,transactionNumber,id,head);
	DbUnitAddress targetNext=new DbUnitAddress();
	ParentsNodeDriver.getNextNode(dbsm,transactionNumber,target,targetNext);
	DbUnitAddress targetPrevious=new DbUnitAddress();
	ParentsNodeDriver.getPreviousNode(dbsm,transactionNumber,target,targetPrevious);

	if (head.equals(target))
	    SsdDescriptorDriver.setHeadParentsCollection(dbsm,transactionNumber,id,targetNext);
	if (!targetPrevious.isNull())
	    ParentsNodeDriver.setNextNode(dbsm,transactionNumber,targetPrevious,targetNext);
	if (!targetNext.isNull())
	    ParentsNodeDriver.setPreviousNode(dbsm,transactionNumber,targetNext,targetPrevious);
	
	ParentsNodeDriver.setPreviousNode(dbsm,transactionNumber,target,DbUnitAddress.NULL);
	ParentsNodeDriver.setNextNode(dbsm,transactionNumber,target,DbUnitAddress.NULL);
    }

    // Pre: node!=null && node.left!=null
    static private DbUnitAddress rotateWithLeftChildInTree(DbStorageManager dbsm, long transactionNumber,
							   PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress m=new DbUnitAddress();
	ParentsNodeDriver.getLeftNode(dbsm,transactionNumber,node,m);
	
	DbUnitAddress aux=new DbUnitAddress();
	ParentsNodeDriver.getRightNode(dbsm,transactionNumber,m,aux);
	ParentsNodeDriver.setLeftNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    ParentsNodeDriver.setParentNode(dbsm,transactionNumber,aux,node);
	ParentsNodeDriver.setRightNode(dbsm,transactionNumber,m,node);
	
	ParentsNodeDriver.getParentNode(dbsm,transactionNumber,node,aux);
	if (aux.isNull())
	    SsdDescriptorDriver.setRootParentsCollection(dbsm,transactionNumber,id,m);
	else
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		ParentsNodeDriver.getLeftNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(node))
		    ParentsNodeDriver.setLeftNode(dbsm,transactionNumber,aux,m);
		else
		    ParentsNodeDriver.setRightNode(dbsm,transactionNumber,aux,m);
	    }
	ParentsNodeDriver.setParentNode(dbsm,transactionNumber,m,aux);
	ParentsNodeDriver.setParentNode(dbsm,transactionNumber,node,m);
	return m;
    }

    // Pre: node!=null && node.right!=null
     static private DbUnitAddress rotateWithRightChildInTree(DbStorageManager dbsm, long transactionNumber,
							     PhysicalID id, DbUnitAddress node)
	 throws StorageException, IOException
    {
	DbUnitAddress m=new DbUnitAddress();
	ParentsNodeDriver.getRightNode(dbsm,transactionNumber,node,m);
	
	DbUnitAddress aux=new DbUnitAddress();
	ParentsNodeDriver.getLeftNode(dbsm,transactionNumber,m,aux);
	ParentsNodeDriver.setRightNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    ParentsNodeDriver.setParentNode(dbsm,transactionNumber,aux,node);
	ParentsNodeDriver.setLeftNode(dbsm,transactionNumber,m,node);
	
	ParentsNodeDriver.getParentNode(dbsm,transactionNumber,node,aux);
	if (aux.isNull())
	    SsdDescriptorDriver.setRootParentsCollection(dbsm,transactionNumber,id,m);
	else
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		ParentsNodeDriver.getLeftNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(node))
		    ParentsNodeDriver.setLeftNode(dbsm,transactionNumber,aux,m);
		else
		    ParentsNodeDriver.setRightNode(dbsm,transactionNumber,aux,m);
	    }
	ParentsNodeDriver.setParentNode(dbsm,transactionNumber,m,aux);
	ParentsNodeDriver.setParentNode(dbsm,transactionNumber,node,m);
	return m;
    }

    // Pre: node!=null
    static private DbUnitAddress skewInTree(DbStorageManager dbsm, long transactionNumber,
					    PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress aux=new DbUnitAddress();
	ParentsNodeDriver.getLeftNode(dbsm,transactionNumber,node,aux);
	
	if (!aux.isNull() && 
	    (ParentsNodeDriver.getHeight(dbsm,transactionNumber,node)==
	     ParentsNodeDriver.getHeight(dbsm,transactionNumber,aux)))
	    return rotateWithLeftChildInTree(dbsm,transactionNumber,id,node);
	else
	    return node;
    }

    // Pre: node!=null
    static private DbUnitAddress splitInTree(DbStorageManager dbsm, long transactionNumber,
					     PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	DbUnitAddress aux=new DbUnitAddress();
	ParentsNodeDriver.getRightNode(dbsm,transactionNumber,node,aux);
	if (!aux.isNull())
	    {
		DbUnitAddress aux2=new DbUnitAddress();
		ParentsNodeDriver.getRightNode(dbsm,transactionNumber,aux,aux2);
		if (!aux2.isNull() &&
		    (ParentsNodeDriver.getHeight(dbsm,transactionNumber,node)==
		     ParentsNodeDriver.getHeight(dbsm,transactionNumber,aux2)))
		    {
			DbUnitAddress aux3=rotateWithRightChildInTree(dbsm,transactionNumber,id,node);
			ParentsNodeDriver.setHeight(dbsm,transactionNumber,aux3,
						      (byte)(ParentsNodeDriver
							     .getHeight(dbsm,transactionNumber,aux3)+1));
			return aux3;
		    }
	    }
	return node;
    }

    static private DbUnitAddress findNearestInTree(DbStorageManager dbsm, long transactionNumber,
						   PhysicalID id, PhysicalID matchID)
	throws StorageException, IOException
    {
	DbUnitAddress current=new DbUnitAddress();
	current.setNull();
	DbUnitAddress target=new DbUnitAddress();
	SsdDescriptorDriver.getRootParentsCollection(dbsm,transactionNumber,id,target);
	PhysicalID aux=new PhysicalID();
	while (!target.isNull())
	    {
		current.copyFrom(target);
		ParentsNodeDriver.getId(dbsm,transactionNumber,current,aux);
		int r=matchID.compareTo(aux);
		if (r<0)
		    ParentsNodeDriver.getLeftNode(dbsm,transactionNumber,current,target);
		else if (r>0)
		    ParentsNodeDriver.getRightNode(dbsm,transactionNumber,current,target);
		else
		    return current;
	    }
	return current;
    }

    static private void adjustTreeAfterInsertion(DbStorageManager dbsm, long transactionNumber,
						  PhysicalID id, DbUnitAddress node)
	throws StorageException, IOException
    {
	while (!node.isNull())
	    {
		node=skewInTree(dbsm,transactionNumber,id,node);
		node=splitInTree(dbsm,transactionNumber,id,node);
		ParentsNodeDriver.getParentNode(dbsm,transactionNumber,node,node);
	    }
    }

    static private void reinsertNodeInTree(DbStorageManager dbsm, long transactionNumber,
					    PhysicalID id, DbUnitAddress isolated, DbUnitAddress deleted)
	throws StorageException, IOException
    {
	if (!isolated.equals(deleted))
	    {
		ParentsNodeDriver.setHeight(dbsm,transactionNumber,isolated,
					    ParentsNodeDriver.getHeight(dbsm,transactionNumber,deleted));
		DbUnitAddress aux=new DbUnitAddress();
		ParentsNodeDriver.getLeftNode(dbsm,transactionNumber,deleted,aux);
		ParentsNodeDriver.setLeftNode(dbsm,transactionNumber,isolated,aux);
		ParentsNodeDriver.getRightNode(dbsm,transactionNumber,deleted,aux);
		ParentsNodeDriver.setRightNode(dbsm,transactionNumber,isolated,aux);
		ParentsNodeDriver.getParentNode(dbsm,transactionNumber,deleted,aux);
		ParentsNodeDriver.setParentNode(dbsm,transactionNumber,isolated,aux);
		
		if (aux.isNull())
		    SsdDescriptorDriver.setRootParentsCollection(dbsm,transactionNumber,id,isolated);
		else
		    {
			DbUnitAddress aux2=new DbUnitAddress();
			ParentsNodeDriver.getLeftNode(dbsm,transactionNumber,aux,aux2);
			if (aux2.equals(deleted))
			    ParentsNodeDriver.setLeftNode(dbsm,transactionNumber,aux,isolated);
			else
			    ParentsNodeDriver.setRightNode(dbsm,transactionNumber,aux,isolated);
		    }

		ParentsNodeDriver.getLeftNode(dbsm,transactionNumber,deleted,aux);
		if (!aux.isNull())
		    ParentsNodeDriver.setParentNode(dbsm,transactionNumber,aux,isolated);
		ParentsNodeDriver.getRightNode(dbsm,transactionNumber,deleted,aux);
		if (!aux.isNull())
		    ParentsNodeDriver.setParentNode(dbsm,transactionNumber,aux,isolated);	
	    }
	
	ParentsNodeDriver.setHeight(dbsm,transactionNumber,deleted,(byte)0);
	ParentsNodeDriver.setLeftNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
	ParentsNodeDriver.setRightNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
	ParentsNodeDriver.setParentNode(dbsm,transactionNumber,deleted,DbUnitAddress.NULL);
    }


    static private void removeFromTree(DbStorageManager dbsm, long transactionNumber,
				       PhysicalID id, DbUnitAddress deleted)
	throws StorageException, IOException
    {
	DbUnitAddress victim=new DbUnitAddress();
	DbUnitAddress adjusted=new DbUnitAddress();
	DbUnitAddress aux=new DbUnitAddress();
	DbUnitAddress aux2=new DbUnitAddress();

	ParentsNodeDriver.getRightNode(dbsm,transactionNumber,deleted,aux);
	if (!aux.isNull())
	    ParentsNodeDriver.getNextNode(dbsm,transactionNumber,deleted,victim);
	else
	    victim.copyFrom(deleted);
	     
	ParentsNodeDriver.getParentNode(dbsm,transactionNumber,victim,aux);
	if (aux.equals(deleted))
	    adjusted.copyFrom(victim);
	else
	    ParentsNodeDriver.getParentNode(dbsm,transactionNumber,victim,adjusted);
	if (aux.isNull())
	    SsdDescriptorDriver.setRootParentsCollection(dbsm,transactionNumber,id,DbUnitAddress.NULL);
	else
	    {
		ParentsNodeDriver.getLeftNode(dbsm,transactionNumber,aux,aux2);
		if (aux2.equals(victim))
		    {
			DbUnitAddress aux3=new DbUnitAddress();
			ParentsNodeDriver.getRightNode(dbsm,transactionNumber,victim,aux3);
			ParentsNodeDriver.setLeftNode(dbsm,transactionNumber,aux,aux3);
			ParentsNodeDriver.getRightNode(dbsm,transactionNumber,victim,aux2);
			if (!aux2.isNull())
			    ParentsNodeDriver.setParentNode(dbsm,transactionNumber,aux2,aux);
		    }
		else
		    {
			ParentsNodeDriver.getRightNode(dbsm,transactionNumber,victim,aux2);
			if (!aux2.isNull())
			    ParentsNodeDriver.setParentNode(dbsm,transactionNumber,aux2,aux);
			ParentsNodeDriver.setRightNode(dbsm,transactionNumber,aux,aux2);
		    }
	    }
	reinsertNodeInTree(dbsm,transactionNumber,id,victim,deleted);

	while (!adjusted.isNull())
	    {
		ParentsNodeDriver.getLeftNode(dbsm,transactionNumber,adjusted,aux);
		ParentsNodeDriver.getRightNode(dbsm,transactionNumber,adjusted,aux2);
		byte adjustedHeight=ParentsNodeDriver.getHeight(dbsm,transactionNumber,adjusted);
		if (((!aux.isNull() && 
		      ParentsNodeDriver.getHeight(dbsm,transactionNumber,aux)<adjustedHeight-1) ||
		     (aux.isNull() && 0<adjustedHeight-1)) ||
		    ((!aux2.isNull() && 
		      ParentsNodeDriver.getHeight(dbsm,transactionNumber,aux2)<adjustedHeight-1) ||
		     (aux2.isNull() && 0<adjustedHeight-1)))
		    {
			ParentsNodeDriver.setHeight(dbsm,transactionNumber,adjusted,
						      (byte)(ParentsNodeDriver.getHeight(dbsm,
											 transactionNumber,
											 adjusted)-1));
			adjustedHeight--;
			if (!aux2.isNull() && 
			    ParentsNodeDriver.getHeight(dbsm,transactionNumber,aux2)>adjustedHeight)
			    ParentsNodeDriver.setHeight(dbsm,transactionNumber,aux2,adjustedHeight);
			adjusted=skewInTree(dbsm,transactionNumber,id,adjusted);
			ParentsNodeDriver.getRightNode(dbsm,transactionNumber,adjusted,aux2);
			if (!aux2.isNull())
			    {
				ParentsNodeDriver.setRightNode(dbsm,transactionNumber,adjusted,
								 skewInTree(dbsm,transactionNumber,id,aux2));
				ParentsNodeDriver.getRightNode(dbsm,transactionNumber,adjusted,aux2);
				DbUnitAddress aux3=new DbUnitAddress();
				ParentsNodeDriver.getRightNode(dbsm,transactionNumber,aux2,aux3);
				if (!aux3.isNull())
				    ParentsNodeDriver.setRightNode(dbsm,transactionNumber,aux2,
								   skewInTree(dbsm,transactionNumber,id,aux3));
			    }
			adjusted=splitInTree(dbsm,transactionNumber,id,adjusted);
			ParentsNodeDriver.getRightNode(dbsm,transactionNumber,adjusted,aux2);
			if (!aux2.isNull())
			    ParentsNodeDriver.setRightNode(dbsm,transactionNumber,adjusted,
							   splitInTree(dbsm,transactionNumber,id,aux2));
		    }
		ParentsNodeDriver.getParentNode(dbsm,transactionNumber,adjusted,adjusted);
	    }	
    }


    // Pre: lssd is valid.
    static private DbUnitAddress insertInTree(DbStorageManager dbsm, long transactionNumber,
					      PhysicalID id, PhysicalID insertedID)
	throws StorageException, IOException, UnavailableUnitException
    {
	DbUnitAddress aux=new DbUnitAddress();
	SsdDescriptorDriver.getRootParentsCollection(dbsm,transactionNumber,id,aux);
	if (aux.isNull())
	    {
		createNode(dbsm,transactionNumber,insertedID,aux);
		SsdDescriptorDriver.setHeadParentsCollection(dbsm,transactionNumber,id,aux);
		SsdDescriptorDriver.setRootParentsCollection(dbsm,transactionNumber,id,aux);
		return aux;
	    }
	else
	    {
		aux=findNearestInTree(dbsm,transactionNumber,id,insertedID);
		PhysicalID auxID=new PhysicalID();
		ParentsNodeDriver.getId(dbsm,transactionNumber,aux,auxID);
		int r=insertedID.compareTo(auxID);
		DbUnitAddress aux2=new DbUnitAddress();
		if (r!=0)
		    {
			createNode(dbsm,transactionNumber,insertedID,aux2);
			if (r<0)
			    {
				ParentsNodeDriver.setLeftNode(dbsm,transactionNumber,aux,aux2);
				insertBeforeInList(dbsm,transactionNumber,id,aux2,aux);
			    }
			else
			    {
				ParentsNodeDriver.setRightNode(dbsm,transactionNumber,aux,aux2);
				insertAfterInList(dbsm,transactionNumber,id,aux2,aux);
			    }
			ParentsNodeDriver.setParentNode(dbsm,transactionNumber,aux2,aux);
			adjustTreeAfterInsertion(dbsm,transactionNumber,id,aux);
			return aux2;
		    }
		else
		    {
			aux2.setNull();
			return aux2;
		    }
	    }
    }


    // Public Methods

    /**
     * Checks if a specified identifier belongs to the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the datum that represents the collection.
     * @param matchID The identifier to search.
     * @return <code>true</code> if the identifier is in the collection,
     * <code>false</code> otherwise.
     */
    static public boolean contains(DbStorageManager dbsm, long transactionNumber,
				   PhysicalID id, PhysicalID matchID)
	throws StorageException, IOException
    {
	PhysicalID auxID=new PhysicalID();
	SsdDescriptorDriver.getFirstParent(dbsm,transactionNumber,id,auxID);
	if (auxID.isNull())
	    return false;
	else if (auxID.compareTo(matchID)==0)
	    return true;
	else
	    {
		DbUnitAddress aux=new DbUnitAddress();
		aux=findNearestInTree(dbsm,transactionNumber,id,matchID);
		ParentsNodeDriver.getId(dbsm,transactionNumber,aux,auxID);
		int r=matchID.compareTo(auxID);
		if (r==0)
		    return true;
		else
		    return false;
	    }
    }

    /**
     * Inserts a specified identifier to the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the datum that represents the collection.
     * @param insertedID The identifier to insert.
     */
    static public void insert(DbStorageManager dbsm, long transactionNumber,
			      PhysicalID id, PhysicalID insertedID)
	throws StorageException, IOException, UnavailableUnitException,
	       AlreadyExistingException
    {
	PhysicalID first=new PhysicalID();
	SsdDescriptorDriver.getFirstParent(dbsm,transactionNumber,id,first);
	if (first.isNull())
	    SsdDescriptorDriver.setFirstParent(dbsm,transactionNumber,id,insertedID);
	else
	    {
		if (first.compareTo(insertedID)==0)
		    throw new AlreadyExistingException();
		else
		    {
			DbUnitAddress aux=insertInTree(dbsm,transactionNumber,id,insertedID);
			if (aux.isNull())
			    throw new AlreadyExistingException();
		    }
	    }
    }

    /**
     * Deletes a specified identifier from the collection.
     * @param dbsm The data base storage manager where the operation will be performed.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of datum that represents the collection.
     * @param deletedID The identifier of the datum to delete.
     */
    static public void delete(DbStorageManager dbsm, long transactionNumber,
			      PhysicalID id, PhysicalID deletedID)
	throws StorageException, IOException, NotExistingException
    {
	PhysicalID auxID=new PhysicalID();
	DbUnitAddress aux=new DbUnitAddress();
	SsdDescriptorDriver.getFirstParent(dbsm,transactionNumber,id,auxID);
	if (auxID.isNull())
	    throw new NotExistingException();
	else if (auxID.compareTo(deletedID)==0)
	    {
		SsdDescriptorDriver.getHeadParentsCollection(dbsm,transactionNumber,id,aux);
		if (!aux.isNull())
		    {
			ParentsNodeDriver.getId(dbsm,transactionNumber,aux,auxID);
			SsdDescriptorDriver.setFirstParent(dbsm,transactionNumber,id,auxID);
			removeFromTree(dbsm,transactionNumber,id,aux);
			removeFromList(dbsm,transactionNumber,id,aux);
			dbsm.dropUnit(transactionNumber,DbStorageManager.PARENTS_SEGMENT,aux);
		    }
		else
		    SsdDescriptorDriver.setFirstParent(dbsm,transactionNumber,id,PhysicalID.NULL);
	    }
	else
	    {
		aux=findNearestInTree(dbsm,transactionNumber,id,deletedID);
		ParentsNodeDriver.getId(dbsm,transactionNumber,aux,auxID);
		int r=deletedID.compareTo(auxID);
		if (r==0)
		    {
			removeFromTree(dbsm,transactionNumber,id,aux);
			removeFromList(dbsm,transactionNumber,id,aux);
			dbsm.dropUnit(transactionNumber,DbStorageManager.PARENTS_SEGMENT,aux);
		    }
		else
		    throw new NotExistingException();
	    }
    }
}
