package DbStorage.primitives;

import DbStorage.*;
import DbStorage.driver.*;
import DbStorage.manager.*;
import storage.*;
import storage.advancedStorage.AdvancedStorageManager;
import ssdbms.ssd.*;
import ssdbms.manager.*;
import ssdbms.manager.primitives.*;
import java.io.*;

/**
 * This class implements athomic operations for a semiestructured data base.
 * @author Egar Arturo Garca Crdenas.
 */
public final class DbManager
{
    DbStorageManager dbsm;
    DbUnitAddress dictionaryDescriptor;

    /** Constructs a new Database manager. */
    public DbManager(DbStorageManager dbsm, DbUnitAddress dictionaryDescriptor)
    {
	this.dbsm=dbsm;
	this.dictionaryDescriptor=dictionaryDescriptor;
    }

    /** Returns the enclosed <code>DbStorageManager</code>. */
    public DbStorageManager getDbStorageManager()
    {
	return dbsm;
    }

    /** Returns the enclosed <code>AdvancedStorageManager</code>. */
    public AdvancedStorageManager getAdvancedStorageManager()
    {
	return dbsm.getAdvancedStorageManager();
    }


    /**
     * Creates a new non primitive semistructured datum.
     * @param transactionNumber The number of transaction. 
     * @return The identifier of the created semistructured datum.
     */
    synchronized public PhysicalID createNonPrimitive(long transactionNumber)
	throws BadFunctionException, FatalErrorException 
    {
	PhysicalID r=new PhysicalID();
	try {
	    IdManager.createNonPrimitive(dbsm,transactionNumber,r);
	    return r;
	} catch (UnavailableUnitException uue) {
	    throw new BadFunctionException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }

    /**
     * Creates a new primitive integer datum.
     * @param value The integer value of the semistructured datum to be created.
     * @param transactionNumber The number of transaction. 
     * @return The identifier of the created semistructured datum.
     */
    synchronized public PhysicalID createInteger(IntegerValue value, long transactionNumber)
	throws BadFunctionException, FatalErrorException 
    {
	PhysicalID r=new PhysicalID();
	try {
	    IdManager.createInteger(dbsm,transactionNumber,value.getValue(),r);
	    return r;
	} catch (UnavailableUnitException uue) {
	    throw new BadFunctionException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }

    /**
     * Creates a new primitive real datum.
     * @param value The real value of the semistructured datum to be created.
     * @param transactionNumber The number of transaction. 
     * @return The identifier of the created semistructured datum.
     */
    synchronized public PhysicalID createReal(RealValue value, long transactionNumber)
	throws BadFunctionException, FatalErrorException 
    {
	PhysicalID r=new PhysicalID();
	try {
	    IdManager.createReal(dbsm,transactionNumber,value.getValue(),r);
	    return r;
	} catch (UnavailableUnitException uue) {
	    throw new BadFunctionException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }

    /**
     * Creates a new primitive string datum.
     * @param value The string value of the semistructured datum to be created.
     * @param transactionNumber The number of transaction. 
     * @return The identifier of the created semistructured datum.
     */
    synchronized public PhysicalID createString(StringValue value, long transactionNumber)
	throws BadFunctionException, FatalErrorException 
    {
	PhysicalID r=new PhysicalID();
	try {
	    IdManager.createString(dbsm,transactionNumber,value.getValue(),r);
	    return r;
	} catch (UnavailableUnitException uue) {
	    throw new BadFunctionException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }

    /**
     * Creates a new primitive byte sequence datum.
     * @param value The byte sequence value of the semistructured datum to be created.
     * @param transactionNumber The number of transaction. 
     * @return The identifier of the created semistructured datum.
     */
    synchronized public PhysicalID createByteSequence(ByteSequenceValue value, long transactionNumber)
	throws BadFunctionException, FatalErrorException 
    {
	PhysicalID r=new PhysicalID();
	try {
	    IdManager.createByteSequence(dbsm,transactionNumber,value.getValue(),r);
	    return r;
	} catch (UnavailableUnitException uue) {
	    throw new BadFunctionException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }

    /**
     * Creates a new primitive BLOB datum.
     * @param transactionNumber The number of transaction. 
     * @return The identifier of the created semistructured datum.
     */
    synchronized public PhysicalID createBLOB(long transactionNumber)
	throws BadFunctionException, FatalErrorException 
    {
	PhysicalID r=new PhysicalID();
	try {
	    IdManager.createBLOB(dbsm,transactionNumber,r);
	    return r;
	} catch (UnavailableUnitException uue) {
	    throw new BadFunctionException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }

    /**
     * Creates a new primitive CLOB datum.
     * @param transactionNumber The number of transaction. 
     * @return The identifier of the created semistructured datum.
     */
    synchronized public PhysicalID createCLOB(long transactionNumber)
	throws BadFunctionException, FatalErrorException 
    {
	PhysicalID r=new PhysicalID();
	try {
	    IdManager.createCLOB(dbsm,transactionNumber,r);
	    return r;
	} catch (UnavailableUnitException uue) {
	    throw new BadFunctionException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }

    /**
     * Checks if a specified semistructured datum exist.
     * @param transactionNumber The number of transaction. 
     * @return <code>true</code> If the datum exists, <code>false</code> otherwise. 
     */
    synchronized public boolean existID(PhysicalID id, long transactionNumber)
	throws FatalErrorException
    {
	try {
	    return dbsm.isOccupiedUnit(transactionNumber,DbStorageManager.ID_SEGMENT,id);
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }


    /**
     * Returns the kind of this semistructured datum.
     * @param id The identifier of the datum to check.
     * @param transactionNumber The number of transaction. 
     * @return The numeric constant that represents the kind of value of this 
     * semistructured datum.
     */
    synchronized public int getType(PhysicalID id, long transactionNumber)
	throws NotExistingDatumException, FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		return SsdDescriptorDriver.getType(dbsm,transactionNumber,id);
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Checks if a semistructured datum with a given identifier is primitive or not.
     * @param id The datum's identifier to be checked.
     * @param transactionNumber The number of transaction. 
     * @return <code>true</code> if the datum is primitive, <code>false</code> otherwise.
     */
    synchronized public boolean isPrimitive(PhysicalID id, long transactionNumber)
	throws NotExistingDatumException, FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		return SsdDescriptorDriver.getType(dbsm,transactionNumber,id)!=SSDValue.NON_PRIMITIVE;
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Adds a labeled semistructured datum in other's content.
     * @param parent The identifier of the semistructured datum which its content will be modified, 
     * this has to be non-primitive.
     * @param childsLabel The label of the datum to be added. 
     * @param child The identifier of the datum to be added. 
     * @param transactionNumber The number of transaction. 
     */
    synchronized public void add(PhysicalID parent, Label childsLabel, PhysicalID child, 
				 long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException,
	       BadFunctionException, FatalErrorException
    {
	if (existID(parent,transactionNumber) && existID(child,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,parent)==SSDValue.NON_PRIMITIVE)
		{
		    LabeledSsdCollectionManager.insert(dbsm,transactionNumber,
						       parent,childsLabel.getText(),child);
		    ParentsCollectionManager.insert(dbsm,transactionNumber,child,parent);
		}
		else
		    throw new InvalidDatumOperationException();
	    } catch (AlreadyExistingException aee) {
	    } catch (UnavailableUnitException uue) {
		throw new BadFunctionException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Removes a labeled semistructured datum from other's content.
     * @param parent The identifier of the semistructured datum which its content will be modified.
     * @param childsLabel The label of the datum to be removed. 
     * @param child The identifier of the datum to be removed. 
     * @param transactionNumber The number of transaction. 
     */
    synchronized public void remove(PhysicalID parent, Label childsLabel, PhysicalID child, 
				    long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException, FatalErrorException
    {
	if (existID(parent,transactionNumber) && existID(child,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,parent)==SSDValue.NON_PRIMITIVE)
		    {
			LabeledSsdCollectionManager.delete(dbsm,transactionNumber,
							   parent,childsLabel.getText(),child);
			if (!LabeledSsdCollectionManager.containsID(dbsm,transactionNumber,
								    parent,child))
			    ParentsCollectionManager.delete(dbsm,transactionNumber,child,parent);
		    }
		else
		    throw new InvalidDatumOperationException();
	    } catch (NotExistingException aee) {
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Removes a datum from other's content.
     * @param parent The identifier of the semistructured datum which its content will be modified.
     * @param child The identifier of the datum to be removed. 
     * @param transactionNumber The number of transaction. 
     */
    synchronized public void removeID(PhysicalID parent, PhysicalID child, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException, FatalErrorException
    {
	if (existID(parent,transactionNumber) && existID(child,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,parent)==SSDValue.NON_PRIMITIVE)
		    {
			LabeledSsdCollectionManager.deleteID(dbsm,transactionNumber,
							     parent,child);
			ParentsCollectionManager.delete(dbsm,transactionNumber,child,parent);
		    }
		else
		    throw new InvalidDatumOperationException();
	    } catch (NotExistingException aee) {
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Removes all subdata from a specified datum with a given label.
     * @param parent The identifier of the semistructured datum which its content will be modified.
     * @param childsLabel The label of the data to be removed. 
     * @param transactionNumber The number of transaction. 
     */
    synchronized public void removeLabel(PhysicalID parent, Label childsLabel, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException, FatalErrorException
    {
	if (existID(parent,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,parent)==SSDValue.NON_PRIMITIVE)
		    {
			DbUnitAddress aux=new DbUnitAddress();
			DbUnitAddress aux2=new DbUnitAddress();
			DbUnitAddress nextAux=new DbUnitAddress();
			PhysicalID auxID=new PhysicalID();
			PhysicalID auxID2=new PhysicalID();
			LabeledSsdCollectionManager.getFirstWithLabel(dbsm,transactionNumber,parent,
								      childsLabel.getText(),aux);
			while (!aux.isNull())
			    {
				ContentNodeDriver.getNextLabelNode(dbsm,transactionNumber,aux,nextAux);
				ContentNodeDriver.getId(dbsm,transactionNumber,aux,auxID);
				LabeledSsdCollectionManager.remove(dbsm,transactionNumber,parent,aux);
				
				boolean deleteIdInParents=false;
				ContentNodeDriver.getNextIdNode(dbsm,transactionNumber,aux,aux2);
				if (!aux2.isNull())
				    {
					ContentNodeDriver.getId(dbsm,transactionNumber,aux2,auxID2);
					if (auxID.compareTo(auxID2)!=0)
					    {
						ContentNodeDriver.getPreviousIdNode(dbsm,transactionNumber,aux,aux2);
						if (!aux2.isNull())
						    {
							ContentNodeDriver.getId(dbsm,transactionNumber,aux2,auxID2);
							if (auxID.compareTo(auxID2)!=0)
							    deleteIdInParents=true;
						    }
						else
						    deleteIdInParents=true;
					    }
				    }
				else
				    {
					ContentNodeDriver.getPreviousIdNode(dbsm,transactionNumber,aux,aux2);
					if (!aux2.isNull())
					    {
						ContentNodeDriver.getId(dbsm,transactionNumber,aux2,auxID2);
						if (auxID.compareTo(auxID2)!=0)
						    deleteIdInParents=true;
					    }
					else
					    deleteIdInParents=true;
				    }
				if (deleteIdInParents)
				    try {
					ParentsCollectionManager.delete(dbsm,transactionNumber,
									auxID,parent);
				    } catch (NotExistingException nee) {}

				if (!nextAux.isNull() &&
				    LabeledSsdCollectionManager.getLabelFromNode(dbsm,transactionNumber,nextAux).
				    compareTo(childsLabel.getText())==0)
				    aux.copyFrom(nextAux);
				else
				    aux.setNull();
			    }
		    }
		else
		    throw new InvalidDatumOperationException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Returns an integer value from a specified datum.
     * @param id The identifier of the datum.
     * @param transactionNumber The number of transaction.
     * @return The content of the datum as an integer value.
     */    
    synchronized public IntegerValue getIntegerValue(PhysicalID id, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException,
	       FatalErrorException
    {	
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.INTEGER)
		    return new IntegerValue(DataConvertion.bytesToLong
					    (IdManager.getSimpleContent(dbsm,transactionNumber,id)));
		else
		    throw new InvalidDatumOperationException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Returns an real value from a specified datum.
     * @param id The identifier of the datum.
     * @param transactionNumber The number of transaction.
     * @return The content of the datum as a real value.
     */
    synchronized public RealValue getRealValue(PhysicalID id, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException,
	       FatalErrorException
    {	
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.REAL)
		    return new RealValue(DataConvertion.bytesToDouble
					    (IdManager.getSimpleContent(dbsm,transactionNumber,id)));
		else
		    throw new InvalidDatumOperationException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Returns an string value from a specified datum.
     * @param id The identifier of the datum.
     * @param transactionNumber The number of transaction.
     * @return The content of the datum as a string value.
     */
    synchronized public StringValue getStringValue(PhysicalID id, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException,
	       BadFunctionException, FatalErrorException
    {	
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.STRING)
		    return new StringValue(DataConvertion.bytesToString
					   (IdManager.getSimpleContent(dbsm,transactionNumber,id)));
		else
		    throw new InvalidDatumOperationException();
	    } catch (BadSSDValueException se) {
		throw new BadFunctionException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Returns an byte sequence value from a specified datum.
     * @param id The identifier of the datum.
     * @param transactionNumber The number of transaction.
     * @return The content of the datum as a byte sequence value.
     */
    synchronized public ByteSequenceValue getByteSequenceValue(PhysicalID id, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException,
	       BadFunctionException, FatalErrorException
    {	
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.BYTE_SEQUENCE)
		    return new ByteSequenceValue(IdManager.getSimpleContent(dbsm,transactionNumber,id));
		else
		    throw new InvalidDatumOperationException();
	    } catch (BadSSDValueException se) {
		throw new BadFunctionException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }


    /**
     * Checks if a given semistructured datum owns another given datum in its content.
     * @param parent The datum's identifier which content will be analized.
     * @param child The datum's identifier will be searched in <code>parent</code>'s content.
     * @param transactionNumber The number of transaction. 
     * @return <code>true</code> if <code>child</code> is in <code>parents</code>'s content,
     * <code>false</code> otherwise. Note: If <code>parent</code> is primitive this method
     * returns false.
     */
    synchronized public boolean contains(PhysicalID parent, PhysicalID child, long transactionNumber)
	throws NotExistingDatumException, FatalErrorException
    {
	if (existID(parent,transactionNumber) && existID(child,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,parent)==SSDValue.NON_PRIMITIVE)
		    return LabeledSsdCollectionManager.containsID(dbsm,transactionNumber,
								  parent,child);
		else
		    return false;
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Checks if a given semistructured belongs to another given datum.
     * @param child The datum's identifier will be searched in <code>parent</code>'s content.
     * @param parent The datum's identifier which content will be analized.
     * @param transactionNumber The number of transaction. 
     * @return <code>true</code> if <code>child</code> is in <code>parents</code>'s content,
     * <code>false</code> otherwise. Note: If <code>parent</code> is primitive this method
     * returns false.
     */
    synchronized public boolean belongs(PhysicalID child, PhysicalID parent, long transactionNumber)
	throws NotExistingDatumException, FatalErrorException
    {
	if (existID(child,transactionNumber) && existID(parent,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,parent)==SSDValue.NON_PRIMITIVE)
		    return ParentsCollectionManager.contains(dbsm,transactionNumber,child,parent);
		else
		    return false;
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Checks if a given semistructured datum owns a specified label.
     * @param parent The datum's identifier which content will be analized.
     * @param childLabel The label to be searched in <code>parent</code>'s content.
     * @param transactionNumber The number of transaction. 
     * @return <code>true</code> if <code>childLabel</code> is in <code>parents</code>'s content,
     * <code>false</code> otherwise. Note: If <code>parent</code> is primitive this method
     * returns false.
     */
    synchronized public boolean containsLabel(PhysicalID parent, Label childLabel, 
					      long transactionNumber)
	throws NotExistingDatumException, FatalErrorException
    {
	if (existID(parent,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,parent)==SSDValue.NON_PRIMITIVE)
		    return LabeledSsdCollectionManager.containsLabel(dbsm,transactionNumber,
								     parent,childLabel.getText());
		else
		    return false;
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }


    /**
     * Adds a SSD Table in the database.
     * @param name The name of the new table.
     * @param root The identifier of the semistructured datum which will be the root of.
     * @param transactionNumber The number of transaction. 
     * the new SSD-Table.
     */
    synchronized public void addSSDTable(Label name, PhysicalID root, long transactionNumber)
	throws NotExistingDatumException, AlreadyDefinedSSDTableException,
	       BadFunctionException, FatalErrorException
    {
	if (existID(root,transactionNumber))
	    
		try {
		    DictionaryManager.insert(dbsm,transactionNumber,dictionaryDescriptor,
					     name.getText(),root);
		    SsdDescriptorDriver.setDictionaryReferences
			(dbsm,transactionNumber,root,
			 SsdDescriptorDriver.getDictionaryReferences(dbsm,transactionNumber,root)+1);
		} catch (AlreadyExistingException aee) {
		    throw new AlreadyDefinedSSDTableException();
		} catch (UnavailableUnitException uue) {
		    throw new BadFunctionException();
		} catch (StorageException se) {
		    throw new FatalErrorException();
		} catch (IOException ioe) {
		    throw new FatalErrorException();
		}
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Returns the identifier of the semistructured datum which is the root of the
     * SSD-Table.
     * @param name The name of the SSD-Table to get its root.
     * @param transactionNumber The number of transaction. 
     * @return The root of the SSD-Table.
     */
    synchronized public PhysicalID getRoot(Label name, long transactionNumber)
	throws NotDefinedSSDTableException, FatalErrorException
    {
	try {
	    PhysicalID r=new PhysicalID();
	    DictionaryManager.getID(dbsm,transactionNumber,dictionaryDescriptor,name.getText(),r);
	    return r;
	} catch (NotExistingException aee) {
	    throw new NotDefinedSSDTableException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }
     
    /**
     * Removes a SSD-Table from the database.
     * @param name The name of the SSD-Table to be removed.
     * @param transactionNumber The number of transaction. 
     */
    synchronized public void removeSSDTable(Label name, long transactionNumber)
	throws NotDefinedSSDTableException, FatalErrorException
    {
	try {
	    PhysicalID r=new PhysicalID();
	    DictionaryManager.getID(dbsm,transactionNumber,dictionaryDescriptor,name.getText(),r);
	    DictionaryManager.deleteName(dbsm,transactionNumber,dictionaryDescriptor,name.getText());
	    SsdDescriptorDriver.setDictionaryReferences
			(dbsm,transactionNumber,r,
			 SsdDescriptorDriver.getDictionaryReferences(dbsm,transactionNumber,r)-1);
	} catch (NotExistingException aee) {
	    throw new NotDefinedSSDTableException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }

    /**
     * Removes the SSD-Tables whith the specified datum.
     * @param id The identifier of the root of the SSD-Tables to be removed.
     * @param transactionNumber The number of transaction. 
     */
    synchronized public void removeSSDTablesWithID(PhysicalID id, long transactionNumber)
	throws NotExistingDatumException, NotDefinedSSDTableException, FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		DbUnitAddress aux=new DbUnitAddress();
		DbUnitAddress aux2=new DbUnitAddress();
		DbUnitAddress nextAux=new DbUnitAddress();
		PhysicalID auxID=new PhysicalID();
		String name="";
		DictionaryManager.getFirstWithID(dbsm,transactionNumber,dictionaryDescriptor,id,aux);
		while (!aux.isNull())
		    {
			DictionaryContentNodeDriver.getNextIdNode(dbsm,transactionNumber,aux,nextAux);
						
			SsdDescriptorDriver.setDictionaryReferences
			    (dbsm,transactionNumber,id,
			     SsdDescriptorDriver.getDictionaryReferences(dbsm,transactionNumber,id)-1);
			
			DictionaryManager.remove(dbsm,transactionNumber,dictionaryDescriptor,aux);

			if (!nextAux.isNull())
			    {
				DictionaryContentNodeDriver.getId(dbsm,transactionNumber,nextAux,auxID);
				if (auxID.compareTo(id)==0)
				    aux.copyFrom(nextAux);
				else
				    aux.setNull();
			    }
			else
			    aux.setNull();
		    }
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Drops a semistructured datum giving its identifier. The semistructured datum to be 
     * droped must be empty (i.e. without subdata) or primitive and 
     * isolated (i.e. without parents and dictionary references).
     * @param id The identifier of the semistructured datum to be dropped.
     * @param transactionNumber The number of transaction. 
     */
    synchronized public void drop(PhysicalID id, long transactionNumber)
	throws NotExistingDatumException, NotIsolatedDatumException, FatalErrorException
    {
	if (existID(id,transactionNumber))
	    {
		try {
		    if (SsdDescriptorDriver.getDictionaryReferences(dbsm,transactionNumber,id)!=0)
			throw new NotIsolatedDatumException();
		    PhysicalID auxID=new PhysicalID();
		    SsdDescriptorDriver.getFirstParent(dbsm,transactionNumber,id,auxID);
		    if (!auxID.isNull())
			throw new NotIsolatedDatumException();

		    if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.NON_PRIMITIVE)
			{
			    DbUnitAddress aux=new DbUnitAddress();
			    NonPrimitiveSsdDescriptorDriver.getIdsHead(dbsm,transactionNumber,id,aux);
			    if (!aux.isNull())
				throw new NotIsolatedDatumException();
			}
		    else
			{
			    int type=SsdDescriptorDriver.getType(dbsm,transactionNumber,id);
			    if (type==SSDValue.BLOB || type==SSDValue.CLOB)
				try {
				    LobManager.setSize(dbsm,transactionNumber,id,0);
				} catch (OutOfLOBException ole) {
				}
			    else
				{
				    SimplePrimitiveSsdDescriptorDriver.
					setContentSize(dbsm,transactionNumber,id,0);
				    DbUnitAddress aux1=new DbUnitAddress();
				    DbUnitAddress aux2=new DbUnitAddress();
				    SimplePrimitiveSsdDescriptorDriver.
					getContentExtension(dbsm,transactionNumber,id,aux1);
				    while (!aux1.isNull())
					{
					    ContentExtensionDriver.getContentExtension(dbsm,transactionNumber,
										       aux1,aux2);
					    dbsm.dropUnit(transactionNumber,DbStorageManager.CONTENT_SEGMENT,aux1);
					    aux1.copyFrom(aux2);
					}
				}
			    dbsm.dropUnit(transactionNumber,DbStorageManager.ID_SEGMENT,id);
			}
		} catch (StorageException se) {
		    throw new FatalErrorException();
		} catch (IOException ioe) {
		    throw new FatalErrorException();
		} 
	    }
	else
	    throw new NotExistingDatumException();
    }


    /**
     * Changes the length of the BLOB datum.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the datum to resize.
     * @param newLength The new length of the datum.
     */
    synchronized public void setLengthOfBLOB(PhysicalID id, long newLength, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException, OutOfBoundsException,
	       FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.BLOB)
		    BlobManager.setLength(dbsm,transactionNumber,id,newLength);
		else
		    throw new InvalidDatumOperationException();
	    } catch (OutOfLOBException ole) {
		throw new OutOfBoundsException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }
    
    /**
     * Returns the length of the BLOB datum.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the datum to resize.
     * @return The length of the datum.
     */
    synchronized public long getLengthOfBLOB(PhysicalID id, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException,
	       FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.BLOB)
		    return BlobManager.getLength(dbsm,transactionNumber,id);
		else
		    throw new InvalidDatumOperationException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }
    
    /**
     * Writes a byte array into a BLOB datum.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the datum to wrote.
     * @param from The position of the BLOB from which the witting starts.
     * @param maxLength The maximum number of bytes to write.
     * @param b The array contains the bytes to write.
     * @param off The offset of <code>b</code> from which the bytes are retrieved.
     * @return The number of written bytes. 
     */
    synchronized public int writeInBLOB(PhysicalID id, long from, int maxLength, 
					byte[] b, int off, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException, OutOfBoundsException,
	       FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.BLOB)
		    return BlobManager.write(dbsm,transactionNumber,id,from,maxLength,b,off);
		else
		    throw new InvalidDatumOperationException();
	    } catch (OutOfLOBException ole) {
		throw new OutOfBoundsException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Writes a byte array into a BLOB datum.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the datum to wrote.
     * @param from The position of the BLOB from which the witting starts.
     * @param b The array contains the bytes to write.
     * @return The number of written bytes. 
     */
    synchronized public int writeInBLOB(PhysicalID id, long from, byte[] b, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException, OutOfBoundsException,
	       FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.BLOB)
		    return BlobManager.write(dbsm,transactionNumber,id,from,b);
		else
		    throw new InvalidDatumOperationException();
	    } catch (OutOfLOBException ole) {
		throw new OutOfBoundsException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Reads from a BLOB datum and stores the readed in a byte array.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the datum to read.
     * @param from The position of the BLOB from which the reading starts.
     * @param maxLength The maximum number of bytes to read.
     * @param b The array where the readed bytes will be stored.
     * @param off The offset of <code>b</code> from which the readed bytes are stored.
     * @return The number of readed bytes. 
     */
    synchronized public int readFromBLOB(PhysicalID id, long from, int maxLength, 
					 byte[] b, int off, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException, OutOfBoundsException,
	       FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.BLOB)
		    return BlobManager.read(dbsm,transactionNumber,id,from,maxLength,b,off);
		else
		    throw new InvalidDatumOperationException();
	    } catch (OutOfLOBException ole) {
		throw new OutOfBoundsException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Reads from a BLOB datum and stores the readed in a byte array.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the datum to read.
     * @param from The position of the BLOB from which the reading starts.
     * @param b The array where the readed bytes will be stored.
     * @return The number of readed bytes. 
     */
    synchronized public int readFromBLOB(PhysicalID id, long from, byte[] b, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException, OutOfBoundsException,
	       FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.BLOB)
		    return BlobManager.read(dbsm,transactionNumber,id,from,b);
		else
		    throw new InvalidDatumOperationException();
	    } catch (OutOfLOBException ole) {
		throw new OutOfBoundsException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }


    /**
     * Changes the length of the CLOB datum.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the datum to resize.
     * @param newLength The new length of the datum (in characters).
     */
    synchronized public void setLengthOfCLOB(PhysicalID id, long newLength, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException, OutOfBoundsException,
	       FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.CLOB)
		    ClobManager.setLength(dbsm,transactionNumber,id,newLength);
		else
		    throw new InvalidDatumOperationException();
	    } catch (OutOfLOBException ole) {
		throw new OutOfBoundsException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Returns the length of the CLOB datum.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the datum to resize.
     * @return The length of the datum (in characters).
     */
    synchronized public long getLengthOfCLOB(PhysicalID id, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException,
	       FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.CLOB)
		    return ClobManager.getLength(dbsm,transactionNumber,id);
		else
		    throw new InvalidDatumOperationException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }
    
    /**
     * Writes a byte array into a CLOB datum.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the datum to wrote.
     * @param from The position of the CLOB from which the witting starts.
     * @param maxLength The maximum number of characters to write.
     * @param str The string contains the characters to write.
     * @param off The offset of <code>str</code> from which the characters are retrieved.
     * @return The number of written characters. 
     */
    synchronized public int writeInCLOB(PhysicalID id, long from, int maxLength, 
					String str, int off, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException, OutOfBoundsException,
	       FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.CLOB)
		    return ClobManager.write(dbsm,transactionNumber,id,from,maxLength,str,off);
		else
		    throw new InvalidDatumOperationException();
	    } catch (OutOfLOBException ole) {
		throw new OutOfBoundsException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Writes a byte array into a CLOB datum.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the datum to wrote.
     * @param from The position of the CLOB from which the witting starts.
     * @param str The string contains the characters to write.
     * @return The number of written characters. 
     */
    synchronized public int writeInCLOB(PhysicalID id, long from, String str, long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException, OutOfBoundsException,
	       FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.CLOB)
		    return ClobManager.write(dbsm,transactionNumber,id,from,str);
		else
		    throw new InvalidDatumOperationException();
	    } catch (OutOfLOBException ole) {
		throw new OutOfBoundsException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Reads from a CLOB datum and stores the readed in a string.
     * @param transactionNumber The number of transaction.
     * @param id The identifier of the datum to read.
     * @param from The position of the CLOB from which the reading starts.
     * @param maxLength The maximum number of characters to read.
     * @return A string contains the readed characters. 
     */
    synchronized public String readFromCLOB(PhysicalID id, long from, int maxLength, 
					    long transactionNumber)
	throws NotExistingDatumException, InvalidDatumOperationException, OutOfBoundsException,
	       FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		if (SsdDescriptorDriver.getType(dbsm,transactionNumber,id)==SSDValue.CLOB)
		    return ClobManager.read(dbsm,transactionNumber,id,from,maxLength);
		else
		    throw new InvalidDatumOperationException();
	    } catch (OutOfLOBException ole) {
		throw new OutOfBoundsException();
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    } 
	else
	    throw new NotExistingDatumException();
    }


    /**
     * Returns the first parent of a datum.
     * @param id The identifier of the datum.
     * @param transactionNumber The number of transaction.
     * @return The identifier of the first parent.
     */
    synchronized public PhysicalID getFirstParent(PhysicalID id, long transactionNumber)
	throws NotExistingDatumException, FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		PhysicalID r=new PhysicalID();
		SsdDescriptorDriver.getFirstParent(dbsm,transactionNumber,id,r);
		return r;
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Returns the head of the parents' collection list of a datum.
     * @param id The identifier of the datum.
     * @param transactionNumber The number of transaction.
     * @return The address of the head node in the PARENTS segment.
     */
    synchronized public DbUnitAddress getParentsCollectionHead(PhysicalID id, long transactionNumber)
	throws NotExistingDatumException, FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		DbUnitAddress r=new DbUnitAddress();
		SsdDescriptorDriver.getHeadParentsCollection(dbsm,transactionNumber,id,r);
		return r;
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Returns the next node of a given node in a parents' collection.
     * @param node The node in the parents' collection.
     * @param transactionNumber The number of transaction.
     * @return The address of the next node in the PARENTS segment.
     */
    synchronized public DbUnitAddress getNextNodeInParentsCollection(DbUnitAddress node, long transactionNumber)
	throws InvalidUnitException, FatalErrorException
    {
	try {
	    if (dbsm.isOccupiedUnit(transactionNumber,DbStorageManager.PARENTS_SEGMENT,node))
		{
		    DbUnitAddress r=new DbUnitAddress();
		    ParentsNodeDriver.getNextNode(dbsm,transactionNumber,node,r);
		    return r;
		}
	    else
		throw new InvalidUnitException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }

    /**
     * Returns the identifier inside a given node in a parents' collection.
     * @param node The node in the parents' collection.
     * @param transactionNumber The number of transaction.
     * @return The identifier inside the node.
     */
    synchronized public PhysicalID getIDFromNodeInParentsCollection(DbUnitAddress node, long transactionNumber)
	throws InvalidUnitException, FatalErrorException
    {
	try {
	    if (dbsm.isOccupiedUnit(transactionNumber,DbStorageManager.PARENTS_SEGMENT,node))
		{
		    PhysicalID r=new PhysicalID();
		    ParentsNodeDriver.getId(dbsm,transactionNumber,node,r);
		    return r;
		}
	    else
		throw new InvalidUnitException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }


    /**
     * Returns the head of the labeled ssd collection list ordered by identifier of a datum.
     * @param id The identifier of the datum.
     * @param transactionNumber The number of transaction.
     * @return The address of the head node in the CONTENT segment.
     */
    synchronized public DbUnitAddress getLabeledSSDCollectionHeadByID(PhysicalID id, long transactionNumber)
	throws NotExistingDatumException, FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		DbUnitAddress r=new DbUnitAddress();
		NonPrimitiveSsdDescriptorDriver.getIdsHead(dbsm,transactionNumber,id,r);
		return r;
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Returns the head of the labeled ssd collection list ordered by label of a datum.
     * @param id The identifier of the datum.
     * @param transactionNumber The number of transaction.
     * @return The address of the head node in the CONTENT segment.
     */
    synchronized public DbUnitAddress getLabeledSSDCollectionHeadByLabel(PhysicalID id, long transactionNumber)
	throws NotExistingDatumException, FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		DbUnitAddress r=new DbUnitAddress();
		NonPrimitiveSsdDescriptorDriver.getLabelsHead(dbsm,transactionNumber,id,r);
		return r;
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Returns the first node of the labeled ssd collection list ordered by identifier, 
     * that contains a specified a datum.
     * @param id The identifier of the datum.
     * @param searchingID The identifier of the datum to locate.
     * @param transactionNumber The number of transaction.
     * @return The address of the head node in the CONTENT segment.
     */
    synchronized public DbUnitAddress getFirstLabeledSSDCollectionNodeWithID(PhysicalID id, 
									     PhysicalID searchingID,
									     long transactionNumber)
	throws NotExistingDatumException, FatalErrorException
    {
	if (existID(id,transactionNumber) && existID(searchingID,transactionNumber))
	    try {
		DbUnitAddress r=new DbUnitAddress();
		LabeledSsdCollectionManager.getFirstWithID(dbsm,transactionNumber,id,searchingID,r);
		return r;
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Returns the first node of the labeled ssd collection list ordered by identifier, 
     * that contains a specified a label.
     * @param id The identifier of the datum.
     * @param searchingLabel The label to locate.
     * @param transactionNumber The number of transaction.
     * @return The address of the head node in the CONTENT segment.
     */
    synchronized public DbUnitAddress getFirstLabeledSSDCollectionNodeWithLabel(PhysicalID id, 
										Label searchingLabel,
										long transactionNumber)
	throws NotExistingDatumException, FatalErrorException
    {
	if (existID(id,transactionNumber))
	    try {
		DbUnitAddress r=new DbUnitAddress();
		LabeledSsdCollectionManager.getFirstWithLabel(dbsm,transactionNumber,id,
							      searchingLabel.getText(),r);
		return r;
	    } catch (StorageException se) {
		throw new FatalErrorException();
	    } catch (IOException ioe) {
		throw new FatalErrorException();
	    }
	else
	    throw new NotExistingDatumException();
    }

    /**
     * Returns the next node of a specified node in a labeled ssd collection list ordered by identifier.
     * @param node The node in the labeled ssd colection ordered by identifier.
     * @param transactionNumber The number of transaction.
     * @return The address of the next node in the CONTENT segment.
     */
    synchronized public DbUnitAddress getNextNodeInLabeledSSDCollectionByID(DbUnitAddress node, 
									    long transactionNumber)
	throws InvalidUnitException, FatalErrorException
    {
	try {
	    if (dbsm.isOccupiedUnit(transactionNumber,DbStorageManager.CONTENT_SEGMENT,node))
		{
		    DbUnitAddress r=new DbUnitAddress();
		    ContentNodeDriver.getNextIdNode(dbsm,transactionNumber,node,r);
		    return r;
		}
	    else
		throw new InvalidUnitException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }

    /**
     * Returns the next node of a specified node in a labeled ssd collection list ordered by label.
     * @param node The node in the labeled ssd colection ordered by label.
     * @param transactionNumber The number of transaction.
     * @return The address of the next node in the CONTENT segment.
     */
    synchronized public DbUnitAddress getNextNodeInLabeledSSDCollectionByLabel(DbUnitAddress node, 
									       long transactionNumber)
	throws InvalidUnitException, FatalErrorException
    {
	try {
	    if (dbsm.isOccupiedUnit(transactionNumber,DbStorageManager.CONTENT_SEGMENT,node))
		{
		    DbUnitAddress r=new DbUnitAddress();
		    ContentNodeDriver.getNextLabelNode(dbsm,transactionNumber,node,r);
		    return r;
		}
	    else
		throw new InvalidUnitException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }

    /**
     * Returns the identifier inside a given node in a labeled ssd collection.
     * @param node The node in the parents' collection.
     * @param transactionNumber The number of transaction.
     * @return The identifier inside the node.
     */
    synchronized public PhysicalID getIDFromNodeInLabeledSSDCollection(DbUnitAddress node, long transactionNumber)
	throws InvalidUnitException, FatalErrorException
    {
	try {
	    if (dbsm.isOccupiedUnit(transactionNumber,DbStorageManager.CONTENT_SEGMENT,node))
		{
		    PhysicalID r=new PhysicalID();
		    ContentNodeDriver.getId(dbsm,transactionNumber,node,r);
		    return r;
		}
	    else
		throw new InvalidUnitException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }

    /**
     * Returns the label inside a given node in a labeled ssd collection.
     * @param node The node in the parents' collection.
     * @param transactionNumber The number of transaction.
     * @return The label inside the node.
     */
    synchronized public Label getLabelFromNodeInLabeledSSDCollection(DbUnitAddress node, long transactionNumber)
	throws BadFunctionException, InvalidUnitException, FatalErrorException
    {
	try {
	    if (dbsm.isOccupiedUnit(transactionNumber,DbStorageManager.CONTENT_SEGMENT,node))
		{
		    String r=LabeledSsdCollectionManager.getLabelFromNode(dbsm,transactionNumber,node);
		    return new Label(r);
		}
	    else
		throw new InvalidUnitException();
	} catch (BadLabelException se) {
	    throw new BadFunctionException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }

    /**
     * Returns the labeled ssd inside a given node in a labeled ssd collection.
     * @param node The node in the parents' collection.
     * @param transactionNumber The number of transaction.
     * @return The labeled ssd inside the node.
     */
    synchronized public LabeledSSD getLabeledSSDFromNodeInLabeledSSDCollection(DbUnitAddress node, 
									       long transactionNumber)
	throws BadFunctionException, InvalidUnitException, FatalErrorException
    {
	try {
	    if (dbsm.isOccupiedUnit(transactionNumber,DbStorageManager.CONTENT_SEGMENT,node))
		{
		    PhysicalID ri=new PhysicalID();
		    ContentNodeDriver.getId(dbsm,transactionNumber,node,ri);
		    String rl=LabeledSsdCollectionManager.getLabelFromNode(dbsm,transactionNumber,node);
		    return new LabeledSSD(new Label(rl),ri);
		}
	    else
		throw new InvalidUnitException();
	} catch (BadLabelException se) {
	    throw new BadFunctionException();
	} catch (SSDException se) {
	    throw new BadFunctionException();
	} catch (StorageException se) {
	    throw new FatalErrorException();
	} catch (IOException ioe) {
	    throw new FatalErrorException();
	}
    }

}



