package concurrency;

import DbStorage.PhysicalID;

/**
 * This class controls the concurrent access to the objects of a
 * semiestructured database. 
 */
public class ConcurrencyManager
{
    LockingTable lockingTable=new LockingTable();
    long nextTransactionNumber;

    /**
     * Constructs a new Concurrency Manager with a initial number of transaction.
     * @param nextTransactionNumber The next number of transaction.
     */
    public ConcurrencyManager(long nextTransactionNumber)
    {
	this.nextTransactionNumber=nextTransactionNumber;
	lockingTable=new LockingTable();
    }

    /**
     * Creates and registres a new transaction.
     * @return A <code>Transaction</code> object that represents the new transaction.
     */
    synchronized public Transaction newTransaction()
    {
	Transaction t=new Transaction(nextTransactionNumber++);
	lockingTable.addTransaction(t);
	return t;
    }

    /**
     * Removes a transaction from the concurrency manager.
     * @param transaction The transaction to remove.
     */
    synchronized public void endTransaction(Transaction transaction)
    {
	lockingTable.removeTransaction(transaction);
	notifyAll();
    }

    /**
     * Locks a database object.
     * @param transaction The transaction that locks.
     * @param obj The object to lock.
     * @param lockmode The locking mode.
     */
    synchronized public void lock(Transaction transaction, LockableObject obj, LockMode lockmode)
	throws NotActiveTransactionException, InterruptedException
    {
	while (!lockingTable.lock(transaction,obj,lockmode))
	    wait();
    }

    /**
     * Locks a database object with a specified time out.
     * @param transaction The transaction that locks.
     * @param obj The object to lock.
     * @param lockmode The locking mode.
     * @param timeOut The time out.
     */
    synchronized public void lock(Transaction transaction, LockableObject obj, LockMode lockmode,
				  long timeOut)
	throws NotActiveTransactionException, TimeOutException, InterruptedException
    {
	if (timeOut<=0)
	    lock(transaction,obj,lockmode);
	else
	    {
		long avTime=timeOut;
		long currTime=System.currentTimeMillis();
		long lastTime=0;
		boolean r=lockingTable.lock(transaction,obj,lockmode);
		while (avTime>0 && !r)
		    {
			wait(avTime);
			lastTime=currTime;
			currTime=System.currentTimeMillis();
			avTime-=currTime-lastTime;
			if (avTime>0)
			    r=lockingTable.lock(transaction,obj,lockmode);
		    }
		if (!r)
		    throw new TimeOutException();
	    }
    }

    /**
     * Locks the view of a datum.
     * @param transaction The transaction that locks.
     * @param id The datum wich view will be locked.
     * @param lockmode The locking mode.
     */
    public void lockView(Transaction transaction, PhysicalID id, LockMode lockmode)
	throws NotActiveTransactionException, InterruptedException
    {
	lock(transaction,new DatumView(id),lockmode);
    }

    /**
     * Locks the view of a datum with a specified time out.
     * @param transaction The transaction that locks.
     * @param id The datum wich view will be locked.
     * @param lockmode The locking mode.
     * @param timeOut The time out.
     */
    public void lockView(Transaction transaction, PhysicalID id, LockMode lockmode, long timeOut)
	throws NotActiveTransactionException, TimeOutException, InterruptedException
    {
	lock(transaction,new DatumView(id),lockmode,timeOut);
    }
    
    /**
     * Locks the content of a datum.
     * @param transaction The transaction that locks.
     * @param id The datum wich content will be locked.
     * @param lockmode The locking mode.
     */
    public void lockContent(Transaction transaction, PhysicalID id, LockMode lockmode)
	throws NotActiveTransactionException, InterruptedException
    {
	lock(transaction,new DatumContent(id),lockmode);
    }

    /**
     * Locks the content of a datum with a specified time out.
     * @param transaction The transaction that locks.
     * @param id The datum wich content will be locked.
     * @param lockmode The locking mode.
     * @param timeOut The time out.
     */
    public void lockContent(Transaction transaction, PhysicalID id, LockMode lockmode, long timeOut)
	throws NotActiveTransactionException, TimeOutException, InterruptedException
    {
	lock(transaction,new DatumContent(id),lockmode,timeOut);
    }

    /**
     * Locks the dictionary.
     * @param transaction The transaction that locks.
     * @param lockmode The locking mode.
     */
    public void lockDictionary(Transaction transaction, LockMode lockmode)
	throws NotActiveTransactionException, InterruptedException
    {
	lock(transaction,Dictionary.DICTIONARY,lockmode);
    }

    /**
     * Locks the dictionary with a specified time out.
     * @param transaction The transaction that locks.
     * @param lockmode The locking mode.
     * @param timeOut The time out.
     */
    public void lockDictionary(Transaction transaction, LockMode lockmode, long timeOut)
	throws NotActiveTransactionException, TimeOutException, InterruptedException
    {
	lock(transaction,Dictionary.DICTIONARY,lockmode,timeOut);
    }
}
