package concurrency;

import java.util.*;

/**
 * The locking table decides when to grant locks and when not.
 */
public final class LockingTable
{
    private TreeMap activeTransactions=new TreeMap(new Transaction.Comparator());
    private TreeMap lockedObjects=new TreeMap(new LockableObject.Comparator());

    /**
     * Adds a new transaction to the locking table.
     * @param trans The transaction to add. 
     */
    synchronized public void addTransaction(Transaction trans)
    {
	if (!activeTransactions.containsKey(trans))
	    {
		TreeMap locks=new TreeMap(LockableObject.comparator);
		activeTransactions.put(trans,locks);
	    }
    }
    
    /**
     * Checks is a lock can be granted and the lock is registred in the table.
     * @param trans The transaction that locks.
     * @param obj The object to lock.
     * @param lockmode The locking mode.
     */
    synchronized public boolean lock(Transaction trans, LockableObject obj, LockMode lockmode)
	throws NotActiveTransactionException
    {
	if (activeTransactions.containsKey(trans))
	    {
		LockMode aux;
		TreeMap lockingTransactions;
		if (lockedObjects.containsKey(obj))
		    {
			lockingTransactions=(TreeMap)(lockedObjects.get(obj));
			if (canBeLocked(trans,lockmode,lockingTransactions))
				{
				    if (lockingTransactions.containsKey(trans))
					{
					    aux=(LockMode)(lockingTransactions.get(trans));
					    aux.setMorePowerful(lockmode);
					}
				    else
					{
					    aux=new LockMode(lockmode);
					    lockingTransactions.put(trans,aux);
					    ((TreeMap)(activeTransactions.get(trans))).put(obj,aux);
					}
				}
			else
			    return false;
		    }
		else
		    {	
			aux=new LockMode(lockmode);
			lockingTransactions=new TreeMap(Transaction.comparator);
			lockingTransactions.put(trans,aux);
			lockedObjects.put(obj,lockingTransactions);
			((TreeMap)(activeTransactions.get(trans))).put(obj,aux);
		    }
		return true;
	    }
	else
	    throw new NotActiveTransactionException();
    }

    /**
     * Removes a transaction from the locking table.
     * @param trans The transaction to remove. 
     */
    synchronized public void removeTransaction(Transaction trans)
    {
	if (activeTransactions.containsKey(trans))
	    {
		Iterator i=((TreeMap)(activeTransactions.get(trans))).entrySet().iterator();
		while (i.hasNext())
		    {
			Map.Entry ei=(Map.Entry)(i.next());
			LockableObject obj=(LockableObject)(ei.getKey());
			TreeMap lockingTransactions=(TreeMap)(lockedObjects.get(obj));
			lockingTransactions.remove(trans);
			if (lockingTransactions.size()==0)
			    lockedObjects.remove(obj);
		    }
		activeTransactions.remove(trans);
	    }
    }

    private boolean canBeLocked(Transaction trans, LockMode lockmode, TreeMap lockingTransactions)
    {
	Iterator i=lockingTransactions.entrySet().iterator();
	while (i.hasNext())
	    {
		Map.Entry e=(Map.Entry)(i.next());
		if (!((Transaction)(e.getKey())).equals(trans) &&
		    !((LockMode)(e.getValue())).isCompatible(lockmode))
		    return false;
	    }
	return true;
    }

    /**
     * Shows to de locking table.
     */
    synchronized public void show()
    {
	System.out.println("Transactions:");
	Iterator i=activeTransactions.entrySet().iterator();
	while (i.hasNext())
	    {
		Map.Entry ei=(Map.Entry)(i.next());
		System.out.print("trans:" + ei.getKey().toString()+" ( ");
		Iterator j=((TreeMap)(ei.getValue())).entrySet().iterator();
		while (j.hasNext())
		    {
			Map.Entry ej=(Map.Entry)(j.next());
			System.out.print("obj:"+ej.getKey().toString()+" mode:"+ej.getValue().toString());
			if (j.hasNext())
			    System.out.print(", ");
		    }
		System.out.println(" )");
	    }

	System.out.println("Objects:");
	i=lockedObjects.entrySet().iterator();
	while (i.hasNext())
	    {
		Map.Entry ei=(Map.Entry)(i.next());
		System.out.print("obj:" + ei.getKey().toString()+" ( ");
		Iterator j=((TreeMap)(ei.getValue())).entrySet().iterator();
		while (j.hasNext())
		    {
			Map.Entry ej=(Map.Entry)(j.next());
			System.out.print("trans:"+ej.getKey().toString()+" mode:"+ej.getValue().toString());
			if (j.hasNext())
			    System.out.print(", ");
		    }
		System.out.println(" )");
	    }
    }
}
