package kernelInterface;

import ssdbms.ssd.*;
import ssdbms.manager.*;
import ssdbms.manager.primitives.*;
import DbStorage.*;
import DbStorage.primitives.*;
import concurrency.*;
import java.io.IOException;

/**
 * This class implements the primitive operations that can be used
 * for manipulation of semistructured data stored in a database.
 *
 * @author Egar Arturo Garca Crdenas.
 */
public class DbInterface implements SSDBPrimitiveManager
{
    DbManager dbm;
    ConcurrencyManager cm;
    Transaction currentTransaction;
    long timeOut;

    final String timeOutMsg="Transaction time out";

    /**
     * Constructs a new database interface with a specified time out. 
     */
    public DbInterface(DbManager dbm, ConcurrencyManager cm, long timeOut)
    {
	this.dbm=dbm;
	this.cm=cm;
	this.currentTransaction=null;
	this.timeOut=timeOut;
    }

    /**
     * Constructs a new database interface without time out. 
     */
    public DbInterface(DbManager dbm, ConcurrencyManager cm)
    {
	this(dbm,cm,0);
    }

    /**
     * Sets a new time out.
     */
    public void setTimeOut(long timeOut)
    {
	this.timeOut=timeOut;
    }

    /**
     * Returns the current time out.
     */
    public long getTimeOut()
    {
	return timeOut;
    }

    void checkTransaction()
    {
	if (currentTransaction==null)
	    currentTransaction=cm.newTransaction();
    }

    /**
     * Returns the enclosed database manager.
     */
    public DbManager getDbManager()
    {
	return dbm;
    }

    // Transaction management

    /**
     * Commits the current Transaction and leaves the system ready to init 
     * another new. 
     */
    public void commit()
	throws FatalErrorException
    {
	cm.endTransaction(currentTransaction);
	currentTransaction=null;
    }
    
    /**
     * Rollbacks the current Transaction and leaves the system ready to init
     * another new. 
     */
    public void rollback()
	throws FatalErrorException
    {
	cm.endTransaction(currentTransaction);
	currentTransaction=null;
    }

    void manageError()
	throws FatalErrorException
    {
	rollback();
    }

    // Data creation
    
    /**
     * Creates a new non primitive semistructured datum.
     * @return The identifier of the created semistructured datum.
     */
    public ID createNonPrimitive()
	throws BadFunctionException, FatalErrorException 
    {
	try {
	    checkTransaction();
	    PhysicalID id=dbm.createNonPrimitive(currentTransaction.getTransactionNumber());
	    cm.lockView(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    cm.lockContent(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    return id;
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	}catch (BadFunctionException bfe) {
	    manageError();
	    throw bfe;
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Creates a new primitive integer datum.
     * @param value The integer value of the semistructured datum to be created.
     * @return The identifier of the created semistructured datum.
     */
    public ID createInteger(IntegerValue value)
	throws BadFunctionException, FatalErrorException 
    {
	try {
	    checkTransaction();
	    PhysicalID id=dbm.createInteger(value,currentTransaction.getTransactionNumber());
	    cm.lockView(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    cm.lockContent(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    return id;
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	}catch (BadFunctionException bfe) {
	    manageError();
	    throw bfe;
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Creates a new primitive real datum.
     * @param value The real value of the semistructured datum to be created.
     * @return The identifier of the created semistructured datum.
     */
    public ID createReal(RealValue value)
	throws BadFunctionException, FatalErrorException 
    {
	try {
	    checkTransaction();
	    PhysicalID id=dbm.createReal(value,currentTransaction.getTransactionNumber());
	    cm.lockView(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    cm.lockContent(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    return id;
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	}catch (BadFunctionException bfe) {
	    manageError();
	    throw bfe;
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Creates a new primitive string datum.
     * @param value The string value of the semistructured datum to be created.
     * @return The identifier of the created semistructured datum.
     */
    public ID createString(StringValue value)
	throws BadFunctionException, FatalErrorException 
    {
	try {
	    checkTransaction();
	    PhysicalID id=dbm.createString(value,currentTransaction.getTransactionNumber());
	    cm.lockView(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    cm.lockContent(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    return id;
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	}catch (BadFunctionException bfe) {
	    manageError();
	    throw bfe;
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Creates a new primitive byte sequence datum.
     * @param value The byte sequence value of the semistructured datum to be created.
     * @return The identifier of the created semistructured datum.
     */
    public ID createByteSequence(ByteSequenceValue value)
	throws BadFunctionException, FatalErrorException 
    {
	try {
	    checkTransaction();
	    PhysicalID id=dbm.createByteSequence(value,currentTransaction.getTransactionNumber());
	    cm.lockView(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    cm.lockContent(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    return id;
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	}catch (BadFunctionException bfe) {
	    manageError();
	    throw bfe;
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Creates a new primitive BLOB datum.
     * @return The identifier of the created semistructured datum.
     */
    public ID createBLOB()
	throws BadFunctionException, FatalErrorException 
    {
	try {
	    checkTransaction();
	    PhysicalID id=dbm.createBLOB(currentTransaction.getTransactionNumber());
	    cm.lockView(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    cm.lockContent(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    return id;
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	}catch (BadFunctionException bfe) {
	    manageError();
	    throw bfe;
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Creates a new primitive CLOB datum.
     * @return The identifier of the created semistructured datum.
     */
    public ID createCLOB()
	throws BadFunctionException, FatalErrorException 
    {
	try {
	    checkTransaction();
	    PhysicalID id=dbm.createCLOB(currentTransaction.getTransactionNumber());
	    cm.lockView(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    cm.lockContent(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    return id;
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	}catch (BadFunctionException bfe) {
	    manageError();
	    throw bfe;
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    // Data content

    /**
     * Returns the kind of this semistructured datum.
     * @param id The identifier of the datum to check.
     * @return The numeric constant that represents the kind of value of this 
     * semistructured datum.
     */
    public int getType(ID id)
	throws BadFunctionException, NotExistingDatumException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockContent(currentTransaction,(PhysicalID)id,LockMode.SHARE_MODE,timeOut);
	    return dbm.getType((PhysicalID)id,currentTransaction.getTransactionNumber());
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Returns the content of a specified datum in a <code>SSDValue</code> object.
     * @param id The identifier of the datum to get the content.
     * @return The content in a <code>SSDValue</code>. 
     * Returns a <code>SetOfLabeledSSD</code> if the datum in non primitive.
     * Returns a <code>IntegerValue</code> if the datum is an integer number.
     * Returns a <code>RealValue</code> if the datum is a real number.
     * Returns a <code>StringValue</code> if the datum is a string.
     * Returns a <code>ByteSequenceValue</code> if the datum is a byte sequence.
     * Returns a <code>BLOBManager</code> if the datum is BLOB.
     * Returns a <code>CLOBManager</code> if the datum is CLOB.
     */
    public SSDValue getContent(ID id)
	throws NotExistingDatumException, BadFunctionException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockContent(currentTransaction,(PhysicalID)id,LockMode.SHARE_MODE,timeOut);
	    int type=dbm.getType((PhysicalID)id,currentTransaction.getTransactionNumber());
	    if (type==SSDValue.NON_PRIMITIVE)
		return new AllNonPrimitiveContent((PhysicalID)id);
	    else if (type==SSDValue.INTEGER)
		return dbm.getIntegerValue((PhysicalID)id,currentTransaction.getTransactionNumber());
	    else if (type==SSDValue.REAL)
		return dbm.getRealValue((PhysicalID)id,currentTransaction.getTransactionNumber());
	    else if (type==SSDValue.STRING)
		return dbm.getStringValue((PhysicalID)id,currentTransaction.getTransactionNumber());
	    else if (type==SSDValue.BYTE_SEQUENCE)
		return dbm.getByteSequenceValue((PhysicalID)id,currentTransaction.getTransactionNumber());
	    else if (type==SSDValue.BLOB)
		return new BLOBContentManager((PhysicalID)id);
	    else if (type==SSDValue.CLOB)
		return new CLOBContentManager((PhysicalID)id);
	    else
		return null;
	} catch (InvalidDatumOperationException idoe) {
	    throw new BadFunctionException();
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Checks if a semistructured datum with a given identifier is primitive or not.
     * @param id The datum's identifier to be checked.
     * @return <code>true</code> if the datum is primitive, <code>false</code> otherwise.
     */
    public boolean isPrimitive(ID id)
	throws BadFunctionException, NotExistingDatumException, FatalErrorException
    {
    	try {
	    checkTransaction();
	    cm.lockContent(currentTransaction,(PhysicalID)id,LockMode.SHARE_MODE,timeOut);
	    return dbm.isPrimitive((PhysicalID)id,currentTransaction.getTransactionNumber());
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Adds a labeled semistructured datum in other's content.
     * @param parent The identifier of the semistructured datum which its content will be modified, 
     * this has to be non-primitive.
     * @param childsLabel The label of the datum to be added. 
     * @param child The identifier of the datum to be added. 
     */
    public void add(ID parent, Label childsLabel, ID child)
	throws NotExistingDatumException, InvalidDatumOperationException,
	       BadFunctionException, FatalErrorException
    {
    	try {
	    checkTransaction();
	    cm.lockContent(currentTransaction,(PhysicalID)parent,LockMode.EXCLUSIVE_MODE,timeOut);
	    cm.lockView(currentTransaction,(PhysicalID)child,LockMode.EXCLUSIVE_MODE,timeOut);
	    dbm.add((PhysicalID)parent,childsLabel,(PhysicalID)child,
		    currentTransaction.getTransactionNumber());
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	}catch (BadFunctionException bfe) {
	    manageError();
	    throw bfe;
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Removes a labeled semistructured datum from other's content.
     * @param parent The identifier of the semistructured datum which its content will be modified.
     * @param childsLabel The label of the datum to be removed. 
     * @param child The identifier of the datum to be removed. 
     */
    public void remove(ID parent, Label childsLabel, ID child)
	throws BadFunctionException, NotExistingDatumException, 
	       InvalidDatumOperationException, FatalErrorException
    {
    	try {
	    checkTransaction();
	    cm.lockContent(currentTransaction,(PhysicalID)parent,LockMode.EXCLUSIVE_MODE,timeOut);
	    cm.lockView(currentTransaction,(PhysicalID)child,LockMode.EXCLUSIVE_MODE,timeOut);
	    dbm.remove((PhysicalID)parent,childsLabel,(PhysicalID)child,
		       currentTransaction.getTransactionNumber());
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Removes a datum from other's content.
     * @param parent The identifier of the semistructured datum which its content will be modified.
     * @param child The identifier of the datum to be removed. 
     */
    public void removeID(ID parent, ID child)
	throws BadFunctionException, NotExistingDatumException, 
	       InvalidDatumOperationException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockContent(currentTransaction,(PhysicalID)parent,LockMode.EXCLUSIVE_MODE,timeOut);
	    cm.lockView(currentTransaction,(PhysicalID)child,LockMode.EXCLUSIVE_MODE,timeOut);
	    dbm.removeID((PhysicalID)parent,(PhysicalID)child,
		       currentTransaction.getTransactionNumber());
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Removes all subdata from a specified datum with a given label.
     * @param parent The identifier of the semistructured datum which its content will be modified.
     * @param childsLabel The label of the data to be removed. 
     */
    public void removeLabel(ID parent, Label childsLabel)
	throws NotExistingDatumException, InvalidDatumOperationException, 
	       BadFunctionException, FatalErrorException
    {
	try {
	    SetOfLabeledSSD s=getContentWithLabel(parent,childsLabel);
	    while (s.hasNext())
		{
		    LabeledSSD lssd=s.next();
		    remove(parent,lssd.getLabel(),lssd.getID());
		}
	} catch (BadFunctionException bfe) {
	    manageError();
	    throw new BadFunctionException();
	}
    }
    

    // Parents checking

    /**
     * Checks if a given semistructured datum owns another given datum in its content.
     * @param parent The datum's identifier which content will be analized.
     * @param child The datum's identifier will be searched in <code>parent</code>'s content.
     * @return <code>true</code> if <code>child</code> is in <code>parents</code>'s content,
     * <code>false</code> otherwise. Note: If <code>parent</code> is primitive this method
     * returns false.
     */
    public boolean contains(ID parent, ID child)
	throws BadFunctionException, NotExistingDatumException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockContent(currentTransaction,(PhysicalID)parent,LockMode.SHARE_MODE,timeOut);
	    //cm.lockView(currentTransaction,(PhysicalID)child,LockMode.SHARE_MODE,timeOut);
	    return dbm.contains((PhysicalID)parent,(PhysicalID)child,
				currentTransaction.getTransactionNumber());
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Checks if a given semistructured belongs to another given datum.
     * @param child The datum's identifier will be searched in <code>parent</code>'s content.
     * @param parent The datum's identifier which content will be analized.
     * @return <code>true</code> if <code>child</code> is in <code>parents</code>'s content,
     * <code>false</code> otherwise. Note: If <code>parent</code> is primitive this method
     * returns false.
     */
    public boolean belongs(ID child, ID parent)
	throws BadFunctionException, NotExistingDatumException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockView(currentTransaction,(PhysicalID)child,LockMode.SHARE_MODE,timeOut);
	    //cm.lockContent(currentTransaction,(PhysicalID)parent,LockMode.SHARE_MODE,timeOut);
	    return dbm.belongs((PhysicalID)child,(PhysicalID)parent,
			       currentTransaction.getTransactionNumber());
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Checks if a given semistructured datum owns a specified label.
     * @param parent The datum's identifier which content will be analized.
     * @param childLabel The label to be searched in <code>parent</code>'s content.
     * @return <code>true</code> if <code>childLabel</code> is in <code>parents</code>'s content,
     * <code>false</code> otherwise. Note: If <code>parent</code> is primitive this method
     * returns false.
     */
    public boolean containsLabel(ID parent, Label childLabel)
	throws BadFunctionException, NotExistingDatumException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockContent(currentTransaction,(PhysicalID)parent,LockMode.SHARE_MODE,timeOut);
	    return dbm.containsLabel((PhysicalID)parent,childLabel,
				     currentTransaction.getTransactionNumber());
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }
    

    // Data dictionary

    /**
     * Adds a SSD Table in the database.
     * @param name The name of the new table.
     * @param root The identifier of the semistructured datum which will be the root of
     * the new SSD-Table.
     */
    public void addSSDTable(Label name, ID root)
	throws NotExistingDatumException, AlreadyDefinedSSDTableException,
	       BadFunctionException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockDictionary(currentTransaction,LockMode.EXCLUSIVE_MODE,timeOut);
	    cm.lockView(currentTransaction,(PhysicalID)root,LockMode.EXCLUSIVE_MODE,timeOut);
	    dbm.addSSDTable(name,(PhysicalID)root,currentTransaction.getTransactionNumber());
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	}catch (BadFunctionException bfe) {
	    manageError();
	    throw bfe;
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Returns the identifier of the semistructured datum which is the root of the
     * SSD-Table.
     * @param name The name of the SSD-Table to get its root.
     * @return The root of the SSD-Table.
     */
    public ID getRoot(Label name)
	throws BadFunctionException, NotDefinedSSDTableException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockDictionary(currentTransaction,LockMode.SHARE_MODE,timeOut);
	    return dbm.getRoot(name,currentTransaction.getTransactionNumber());
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }
    
    /**
     * Removes a SSD-Table from the database.
     * @param name The name of the SSD-Table to be removed.
     */
    public void removeSSDTable(Label name)
	throws BadFunctionException, NotDefinedSSDTableException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockDictionary(currentTransaction,LockMode.EXCLUSIVE_MODE,timeOut);
	    PhysicalID id=dbm.getRoot(name,currentTransaction.getTransactionNumber());
	    cm.lockView(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
	    dbm.removeSSDTable(name,currentTransaction.getTransactionNumber());
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Removes the SSD-Tables whith the specified datum.
     * @param id The identifier of the root of the SSD-Tables to be removed.
     */
    public void removeSSDTablesWithID(ID id)
	throws NotExistingDatumException, NotDefinedSSDTableException, 
	       BadFunctionException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockDictionary(currentTransaction,LockMode.EXCLUSIVE_MODE,timeOut);
	    cm.lockView(currentTransaction,(PhysicalID)id,LockMode.EXCLUSIVE_MODE,timeOut);
	    dbm.removeSSDTablesWithID((PhysicalID)id,currentTransaction.getTransactionNumber());
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    // Dropping

    /**
     * Drops a semistructured datum giving its identifier. The semistructured datum to be 
     * droped must be empty (i.e. without subdata) or primitive and 
     * isolated (i.e. without parents and dictionary references).
     * @param id The identifier of the semistructured datum to be dropped.
     */
    public void drop(ID id)
	throws NotExistingDatumException, NotIsolatedDatumException, 
	       BadFunctionException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockView(currentTransaction,(PhysicalID)id,LockMode.EXCLUSIVE_MODE,timeOut);
	    cm.lockContent(currentTransaction,(PhysicalID)id,LockMode.EXCLUSIVE_MODE,timeOut);
	    dbm.drop((PhysicalID)id,currentTransaction.getTransactionNumber());
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    
    // BLOB operations

    class BLOBContentManager implements BLOBManager
    {
	PhysicalID id;

	BLOBContentManager(PhysicalID id)
	{
	    this.id=id;
	}

	public void setLength(long newLength)
	    throws OutOfBoundsException, BadFunctionException, FatalErrorException
	{
	    try {
		checkTransaction();
		cm.lockContent(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
		dbm.setLengthOfBLOB(id,newLength,currentTransaction.getTransactionNumber());
	    } catch (TimeOutException toe) {
		manageError();
		throw new BadFunctionException(timeOutMsg);
	    } catch (InterruptedException ie) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotActiveTransactionException nate) {
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    } catch (InvalidDatumOperationException idoe) {
		manageError();
		throw new BadFunctionException();
	    }
	}
    
	public long length()
	    throws BadFunctionException, FatalErrorException
	{
	    try {
		checkTransaction();
		cm.lockContent(currentTransaction,id,LockMode.SHARE_MODE,timeOut);
		return dbm.getLengthOfBLOB(id,currentTransaction.getTransactionNumber());
	    } catch (TimeOutException toe) {
		manageError();
		throw new BadFunctionException(timeOutMsg);
	    } catch (InterruptedException ie) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotActiveTransactionException nate) {
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    } catch (InvalidDatumOperationException idoe) {
		manageError();
		throw new BadFunctionException();
	    }
	}
	
	public int write(long from, byte[] b, int off, int maxLength)
	    throws OutOfBoundsException, BadFunctionException, FatalErrorException
	{
	    try {
		checkTransaction();
		cm.lockContent(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
		return dbm.writeInBLOB(id,from,maxLength,b,off,
				       currentTransaction.getTransactionNumber());
	    } catch (TimeOutException toe) {
		manageError();
		throw new BadFunctionException(timeOutMsg);
	    } catch (InterruptedException ie) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotActiveTransactionException nate) {
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    } catch (InvalidDatumOperationException idoe) {
		manageError();
		throw new BadFunctionException();
	    }
	}
	
	public int write(long from, byte[] b)
	    throws OutOfBoundsException, BadFunctionException, FatalErrorException
	{
	    try {
		checkTransaction();
		cm.lockContent(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
		return dbm.writeInBLOB(id,from,b,currentTransaction.getTransactionNumber());
	    } catch (TimeOutException toe) {
		manageError();
		throw new BadFunctionException(timeOutMsg);
	    } catch (InterruptedException ie) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotActiveTransactionException nate) {
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    } catch (InvalidDatumOperationException idoe) {
		manageError();
		throw new BadFunctionException();
	    }
	}
	
	public int read(long from, byte[] b, int off, int maxLength)
	    throws OutOfBoundsException, BadFunctionException, FatalErrorException
	{
	    try {
		checkTransaction();
		cm.lockContent(currentTransaction,id,LockMode.SHARE_MODE,timeOut);
		return dbm.readFromBLOB(id,from,maxLength,b,off,
					currentTransaction.getTransactionNumber());
	    } catch (TimeOutException toe) {
		manageError();
		throw new BadFunctionException(timeOutMsg);
	    } catch (InterruptedException ie) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotActiveTransactionException nate) {
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    } catch (InvalidDatumOperationException idoe) {
		manageError();
		throw new BadFunctionException();
	    }
	}
	
	public int read(long from, byte[] b)
	    throws OutOfBoundsException, BadFunctionException, FatalErrorException
	{
	    try {
		checkTransaction();
		cm.lockContent(currentTransaction,id,LockMode.SHARE_MODE,timeOut);
		return dbm.readFromBLOB(id,from,b,currentTransaction.getTransactionNumber());
	    } catch (TimeOutException toe) {
		manageError();
		throw new BadFunctionException(timeOutMsg);
	    } catch (InterruptedException ie) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotActiveTransactionException nate) {
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    } catch (InvalidDatumOperationException idoe) {
		manageError();
		throw new BadFunctionException();
	    }
	}
    }

    
    // CLOB Operations

    class CLOBContentManager implements CLOBManager
    {
	PhysicalID id;

	CLOBContentManager(PhysicalID id)
	{
	    this.id=id;
	}

	public void setLength(long newLength)
	    throws OutOfBoundsException, BadFunctionException, FatalErrorException
	{
	    try {
		checkTransaction();
		cm.lockContent(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
		dbm.setLengthOfCLOB(id,newLength,currentTransaction.getTransactionNumber());
	    } catch (TimeOutException toe) {
		manageError();
		throw new BadFunctionException(timeOutMsg);
	    } catch (InterruptedException ie) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotActiveTransactionException nate) {
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    } catch (InvalidDatumOperationException idoe) {
		manageError();
		throw new BadFunctionException();
	    }
	}
    
	public long length()
	    throws BadFunctionException, FatalErrorException
	{
	    try {
		checkTransaction();
		cm.lockContent(currentTransaction,id,LockMode.SHARE_MODE,timeOut);
		return dbm.getLengthOfCLOB(id,currentTransaction.getTransactionNumber());
	    } catch (TimeOutException toe) {
		manageError();
		throw new BadFunctionException(timeOutMsg);
	    } catch (InterruptedException ie) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotActiveTransactionException nate) {
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    } catch (InvalidDatumOperationException idoe) {
		manageError();
		throw new BadFunctionException();
	    }
	}
	
	public int write(long from, String str, int off, int maxLength)
	    throws OutOfBoundsException, BadFunctionException, FatalErrorException
	{
	    try {
		checkTransaction();
		cm.lockContent(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
		return dbm.writeInCLOB(id,from,maxLength,str,off,
				       currentTransaction.getTransactionNumber());
	    } catch (TimeOutException toe) {
		manageError();
		throw new BadFunctionException(timeOutMsg);
	    } catch (InterruptedException ie) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotActiveTransactionException nate) {
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    } catch (InvalidDatumOperationException idoe) {
		manageError();
		throw new BadFunctionException();
	    }
	}
	
	public int write(long from, String str)
	    throws OutOfBoundsException, BadFunctionException, FatalErrorException
	{
	    try {
		checkTransaction();
		cm.lockContent(currentTransaction,id,LockMode.EXCLUSIVE_MODE,timeOut);
		return dbm.writeInCLOB(id,from,str,currentTransaction.getTransactionNumber());
	    } catch (TimeOutException toe) {
		manageError();
		throw new BadFunctionException(timeOutMsg);
	    } catch (InterruptedException ie) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotActiveTransactionException nate) {
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    } catch (InvalidDatumOperationException idoe) {
		manageError();
		throw new BadFunctionException();
	    }
	}
	
		
	public String read(long from, int maxLength)
	    throws OutOfBoundsException, BadFunctionException, FatalErrorException
	{
	    try {
		checkTransaction();
		cm.lockContent(currentTransaction,id,LockMode.SHARE_MODE,timeOut);
		return dbm.readFromCLOB(id,from,maxLength,
					currentTransaction.getTransactionNumber());
	    } catch (TimeOutException toe) {
		manageError();
		throw new BadFunctionException(timeOutMsg);
	    } catch (InterruptedException ie) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotActiveTransactionException nate) {
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    } catch (InvalidDatumOperationException idoe) {
		manageError();
		throw new BadFunctionException();
	    }
	}
    }

    
    class ParentsSet implements SetOfID
    {
	PhysicalID id, nextID;
	DbUnitAddress current, next;
	boolean first;
	
	ParentsSet(PhysicalID id)
	{
	    this.id=id;
	    first=true;
	}

	void skip()
	    throws BadFunctionException, FatalErrorException
	{
	    try {
		if (first)
		    {
			nextID=dbm.getFirstParent(id,currentTransaction.getTransactionNumber());
			current=null;
			next=dbm.getParentsCollectionHead(id,currentTransaction.getTransactionNumber());
			first=false;
		    }
		else
		    {
			current=next;
			if (!current.isNull())
			    {
				next=dbm.getNextNodeInParentsCollection(current,
							      currentTransaction.getTransactionNumber());
				nextID=dbm.getIDFromNodeInParentsCollection(current,
							      currentTransaction.getTransactionNumber());
			    }
			else
			    nextID=PhysicalID.NULL;
		    }
	    } catch (InvalidUnitException iue) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    }
	}

	public boolean hasNext()
	    throws BadFunctionException, FatalErrorException
	{
	    if (first)
		skip();
	    return !nextID.isNull();
	}

	public ID next()
	    throws BadFunctionException, FatalErrorException
	{
	    if (first)
		skip();
	    ID i=nextID;
	    skip();
	    return i;
	}
    }

    /**
     * Returns the parents of a given semistructured datum. 
     * @param id The datum's identifier to obtain its parents.
     * @return A set contains the identifiers of the datum's parents.
     */
    public SetOfID getParents(ID id)
	throws NotExistingDatumException, BadFunctionException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockView(currentTransaction,(PhysicalID)id,LockMode.SHARE_MODE,timeOut);
	    if (dbm.existID((PhysicalID)id,currentTransaction.getTransactionNumber()))
		return new ParentsSet((PhysicalID)id);
	    else
		throw new NotExistingDatumException();
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }


    abstract class NonPrimitiveContent implements SetOfLabeledSSD
    {
	PhysicalID id; 
	LabeledSSD nextLSSD;
	DbUnitAddress current, next;
	boolean first;
	
	NonPrimitiveContent(PhysicalID id)
	{
	    this.id=id;
	    first=true;
	}
	
	abstract void skip() throws BadFunctionException, FatalErrorException;


	public boolean hasNext()
	    throws BadFunctionException, FatalErrorException
	{
	    if (first)
		skip();
	    return nextLSSD!=null;
	}

	public LabeledSSD next()
	    throws BadFunctionException, FatalErrorException
	{
	    if (first)
		skip();
	    LabeledSSD lssd=nextLSSD;
	    skip();
	    return lssd;
	}
    }

    class AllNonPrimitiveContent extends NonPrimitiveContent
    {
	AllNonPrimitiveContent(PhysicalID id)
	{
	    super(id);
	}

	void skip()
	    throws BadFunctionException, FatalErrorException
	{
	    try {
		if (first)
		    {
			current=dbm.getLabeledSSDCollectionHeadByID(id,currentTransaction.getTransactionNumber());
			first=false;
		    }
		else
		    current=next;
		    		
		if (!current.isNull())
		    {
			nextLSSD=dbm.getLabeledSSDFromNodeInLabeledSSDCollection(current,
								       currentTransaction.getTransactionNumber());
			next=dbm.getNextNodeInLabeledSSDCollectionByID(current,
								       currentTransaction.getTransactionNumber());
		    }
		else
		    nextLSSD=null;

	    } catch (InvalidUnitException iue) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    }
	}
    }

    class ByIDNonPrimitiveContent extends NonPrimitiveContent
    {
	PhysicalID child;
	
	ByIDNonPrimitiveContent(PhysicalID parent, PhysicalID child)
	{
	    super(parent);
	    this.child=child;
	}

	void skip()
	    throws BadFunctionException, FatalErrorException
	{
	    try {
		if (first)
		    {
			current=dbm.getFirstLabeledSSDCollectionNodeWithID(id,child,
						       currentTransaction.getTransactionNumber());
			first=false;
		    }
		else
		    current=next;
		    		
		if (!current.isNull())
		    {
			if (dbm.getIDFromNodeInLabeledSSDCollection(current,
								    currentTransaction.getTransactionNumber()).
			    compareTo(child)==0)
			    {
				nextLSSD=dbm.getLabeledSSDFromNodeInLabeledSSDCollection(current,
								  currentTransaction.getTransactionNumber());
				next=dbm.getNextNodeInLabeledSSDCollectionByID(current,
								  currentTransaction.getTransactionNumber());
			    }
			else
			    {
				nextLSSD=null;
				next=null;
			    }
		    }
		else
		    nextLSSD=null;

	    } catch (InvalidUnitException iue) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    }
	}
    }


    class ByLabelNonPrimitiveContent extends NonPrimitiveContent
    {
	Label child;
	
	ByLabelNonPrimitiveContent(PhysicalID parent, Label child)
	{
	    super(parent);
	    this.child=child;
	}

	void skip()
	    throws BadFunctionException, FatalErrorException
	{
	    try {
		if (first)
		    {
			current=dbm.getFirstLabeledSSDCollectionNodeWithLabel(id,child,
						       currentTransaction.getTransactionNumber());
			first=false;
		    }
		else
		    current=next;
		    		
		if (!current.isNull())
		    {
			if (dbm.getLabelFromNodeInLabeledSSDCollection(current,
						    currentTransaction.getTransactionNumber()).getText().
			    compareTo(child.getText())==0)
			    {
				nextLSSD=dbm.getLabeledSSDFromNodeInLabeledSSDCollection(current,
								  currentTransaction.getTransactionNumber());
				next=dbm.getNextNodeInLabeledSSDCollectionByLabel(current,
								  currentTransaction.getTransactionNumber());
			    }
			else
			    {
				nextLSSD=null;
				next=null;
			    }
		    }
		else
		    nextLSSD=null;

	    } catch (InvalidUnitException iue) {
		manageError();
		throw new BadFunctionException();
	    } catch (NotExistingDatumException nede) {
		manageError();
		throw new BadFunctionException();
	    }
	}
    }

    /**
     * Given a semistructured datum, this returns the semistructured data in its 
     * content with the specified identifier.
     * @param parent The datum's identifier which its content will be analized.
     * @param child The datum to search in <code>parent</code>'s content.
     * @return A set of labeled semistructured data in <code>parent</code>'s content with 
     * the <code>child</code> identifier.
     */
    public SetOfLabeledSSD getContentWithID(ID parent, ID child)
	throws NotExistingDatumException, InvalidDatumOperationException,
	       BadFunctionException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockContent(currentTransaction,(PhysicalID)parent,LockMode.SHARE_MODE,timeOut);
	    if (dbm.getType((PhysicalID)parent,currentTransaction.getTransactionNumber())==SSDValue.NON_PRIMITIVE)
		return new ByIDNonPrimitiveContent((PhysicalID)parent,(PhysicalID)child);
	    else 
		throw new InvalidDatumOperationException();
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }
    
    /**
     * Given a semistructured datum, this returns the semistructured data in its 
     * content with the specified label.
     * @param parent The datum's identifier which its content will be analized.
     * @param childLabel The label which data of <code>parent</code>'s content
     * must match to be returned.
     * @return A set of labeled semistructured data in <code>parent</code>'s content labeled 
     * as <code>childrenLabel</code>.
     */
    public SetOfLabeledSSD getContentWithLabel(ID parent, Label childLabel)
	throws NotExistingDatumException, InvalidDatumOperationException,
	       BadFunctionException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockContent(currentTransaction,(PhysicalID)parent,LockMode.SHARE_MODE,timeOut);
	    if (dbm.getType((PhysicalID)parent,currentTransaction.getTransactionNumber())==SSDValue.NON_PRIMITIVE)
		return new ByLabelNonPrimitiveContent((PhysicalID)parent,childLabel);
	    else 
		throw new InvalidDatumOperationException();
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }

    /**
     * Given a semistructured datum, this returns the semistructured data in its 
     * content.
     * @param parent The datum's identifier which its content will be analized.
     * @return The set of labeled semistructured data represents the <code>parent</code>'s content.
     */
    public SetOfLabeledSSD getAllContent(ID parent)
	throws NotExistingDatumException, InvalidDatumOperationException,
	       BadFunctionException, FatalErrorException
    {
	try {
	    checkTransaction();
	    cm.lockContent(currentTransaction,(PhysicalID)parent,LockMode.SHARE_MODE,timeOut);
	    if (dbm.getType((PhysicalID)parent,currentTransaction.getTransactionNumber())==SSDValue.NON_PRIMITIVE)
		return new AllNonPrimitiveContent((PhysicalID)parent);
	    else 
		throw new InvalidDatumOperationException();
	} catch (TimeOutException toe) {
	    manageError();
	    throw new BadFunctionException(timeOutMsg);
	} catch (InterruptedException ie) {
	    manageError();
	    throw new BadFunctionException();
	} catch (NotActiveTransactionException nate) {
	    throw new BadFunctionException();
	}
    }
}



