package storage;

import java.io.*;
import storage.primitiveStorage.*;
import storage.intermediateStorage.*;
import storage.advancedStorage.*;
import storage.*;

/**
 * This class provides methods for using units. That methods are shortcuts of
 * the underlying methods in the advanced storage component.
 */
public final class UnitsStorageManager
{
    private AdvancedStorageManager asm;

    /** 
     * Finds the first free unit in the segment.
     * @param segmentNumber The number of the segment.
     * @return A <code>UnitAddress</code> objec represents the unit found,
     *         <code>null</code> if a free unit is not found.
     */
    public UnitAddress findFreeUnit(int segmentNumber)
	throws StorageException, IOException
    {
	AdvancedStorageManager.Segment segment=asm.getSegment(segmentNumber);
	AdvancedStorageManager.Segment.SpaceNumberIterator i=segment.getSpaceNumbers();
	while (i.hasNext())
	    {
		int spaceNumber=i.next();
		AdvancedStorageManager.Segment.Space space=segment.getSpace(spaceNumber);
		long unit=space.findFirstFreeUnit();
		if (unit>=0)
		    return new UnitAddress(spaceNumber,unit);
	    }
	return null;
    }

    /** 
     * Finds the first free unit in the segment.
     * @param segmentNumber The number of the segment.
     * @param freeUnit Where the address of the free unit will be stored.
     * @return <code>true</code> if the free unit is found, <code>false</code> otherwise. 
     */
    public boolean findFreeUnit(int segmentNumber, UnitAddress freeUnit)
	throws StorageException, IOException
    {
	AdvancedStorageManager.Segment segment=asm.getSegment(segmentNumber);
	AdvancedStorageManager.Segment.SpaceNumberIterator i=segment.getSpaceNumbers();
	while (i.hasNext())
	    {
		int spaceNumber=i.next();
		AdvancedStorageManager.Segment.Space space=segment.getSpace(spaceNumber);
		long unit=space.findFirstFreeUnit();
		if (unit>=0)
		    {
			freeUnit.set(spaceNumber,unit);
			return true;
		    }
	    }
	return false;
    }


    /** 
     * Tries to find a free unit, first in the same page, second in the same space and 
     * finally in the same segment of a specified unit.
     * @param segmentNumber The number of the segment of the close unit.
     * @param spaceNumber The number of the space of the close unit.
     * @param unitNumber The unit number of the close unit. 
     * @return A <code>UnitAddress</code> objec represents the unit found,
     *         <code>null</code> if a free unit is not found.
     */
    public UnitAddress findFreeUnitAsCloserAsPosibleOf(int segmentNumber, int spaceNumber, long unitNumber)
	throws StorageException, IOException
    {
	AdvancedStorageManager.Segment segment=asm.getSegment(segmentNumber);
	AdvancedStorageManager.Segment.Space space=segment.getSpace(spaceNumber);
	
	// In the same page
	long unit=space.findFreeUnitCloseOf(unitNumber);
	if (unit>=0)
	    return new UnitAddress(spaceNumber,unit);
	
	// In the same segment
	unit=space.findFirstFreeUnit();
	if (unit>=0)
	    return new UnitAddress(spaceNumber,unit);
	
	// In all segments
	AdvancedStorageManager.Segment.SpaceNumberIterator i=segment.getSpaceNumbers();
	while (i.hasNext())
	    {
		int otherSpaceNumber=i.next();
		if (otherSpaceNumber!=spaceNumber)
		    {
			space=segment.getSpace(otherSpaceNumber);
			unit=space.findFirstFreeUnit();
			if (unit>=0)
			    return new UnitAddress(otherSpaceNumber,unit);
		    }
	    }

	return null;
    }

    /** 
     * Tries to find a free unit, first in the same page, second in the same space and 
     * finally in the same segment of a specified unit.
     * @param segmentNumber The number of the segment of the close unit.
     * @param unitAddress unit address of the close unit. 
     * @return A <code>UnitAddress</code> objec represents the unit found,
     *         <code>null</code> if a free unit is not found.
     */
    public UnitAddress findFreeUnitAsCloserAsPosibleOf(int segmentNumber, UnitAddress unitAddress)
	throws StorageException, IOException
    {
	return findFreeUnitAsCloserAsPosibleOf(segmentNumber,unitAddress.getSpaceNumber(),
					       unitAddress.getUnitNumber());
    }

    /** 
     * Tries to find a free unit, first in the same page, second in the same space and 
     * finally in the same segment of a specified unit.
     * @param segmentNumber The number of the segment of the close unit.
     * @param spaceNumber The number of the space of the close unit.
     * @param unitNumber The unit number of the close unit. 
     * @param freeUnit Where the address of the new unit will be stored.
     * @return <code>true</code> if the unit is found,
     *         <code>false</code> otherwise.
     */
    public boolean findFreeUnitAsCloserAsPosibleOf(int segmentNumber, int spaceNumber, long unitNumber, 
					      UnitAddress freeUnit)
	throws StorageException, IOException
    {
	AdvancedStorageManager.Segment segment=asm.getSegment(segmentNumber);
	AdvancedStorageManager.Segment.Space space=segment.getSpace(spaceNumber);
	
	// In the same page
	long unit=space.findFreeUnitCloseOf(unitNumber);
	if (unit>=0)
	    {
		freeUnit.set(spaceNumber,unitNumber);
		return true;
	    }

	// In the same segment
	unit=space.findFirstFreeUnit();
	if (unit>=0)
	    {
		freeUnit.set(spaceNumber,unitNumber);
		return true;
	    }
	
	// In all segments
	AdvancedStorageManager.Segment.SpaceNumberIterator i=segment.getSpaceNumbers();
	while (i.hasNext())
	    {
		int otherSpaceNumber=i.next();
		if (otherSpaceNumber!=spaceNumber)
		    {
			space=segment.getSpace(otherSpaceNumber);
			unit=space.findFirstFreeUnit();
			if (unit>=0)
			    {
				freeUnit.set(spaceNumber,unitNumber);
				return true;
			    }
		    }
	    }

	return false;
    }

    /** 
     * Tries to find a free unit, first in the same page, second in the same space and 
     * finally in the same segment of a specified unit.
     * @param segmentNumber The number of the segment of the close unit.
     * @param unitAddress unit address of the close unit. 
     * @param freeUnit Where the address of the new unit will be stored.
     * @return <code>true</code> if the unit is found,
     *         <code>false</code> otherwise.
     */
    public boolean findFreeUnitAsCloserAsPosibleOf(int segmentNumber, UnitAddress unitAddress, 
						   UnitAddress freeUnit)
	throws StorageException, IOException
    {
	return findFreeUnitAsCloserAsPosibleOf(segmentNumber,unitAddress.getSpaceNumber(),
					       unitAddress.getUnitNumber(),freeUnit);
    }

    // ----

    /** 
     * Adds a unit with specifified segment, space and unit numbers, i.e. occups that unit.
     * @param segmentNumber The number of the segment of the specified unit.
     * @param spaceNumber The number of the space of the specified unit.
     * @param unitNumber The unit number of the specified unit. 
     */
    public void addUnit(int segmentNumber, int spaceNumber, long unitNumber)
	throws StorageException, IOException
    {
	AdvancedStorageManager.Segment.Space space=asm.getSegment(segmentNumber).getSpace(spaceNumber);
	space.occupUnit(unitNumber);
    }

    /** 
     * Adds a unit with specifified segment and unit address, i.e. occups that unit.
     * @param segmentNumber The number of the segment of the specified unit.
     * @param unitAddress unit address of the specified unit. 
     */
    public void addUnit(int segmentNumber, UnitAddress unitAddress)
	throws StorageException, IOException
    {
	addUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber());
    }

    
    /**
     * Drops a specific unit, i.e. frees the specified unit.
     * @param segmentNumber The number of the segment.
     * @param spaceNumber The number of the space.
     * @param unitNumber The number of the unit.                .
     */
    public void dropUnit(int segmentNumber, int spaceNumber, long unitNumber)
	throws StorageException, IOException
    {
	asm.getSegment(segmentNumber).getSpace(spaceNumber).freeUnit(unitNumber);
    }

    /**
     * Drops a specific unit, i.e. frees the specified unit.
     * @param segmentNumber The number of the segment.
     * @param unitAddress unit address of the specified unit. 
     */
    public void dropUnit(int segmentNumber, UnitAddress unitAddress)
	throws StorageException, IOException
    {
	dropUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber());
    }

    
    /**
     * Reads fully a unit. This method does not consider anything about the free status of the unit.
     * @param segmentNumber The number of the segment.
     * @param spaceNumber The number of the space.
     * @param unitNumber The number of the unit.                .
     * @param b The byte array which the content of the readed unit will be stored.
     * @param off The offset of <code>b</code> from which the readed content will be stored.
     */
    public void readUnit(int segmentNumber, int spaceNumber, long unitNumber, byte[] b, int off)
	throws StorageException, IOException
    {
	asm.getSegment(segmentNumber).getSpace(spaceNumber).readUnit(unitNumber,b,off);
    }

    /**
     * Reads fully a unit. This method does not consider anything about the free status of the unit.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit.                .
     * @param b The byte array which the content of the readed unit will be stored.
     * @param off The offset of <code>b</code> from which the readed content will be stored.
     */
    public void readUnit(int segmentNumber, UnitAddress unitAddress, byte[] b, int off)
	throws StorageException, IOException
    {
	readUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber(),b,off);
    }

    
    /**
     * Reads partially a unit starting from an offset and reading a specific length.
     * This method does not consider anything about the free status of the unit.
     * @param segmentNumber The number of the segment.
     * @param spaceNumber The number of the space.
     * @param unitNumber The number of the unit.                .
     * @param unitOffset The position of the unit from which the reading starts.
     * @param length The number of bytes to be readed.
     * @param b The byte array which the content of the readed unit will be stored.
     * @param off The offset of <code>b</code> from which the readed content will be stored.
     */
    public void readUnit(int segmentNumber, int spaceNumber, long unitNumber, 
			 int unitOffset, int length, byte[] b, int off)
	throws StorageException, IOException
    {
	asm.getSegment(segmentNumber).getSpace(spaceNumber).readUnit(unitNumber,unitOffset,length,b,off);
    }

    /**
     * Reads partially a unit starting from an offset and reading a specific length.
     * This method does not consider anything about the free status of the unit.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit.                .
     * @param unitOffset The position of the unit from which the reading starts.
     * @param length The number of bytes to be readed.
     * @param b The byte array which the content of the readed unit will be stored.
     * @param off The offset of <code>b</code> from which the readed content will be stored.
     */
    public void readUnit(int segmentNumber, UnitAddress unitAddress, 
			 int unitOffset, int length, byte[] b, int off)
	throws StorageException, IOException
    {
	readUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber(),unitOffset,length,b,off);
    }

    
     /**
     * Writes fully a unit.
     * This method does not consider anything about the free status of the unit.
     * @param segmentNumber The number of the segment.
     * @param spaceNumber The number of the space.
     * @param unitNumber The number of the unit.
     * @param b The byte array from which the units's content will be get.
     * @param off The offset of <code>b</code> from which the unit's content will be get.
     */
    public void writeUnit(int segmentNumber, int spaceNumber, long unitNumber, byte[] b, int off)
	throws StorageException, IOException
    {
	asm.getSegment(segmentNumber).getSpace(spaceNumber).writeUnit(unitNumber,b,off);
    }

     /**
     * Writes fully a unit.
     * This method does not consider anything about the free status of the unit.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit.
     * @param b The byte array from which the units's content will be get.
     * @param off The offset of <code>b</code> from which the unit's content will be get.
     */
    public void writeUnit(int segmentNumber, UnitAddress unitAddress, byte[] b, int off)
	throws StorageException, IOException
    {
	writeUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber(),b,off);
    }


    /**
     * Writes partially a unit starting from an offset and writing a specific length of bytes.
     * This method does not consider anything about the free status of the unit.
     * @param segmentNumber The number of the segment.
     * @param spaceNumber The number of the space.
     * @param unitNumber The number of the unit.
     * @param unitOffset The position of the unit from which the writing starts.
     * @param length The number of bytes to be wrote.
     * @param b The byte array from which the units's content will be get.
     * @param off The offset of <code>b</code> from which the unit's content will be get.
     */
    public void writeUnit(int segmentNumber, int spaceNumber, long unitNumber, 
			  int unitOffset, int length, byte[] b, int off)
	throws StorageException, IOException
    {
	asm.getSegment(segmentNumber).getSpace(spaceNumber).writeUnit(unitNumber,unitOffset,length,b,off);
    }

    /**
     * Writes partially a unit starting from an offset and writing a specific length of bytes.
     * This method does not consider anything about the free status of the unit.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit.
     * @param unitOffset The position of the unit from which the writing starts.
     * @param length The number of bytes to be wrote.
     * @param b The byte array from which the units's content will be get.
     * @param off The offset of <code>b</code> from which the unit's content will be get.
     */
    public void writeUnit(int segmentNumber, UnitAddress unitAddress, 
			  int unitOffset, int length, byte[] b, int off)
	throws StorageException, IOException
    {
	writeUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber(),unitOffset,length,b,off);
    }


    /**
     * Checks if a unit is free.
     * @param segmentNumber The number of the segment.
     * @param spaceNumber The number of the space.
     * @param unitNumber The number of the unit.
     * @return <code>true</code> if the specified unit is free, <code>false</code> otherwise.
     */
    public boolean isFreeUnit(int segmentNumber, int spaceNumber, long unitNumber)
	throws StorageException, IOException
    {
	return asm.getSegment(segmentNumber).getSpace(spaceNumber).isFreeUnit(unitNumber);
    }

    /**
     * Checks if a unit is free.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit.
     * @return <code>true</code> if the specified unit is free, <code>false</code> otherwise.
     */
    public boolean isFreeUnit(int segmentNumber, UnitAddress unitAddress)
	throws StorageException, IOException
    {
	return isFreeUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber());
    }


    /**
     * Checks if a unit is occupied.
     * @param segmentNumber The number of the segment.
     * @param spaceNumber The number of the space.
     * @param unitNumber The number of the unit.
     * @return <code>true</code> if the specified unit is occupied, <code>false</code> otherwise.
     */
    public boolean isOccupiedUnit(int segmentNumber, int spaceNumber, long unitNumber)
	throws StorageException, IOException
    {
	return asm.getSegment(segmentNumber).getSpace(spaceNumber).isOccupiedUnit(unitNumber);
    }

    /**
     * Checks if a unit is occupied.
     * @param segmentNumber The number of the segment.
     * @param unitAddress The address of the unit.
     * @return <code>true</code> if the specified unit is occupied, <code>false</code> otherwise.
     */
    public boolean isOccupiedUnit(int segmentNumber, UnitAddress unitAddress)
	throws StorageException, IOException
    {
	return isOccupiedUnit(segmentNumber,unitAddress.getSpaceNumber(),unitAddress.getUnitNumber());
    }


    /**
     * Returns the underlying primitive storage manager.
     */
    public PrimitiveStorageManager getPrimitiveStorageManager()
    {
	return asm.getPrimitiveStorageManager();
    }

    /**
     * Returns the underlying intermediate storage manager.
     */
    public IntermediateStorageManager getIntermediateStorageManager()
    {
	return asm.getIntermediateStorageManager();
    }
    
    /**
     * Returns the underlying advanced storage manager.
     */
    public AdvancedStorageManager getAdvancedStorageManager()
    {
	return asm;
    }

    /**
     * Constructs a units storage manager from an advanced storage manager.
     */
    public UnitsStorageManager(AdvancedStorageManager asm)
    {
	this.asm=asm;
    }
}
