package storage.advancedStorage;

import java.io.*;
import storage.primitiveStorage.*;
import storage.intermediateStorage.*;
import storage.*;
import java.util.*;

/**
 * This class is used to manage the operatios of the advanced storage component. 
 * The advanced storage component provides a high level storage organization.
 * In advanced storage component, the storage is divided in segments, a segmet consists 
 * of spaces, and a space consists of units. 
 * Each segment has is own unit size, the unit sizes depend on the specific needs
 * of the aplication.
 */
public final class AdvancedStorageManager
{
    private int pageSize;
    private IntermediateStorageManager ism;
        
    /** The maximum space number.*/
    static public final int MAX_SPACE_NUMBER=0x7FFFFFFF; 

    static private final class SpaceNumber
    {
	private int content;
	
	public SpaceNumber(int content)
	{
	    this.content=content;
	}
	
	public int intValue()
	{
	    return content;
	}
	
	public void setValue(int content)
	{
	    this.content=content;
	}
	
	public boolean equals(SpaceNumber cn)
	{
	    return this.content==cn.content;
	}
	
	public int compare(SpaceNumber cn)
	{
	    if (this.content<cn.content)
		return -1;
	    else if (this.content>cn.content)
		return 1;
	    else
		return 0;
	}
    }
	
    static private final class SpaceNumberComparator implements Comparator
    {
	public int compare(Object o1, Object o2)
	{
	    int i1=((SpaceNumber)o1).intValue();
	    int i2=((SpaceNumber)o2).intValue();
	    if (i1<i2)
		return -1;
	    else if (i1>i2)
		return 1;
	    else
		return 0;
	}
	
	public boolean equals(Object obj) 
	{
	    return obj instanceof SpaceNumberComparator;
	}
    }
    
    private SpaceNumber auxSpaceNumber;
    static private final SpaceNumberComparator spaceNumberComparator=new SpaceNumberComparator();

    /**
     * This class represents a segment, a segmen is a set of spaces, 
     * each space inside a segment is
     * identified by a number.
     */
    public final class Segment
    {
	private int unitSize;

	/**
	 * This class represents a space, a space is a set of units,
	 * each unit inside a space is identified by a unique number.
	 * Spaces provides methods for the free space management of units.
	 * Each space is mounted over a chunk (of primitive storage component).  
	 */
	public final class Space
	{
	    private int chunkNumber;
	    
	    private int unitsBitMapSize;
	    
	    private int numberOfUnitsPerUnitsPage;
	    private long numberOfUnitsPagesPerExtension;
	    private long numberOfExtensionsPerArea;
	    
	    private long numberOfPagesPerExtension;
	    private long numberOfPagesPerArea;
	    
	    private long maxChunkSize;
	    private long maxNumberOfPages;	
	    private long maxNumberOfAreas;
	    private long maxNumberOfExtensions;
	    private long maxNumberOfUnitsPages;
	    private long maxNumberOfUnits;	
	    
	    private long lastUnitsPageNumber;
	    
	    Space(int chunkNumber, boolean overwrite)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		this.chunkNumber=chunkNumber;
				
		if (unitSize<(int)((pageSize+1)/2))
		    {
			unitsBitMapSize=(pageSize-1)/(8*unitSize+1)+1;
			numberOfUnitsPerUnitsPage=(pageSize-unitsBitMapSize)/unitSize;
		    }
		else
		    {
			numberOfUnitsPerUnitsPage=1;
			unitsBitMapSize=0;
		    }
		numberOfUnitsPagesPerExtension = 8*pageSize;
		numberOfExtensionsPerArea=8*pageSize;
		
		numberOfPagesPerExtension = numberOfUnitsPagesPerExtension+1;	    
		numberOfPagesPerArea = numberOfExtensionsPerArea*numberOfPagesPerExtension+1;
		
		maxChunkSize = ism.getPrimitiveStorageManager().getChunk(chunkNumber).getMaxSize();
		maxNumberOfPages = maxChunkSize/pageSize;   
		if (maxNumberOfPages>3)
		    {
			long maxNumberOfAreaDescriptors = (maxNumberOfPages-1-1)/numberOfPagesPerArea+1;
			long maxNumberOfExtensionDescriptors 
			    =(maxNumberOfPages-1-maxNumberOfAreaDescriptors-1)/numberOfPagesPerExtension+1;
			maxNumberOfUnitsPages 
			    =maxNumberOfPages-1-maxNumberOfAreaDescriptors-maxNumberOfExtensionDescriptors;
			maxNumberOfUnits = maxNumberOfUnitsPages*numberOfUnitsPerUnitsPage;
			maxNumberOfExtensions = (maxNumberOfUnitsPages-1)/numberOfUnitsPagesPerExtension+1;
			maxNumberOfAreas = (maxNumberOfExtensions-1)/numberOfExtensionsPerArea+1;
		    }
		else
		    {
			maxNumberOfUnitsPages=0;
			maxNumberOfUnits=0;
			maxNumberOfExtensions=0;
			maxNumberOfAreas=0;
		    }
		if (maxNumberOfPages>0)
		    {
			if (overwrite)
			    writeLastUnitsPageNumberToSecundaryStorage(-1);
			else
			    lastUnitsPageNumber=readLastUnitsPageNumberFromSecundaryStorage();
		    }
	    }
	    
	    private void writeLastUnitsPageNumberToSecundaryStorage(long lastUnitsPageNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		this.lastUnitsPageNumber=lastUnitsPageNumber;
		ism.writeLong(chunkNumber,0,0,this.lastUnitsPageNumber);
		//ism.writeShort(chunkNumber,0,0,(short)(0xFFFFl&this.lastUnitsPageNumber));
	    }
	    
	    
	    private long readLastUnitsPageNumberFromSecundaryStorage()
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		return ism.readLong(chunkNumber,0,0);
		//return ism.readShort(chunkNumber,0,0);
	    }
	    
	    
	    // Information
	    
	    /**
	     * Returns the number (in primitive storage component) of the chunk in 
	     * wich the space is mounted.
	     */
	    public int getChunkNumber()
	    {
		return chunkNumber;
	    }
	    
	    /**
	     * Returns the maximun number of units that can be stored in the space.
	     */
	    public long getMaxNumberOfUnits()
	    {
		return maxNumberOfUnits;
	    }
	    
	    /**
	     * Returns the number of bytes that the space is really using.
	     */
	    public long getUsedSize()
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		return (pageNumberOfUnitsPage(lastUnitsPageNumber)+1)*pageSize;
	    }
	    
	    /**
	     *  Returns <code>true</code> if the space is empty, i.e. all is units are free.
	     *  Returns <code>false</code> otherwise.
	     */
	    public boolean isEmpty()
	    {
		return lastUnitsPageNumber<0;
	    }
	    
	    // + Input/Output operations
	    
	    // - Reading
	    
	    /**
	     * Reads partially a unit starting from an offset and reading a specific length.
	     * @param unitNumber The number of the unit inside the space.
	     * @param unitOffset The position of the unit from which the reading starts.
	     * @param length The number of bytes to be readed.
	     * @param b The byte array which the content of the readed unit will be stored.
	     * @param off The offset of <code>b</code> from which the readed content will be stored.
	     */
	    public void readUnit(long unitNumber, int unitOffset, int length, byte[] b, int off)
		throws PrimitiveStorageException, IntermediateStorageException, IOException,
		       UnitNumberOutOfRangeException, InvalidUnitOffsetException, OutOfUnitException
	    {
		if (unitNumber>maxNumberOfUnits)
		    throw new UnitNumberOutOfRangeException();
		if (unitOffset<0 || unitOffset>=unitSize)
		    throw new InvalidUnitOffsetException();
		if (unitOffset+length>unitSize)
		    throw new OutOfUnitException();
		
		ism.read(chunkNumber, pageNumberOfUnit(unitNumber), 
			 (int)((unitNumber%numberOfUnitsPerUnitsPage)*unitSize+unitOffset), 
			 b, off, length); 
	    }

	    /**
	     * Reads fully a unit.
	     * @param unitNumber The number of the unit inside the space.
	     * @param b The byte array which the content of the readed unit will be stored.
	     * @param off The offset of <code>b</code> from which the readed content will be stored.
	     */
	    public void readUnit(long unitNumber, byte[] b, int off)
		throws PrimitiveStorageException, IntermediateStorageException, IOException,
		       UnitNumberOutOfRangeException
	    {
		if (unitNumber>maxNumberOfUnits)
		    throw new UnitNumberOutOfRangeException();
		
		ism.read(chunkNumber, pageNumberOfUnit(unitNumber), 
			 (int)((unitNumber%numberOfUnitsPerUnitsPage)*unitSize), 
			 b, off, unitSize); 
	    }
	    
	    
	    // - Writing
	    
	    /**
	     * Writes partially a unit starting from an offset and writing a specific length of bytes.
	     * @param unitNumber The number of the unit.
	     * @param unitOffset The position of the unit from which the writing starts.
	     * @param b The byte array from which the units's content will be get.
	     * @param off The offset of <code>b</code> from which the unit's content will be get.
	     * @param length The number of bytes to be wrote.
	     */
	    public void writeUnit(long unitNumber, int unitOffset, int length, byte[] b, int off)
		throws PrimitiveStorageException, IntermediateStorageException, IOException,
		       UnitNumberOutOfRangeException, InvalidUnitOffsetException, OutOfUnitException
	    {
		if (unitNumber>maxNumberOfUnits)
		    throw new UnitNumberOutOfRangeException();
		if (unitOffset<0 || unitOffset>=unitSize)
		    throw new InvalidUnitOffsetException();
		if (unitOffset+length>unitSize)
		    throw new OutOfUnitException();
		
		ism.write(chunkNumber, pageNumberOfUnit(unitNumber), 
			  (int)((unitNumber%numberOfUnitsPerUnitsPage)*unitSize+unitOffset), 
			  b, off, length);
	    }

	    /**
	     * Writes fully a unit.
	     * @param unitNumber The number of the unit.
	     * @param b The byte array from which the units's content will be get.
	     * @param off The offset of <code>b</code> from which the unit's content will be get.
	     */
	    public void writeUnit(long unitNumber, byte[] b, int off)
		throws PrimitiveStorageException, IntermediateStorageException, IOException,
		       UnitNumberOutOfRangeException
	    {
		if (unitNumber>maxNumberOfUnits)
		    throw new UnitNumberOutOfRangeException();
		
		ism.write(chunkNumber, pageNumberOfUnit(unitNumber), 
			  (int)((unitNumber%numberOfUnitsPerUnitsPage)*unitSize), 
			  b, off, unitSize);
	    }
	    

	    
	    // + Addressing
	    
	    private long unitsPageNumberOfUnit(long unitNumber)
	    {
		return unitNumber/numberOfUnitsPerUnitsPage;
	    }
	    
	    private long extensionNumberOfUnitsPage(long unitsPageNumber)
	    {
		return unitsPageNumber/numberOfUnitsPagesPerExtension;
	    }
	    
	    private long areaNumberOfExtension(long extensionNumber)
	    {
		return extensionNumber/numberOfExtensionsPerArea;
	    }
	    
	    private long pageNumberOfUnitsPage(long unitsPageNumber)
	    {
		long extension=extensionNumberOfUnitsPage(unitsPageNumber);;
		long area=areaNumberOfExtension(extension);
		long page=unitsPageNumber+(extension+1)+(area+1)+1;
		return page;
	    }
	    
	    private long pageNumberOfUnit(long unitNumber)
	    {
		return pageNumberOfUnitsPage(unitNumber/numberOfUnitsPerUnitsPage);
	    }
	    
	    
	    // + Free space structure
	    
	    private long descriptionPageOfExtension(long extensionNumber)
	    {
		return extensionNumber*numberOfPagesPerExtension+(areaNumberOfExtension(extensionNumber)+1)+1;
	    }
	    
	    private long descriptionPageOfArea(long areaNumber)
	    {
		return areaNumber*numberOfPagesPerArea+1;
	    }
	    
	    private void setInBitMap(long pageNumber, long pos, boolean occupied)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		ism.writeBit(chunkNumber,pageNumber,pos,occupied);
	    }
	    
	    private boolean getOfBitMap(long pageNumber, long pos)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		return ism.readBit(chunkNumber,pageNumber,pos);
	    }
	    
	    private void setInUnitsBitMap(long pageNumber, int unitPos, boolean occupied)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		long absolutePos=(long)(pageSize-unitsBitMapSize)*8+unitPos;
		ism.writeBit(chunkNumber,pageNumber,absolutePos,occupied);
	    }
	    
	    private boolean getOfUnitsBitMap(long pageNumber, int unitPos)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		long absolutePos=(long)(pageSize-unitsBitMapSize)*8+unitPos;
		return ism.readBit(chunkNumber,pageNumber,absolutePos);
	    }


	    // + Free space management
	    
	    // - Space Assignation
	    
	    private boolean isAreaAssigned(long areaNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		if (areaNumber*numberOfExtensionsPerArea*numberOfUnitsPagesPerExtension>lastUnitsPageNumber)
		    return false;
		else
		    return true;
	    }
	    
	    private boolean isExtensionAssigned(long extensionNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		if (extensionNumber*numberOfUnitsPagesPerExtension>lastUnitsPageNumber)
		    return false;
		else
		    return true;
	    }

	    private boolean isUnitsPageAssigned(long unitsPageNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		if (unitsPageNumber>lastUnitsPageNumber)
		    return false;
		else
		    return true;
	    }
	    
	    private void assignArea(long areaNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		long desc=descriptionPageOfArea(areaNumber);
		for(int i=0; i<numberOfExtensionsPerArea; i++)
		    setInBitMap(desc,i,false);
	    }
	    
	    private void assignExtension(long extensionNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		long desc=descriptionPageOfExtension(extensionNumber);
		for(int i=0; i<numberOfUnitsPagesPerExtension; i++)
		    setInBitMap(desc,i,false);
	    }
	    
	    private void assignUnitsPage(long unitsPageNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		long pageNumber=pageNumberOfUnitsPage(unitsPageNumber);
		for(int i=0; i<numberOfUnitsPerUnitsPage; i++)
		    setInUnitsBitMap(pageNumber,i,false);
	    }
	    
	    
	    
	    // - Space Availability
	
	    private long findFirstAvailableExtensionInArea(long areaNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    { 
		if (isAreaAssigned(areaNumber))
		    {
			long desc=descriptionPageOfArea(areaNumber);
			for(long i=0; 
			    i<numberOfExtensionsPerArea && 
				areaNumber*numberOfExtensionsPerArea+i<maxNumberOfExtensions; 
			    i++)
			    if (!getOfBitMap(desc,i))
				return areaNumber*numberOfExtensionsPerArea+i;
			return -1;
		    }
		else
		    return areaNumber*numberOfExtensionsPerArea;
	    }
	    
	    private long findFirstAvailableUnitsPageInExtension(long extensionNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    { 
		if (isExtensionAssigned(extensionNumber))
		    {
			long desc=descriptionPageOfExtension(extensionNumber);
			for(long i=0; 
			    i<numberOfUnitsPagesPerExtension && 
				extensionNumber*numberOfUnitsPagesPerExtension+i<maxNumberOfUnitsPages; 
			    i++)
			    if (!getOfBitMap(desc,i))
				return extensionNumber*numberOfUnitsPagesPerExtension+i;
			return -1;
		    }
		else
		    return extensionNumber*numberOfUnitsPagesPerExtension;
	    }
	    
	    private long findFirstAvailableUnitInUnitsPage(long unitsPageNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    { 
		if (isUnitsPageAssigned(unitsPageNumber))
		    {
			if (numberOfUnitsPerUnitsPage > 1)
			    {
				long page=pageNumberOfUnitsPage(unitsPageNumber);
				for(int i=0; i<numberOfUnitsPerUnitsPage; i++)
				    if (!getOfUnitsBitMap(page,i))
					return unitsPageNumber*numberOfUnitsPerUnitsPage+i;
				return -1;
			    }
			else
			    {
				if (getOfBitMap(descriptionPageOfExtension(extensionNumberOfUnitsPage(unitsPageNumber)),
						unitsPageNumber%numberOfUnitsPagesPerExtension))
				    return -1;
				else
				    return unitsPageNumber;
			    }
		    }
		else
		    return unitsPageNumber*numberOfUnitsPerUnitsPage;
	    }

	    private boolean isAreaFull(long areaNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		return findFirstAvailableExtensionInArea(areaNumber)<0;
	    }
	    
	    private boolean isExtensionFull(long extensionNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		return findFirstAvailableUnitsPageInExtension(extensionNumber)<0;
	    }
	    
	    private boolean isUnitsPageFull(long unitsPageNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		return findFirstAvailableUnitInUnitsPage(unitsPageNumber)<0;
	    }
	    
	    
	    private boolean isAssignedUnitsPageEmpty(long unitsPageNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		if (numberOfUnitsPerUnitsPage > 1)
		    {
			boolean empty=true;
			long pageNumber=pageNumberOfUnitsPage(unitsPageNumber);
			for (int i=0; i<numberOfUnitsPerUnitsPage; i++)
			    if (getOfUnitsBitMap(pageNumber,i))
				empty=false;
			return empty;
		    }
		else
		    {
			return !getOfBitMap(descriptionPageOfExtension(extensionNumberOfUnitsPage(unitsPageNumber)),
					    unitsPageNumber%numberOfUnitsPagesPerExtension);
		    }
	    }
	    
	    private long findFirstAvailableArea()
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		long area;
		for (area=0; area<maxNumberOfAreas && isAreaFull(area); area++);
		if (area<maxNumberOfAreas)
		    return area;
		else
		    return -1;
	    }
	    
	    private void setAssignedUnitStatus(long unitNumber, boolean occupied)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		long unitsPageNumber=unitsPageNumberOfUnit(unitNumber);
		long extensionNumber=extensionNumberOfUnitsPage(unitsPageNumber);
		long areaNumber=areaNumberOfExtension(extensionNumber);
		if (numberOfUnitsPerUnitsPage>1)
		    {
			setInUnitsBitMap(pageNumberOfUnitsPage(unitsPageNumber),
					 (int)(unitNumber%numberOfUnitsPerUnitsPage),occupied);
			setInBitMap(descriptionPageOfExtension(extensionNumber),
				    unitsPageNumber%numberOfUnitsPagesPerExtension,
				    isUnitsPageFull(unitsPageNumber));
		    }
		else
		    setInBitMap(descriptionPageOfExtension(extensionNumber),
				unitsPageNumber%numberOfUnitsPagesPerExtension,
				occupied);
		setInBitMap(descriptionPageOfArea(areaNumber),
			    extensionNumber%numberOfExtensionsPerArea,
			    isExtensionFull(extensionNumber));
	    }


	    private boolean getAssignedUnitStatus(long unitNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		long unitsPageNumber=unitsPageNumberOfUnit(unitNumber);
		long extensionNumber=extensionNumberOfUnitsPage(unitsPageNumber);
		long areaNumber=areaNumberOfExtension(extensionNumber);
		if (!getOfBitMap(descriptionPageOfArea(areaNumber), extensionNumber%numberOfExtensionsPerArea))
		    {
			if (!getOfBitMap(descriptionPageOfExtension(extensionNumber),
					 unitsPageNumber%numberOfUnitsPagesPerExtension))
			    {
				if (numberOfUnitsPerUnitsPage>1)
				    return getOfUnitsBitMap(pageNumberOfUnitsPage(unitsPageNumber),
								(int)(unitNumber%numberOfUnitsPerUnitsPage));
				else
				    return false;
			    }
			else
			    return true;
		    }
		else
		    return true;
	    }

	
	    /**
	     * This function finds the first available unit in the space.
	     * @return The number of the first free unit, 
	     * or <code>-1</code> if there are not free units.
	     */
	    public long findFirstFreeUnit()
		throws PrimitiveStorageException, IntermediateStorageException, IOException,
		       InconsistentStorageStructureException
	    {
		long area=findFirstAvailableArea();
		if (area<0)
		    return -1;   
		long extension=findFirstAvailableExtensionInArea(area);
		if (extension<0)
		    throw new InconsistentStorageStructureException();
		long unitsPage=findFirstAvailableUnitsPageInExtension(extension);
		if (unitsPage<0)
		    throw new InconsistentStorageStructureException();
		long unit=findFirstAvailableUnitInUnitsPage(unitsPage);
		if (unit<0)
		    throw new InconsistentStorageStructureException();
		return unit;
	    }
	
	    /**
	     * Find a free unit in the same page of a given unit.
	     * @param unitNumber The number of the unit that has to be in the same page 
	     *                   that the unit to be found.
	     * @return The number of the found free unit, 
	     *         or <code>-1</code> if there are not free units 
	     *         in the same page that <code>unitNumber</code>.
	     */
	    public long findFreeUnitCloseOf(long unitNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException,
		       UnitNumberOutOfRangeException
	    {
		if (unitNumber>maxNumberOfUnits)
		    throw new UnitNumberOutOfRangeException();
		return findFirstAvailableUnitInUnitsPage(unitsPageNumberOfUnit(unitNumber));
	    }

	    /**
	     * Sets a specific unit as occupied.
	     * @param unitNumber The number of the unit to occup. 
	     */
	    public void occupUnit(long unitNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException,
		       UnitNumberOutOfRangeException
	    {
		if (unitNumber>maxNumberOfUnits)
		    throw new UnitNumberOutOfRangeException();
		long unitsPageNumber=unitsPageNumberOfUnit(unitNumber);
		if (unitsPageNumber>lastUnitsPageNumber)
		    {
			for (long unitsPage=lastUnitsPageNumber+1; unitsPage<=unitsPageNumber; unitsPage++)
			    {
				long extension=extensionNumberOfUnitsPage(unitsPage);
				long area=areaNumberOfExtension(extension);
				if (extension%numberOfExtensionsPerArea==0)
				    assignArea(area);
				if (unitsPage%numberOfUnitsPagesPerExtension==0)
				    assignExtension(extension);
				if (numberOfUnitsPerUnitsPage>1)
				    assignUnitsPage(unitsPage);
			    }
			writeLastUnitsPageNumberToSecundaryStorage(unitsPageNumber);
		    }	    
		setAssignedUnitStatus(unitNumber,true);
	    }
	
	    /**
	     * Sets a specific unit as free.
	     * @param unitNumber The number of the unit to free. 
	     */
	    public void freeUnit(long unitNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException,
		       UnitNumberOutOfRangeException
	    {
		if (unitNumber>maxNumberOfUnits)
		    throw new UnitNumberOutOfRangeException();
		long unitsPageNumber=unitsPageNumberOfUnit(unitNumber);
		if (unitsPageNumber<=lastUnitsPageNumber)
		    setAssignedUnitStatus(unitNumber,false);
	    }

	    /**
	     * Checks if a unit is occupied.
	     * @param unitNumber The number of the unit.
	     * @return <code>true</code> if the specified unit is occupied, <code>false</code> otherwise.
	     */
	    public boolean isOccupiedUnit(long unitNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException,
		       UnitNumberOutOfRangeException
	    {
		if (unitNumber>maxNumberOfUnits)
		    throw new UnitNumberOutOfRangeException();
		long unitsPageNumber=unitsPageNumberOfUnit(unitNumber);
		if (unitsPageNumber<=lastUnitsPageNumber)
		    return getAssignedUnitStatus(unitNumber);
		else
		    return false;
	    }

	    /**
	     * Checks if a unit is free.
	     * @param unitNumber The number of the unit.
	     * @return <code>true</code> if the specified unit is free, <code>false</code> otherwise.
	     */
	    public boolean isFreeUnit(long unitNumber)
		throws PrimitiveStorageException, IntermediateStorageException, IOException,
		       UnitNumberOutOfRangeException
	    {
		return !isOccupiedUnit(unitNumber);
	    }

	    
	    /**
	     * Reduce the logic assigned space, later the physical space can be reduced. 
	     */
	    public void reduce()
		throws PrimitiveStorageException, IntermediateStorageException, IOException
	    {
		while (lastUnitsPageNumber>-1 && isAssignedUnitsPageEmpty(lastUnitsPageNumber))
		    writeLastUnitsPageNumberToSecundaryStorage(lastUnitsPageNumber-1);
	    }
	    
	}

	TreeMap spacesTable;
	
	Segment(int unitSize)
	    throws InvalidUnitSizeException
	{
	    if (unitSize<=0 || unitSize>pageSize)
		    throw new InvalidUnitSizeException();
	    this.unitSize=unitSize;
	    spacesTable=new TreeMap(spaceNumberComparator);
	}

	/**
	 * Load a space with a specified space number.
	 * @param spaceNumber The number which the space will be identified inside the segment.
	 * @param chunkNumber The number of the chunk to be used for mounting the space.
	 * @param overwrite If it is <code>false</code> the previous chunk's content will not be changed, 
	 *                  if it is <code>true</code> the space will be formated.
	 */
	public void loadSpace(int spaceNumber, int chunkNumber, boolean overwrite)
	    throws InvalidSpaceNumberException, AlreadyExistingSpaceNumberException,
		   PrimitiveStorageException, IntermediateStorageException, IOException
	{
	    if (spaceNumber<0 || spaceNumber>=MAX_SPACE_NUMBER)
		throw new InvalidSpaceNumberException();
	    auxSpaceNumber.setValue(spaceNumber);
	    if (spacesTable.containsKey(auxSpaceNumber))
		throw new AlreadyExistingSpaceNumberException();
	    else
		spacesTable.put(new SpaceNumber(spaceNumber), new Space(chunkNumber,overwrite));
	}

	/**
	 * Removes the space with the specified number inside the segment.
	 * @param spaceNumber The number of the space to be removed.
	 */
	public void removeSpace(int spaceNumber)
	    throws NotExistingSpaceNumberException
	{
	    auxSpaceNumber.setValue(spaceNumber);
	    if (!spacesTable.containsKey(auxSpaceNumber))
		throw new NotExistingSpaceNumberException();
	    else
		spacesTable.remove(auxSpaceNumber);
	}
	
	/**
	 * Returns the space with the specified number.
	 * @param spaceNumber The number of the space to be retrieved.
	 * @return The <code>Space</code> object requested.
	 */
	public Space getSpace(int spaceNumber)
	{
	    auxSpaceNumber.setValue(spaceNumber);
	    if (spacesTable.containsKey(auxSpaceNumber))
		return (Space)(spacesTable.get(auxSpaceNumber));
	    else
		return null;
	}

	/**
	 * This class is used to retrieve the used space numbers inside the segment. 
	 * Space numbers are retrived in a sequence.
	 */
	public class SpaceNumberIterator
	{
	    private Iterator i;

	    SpaceNumberIterator()
	    {
		i=spacesTable.keySet().iterator();
	    }
	    
	    /**
	     * Indicates if there are more space numbers in the sequence.
	     */
	    public boolean hasNext()
	    {
		return i.hasNext();
	    }

	    /**
	     * Returns the next space number in the sequence.
	     */
	    public int next()
	    {
		return ((SpaceNumber)(i.next())).intValue();
	    }
	}

	/**
	 * Returns the sequence of used space numbers by the segment.
	 */
	public SpaceNumberIterator getSpaceNumbers()
	{
	    return new SpaceNumberIterator();
	}

	/**
	 * Returns the size of the units in the segmet. 
	 */
	public int getUnitSize()
	{
	    return unitSize;
	}
    } 
    

    private Segment[] segmentsTable;

    /** 
     * Constructs an advanced storage manager from a intrmedite storage manager and 
     * an array that specified the number of segments and ther unit sizes.
     */
    public AdvancedStorageManager(IntermediateStorageManager ism, int[] segmentsUnitSizes)
	throws InvalidUnitSizeException
    {
	this.ism=ism;
	this.pageSize=getPrimitiveStorageManager().getPageSize();
	this.segmentsTable=new Segment[segmentsUnitSizes.length];
	for (int i=0; i<segmentsUnitSizes.length; i++)
	    segmentsTable[i]=new Segment(segmentsUnitSizes[i]);
	auxSpaceNumber=new SpaceNumber(0);
    }

    /**
     * Returns the segment identified with the specified number.
     * @param segmentNumber The number of the segment.
     * @return The segment identified with <code>segmentNumber</code>
     */
    public Segment getSegment(int segmentNumber)
    {
	return segmentsTable[segmentNumber];
    }

    /**
     * Returns the number of existing segments.
     */
    public int getNumberOfSegments()
    {
	return segmentsTable.length;
    }

    /**
     * Returns the underlying primitive storage manager.
     */
    public PrimitiveStorageManager getPrimitiveStorageManager()
    {
	return ism.getPrimitiveStorageManager();
    }

    /**
     * Returns the underlying intermediate storage manager.
     */
    public IntermediateStorageManager getIntermediateStorageManager()
    {
	return ism;
    }

    /**
     * Synchronizes with the secundary storage.
     */
    public void synchronize()
	throws IOException
    {
	ism.synchronizeAndTransfer(); 
    } 
}
