package storage.intermediateStorage;

import storage.primitiveStorage.*;
import java.io.*;

/**
 * This class is used to manage the operatios of the intermediate storage component. 
 * The intermediate storage component controls operations of higer level 
 * than primitive storage, partial writing or reading from pages
 * is provided. Intermediate storage manager also uses a little buffer
 * for improving the massive I/O instructions over the same page.
 */
public final class IntermediateStorageManager
{
    private PrimitiveStorageManager psm;
    private int pageSize;

    private byte[] lastPage;
    private int lastChunkNumber;
    private long lastPageNumber;
    private boolean lastUpdated;
    private boolean existLast;

    /** 
     * Constructs an intermediate storage manager from a primitive storage manager.
     */
    public IntermediateStorageManager(PrimitiveStorageManager psm)
    {
	this.psm=psm;
	this.pageSize=psm.getPageSize();
	this.lastPage=new byte[this.pageSize];
	lastChunkNumber=-1;
	lastPageNumber=-1;
	lastUpdated=false;
	existLast=false;
    }    

    /**
     * Returns the underlying primitive storage manager.
     */
    public PrimitiveStorageManager getPrimitiveStorageManager()
    {
	return psm;
    }

    private void readInLastPage(int chunkNumber, long pageNumber)
	throws PrimitiveStorageException, IOException
    {
	if (!(existLast && chunkNumber==lastChunkNumber && pageNumber==lastPageNumber))
	    {
		if (existLast)
		    {
			if (lastUpdated)
			    psm.writePage(lastChunkNumber,lastPageNumber,lastPage);
		    }
		else
		    existLast=true;
		psm.readPage(chunkNumber,pageNumber,lastPage);
		lastChunkNumber=chunkNumber;
		lastPageNumber=pageNumber;
		lastUpdated=false;
	    }
    }

    /**
     * Syncronizes the local buffer with the underlying primitive storage manager.
     */
    public void synchronize()
	throws IOException
    {
	if (existLast && lastUpdated)
	    {
		try {
		    psm.writePage(lastChunkNumber,lastPageNumber,lastPage);
		} catch (NotExistingChunkNumberException e1) {
		} catch (PageNumberOutOfRangeException e2) {}
	    }
    }

    /**
     * Syncronizes itself and the underlying primitive storage manager with secundary storage.
     */
    public void synchronizeAndTransfer()
	throws IOException
    {
	synchronize();
	psm.transferBuffer();
    }

    /**
     * Syncronizes itself and the underlying primitive storage manager with secundary storage,
     * also crear the buffer.
     */
    public void synchronizeTransferAndClear()
	throws IOException
    {
	synchronize();
	psm.transferAndClearBuffer();
    }

    /**
     * Reads partially a page starting from an offset and reading a specific length of bytes.
     * @param chunkNumber The number of the chunk.
     * @param pageNumber The number of the page inside the chunk.
     * @param pageOffset The position of the page from which the reading starts.
     * @param b The byte array which the content of the readed page will be stored.
     * @param off The offset of <code>b</code> from which the readed content will be stored.
     * @param length The number of bytes to be readed.
     */
    public void read(int chunkNumber, long pageNumber, int pageOffset, 
		     byte[] b, int off, int length)
	throws PrimitiveStorageException,
	       InvalidPageOffsetException, OutOfPageException,  
	       IOException
    {
	if (pageOffset<0 || pageOffset>=pageSize)
	    throw new InvalidPageOffsetException();
	if (pageOffset+length>pageSize)
	    throw new OutOfPageException();
	readInLastPage(chunkNumber,pageNumber);
	for (int i=0; i<length; i++)
	    b[i+off]=lastPage[pageOffset+i];
	
    }
    
    /**
     * Writes partially a page starting from an offset and writing a specific length of bytes.
     * @param chunkNumber The number of the chunk.
     * @param pageNumber The number of the page inside the chunk.
     * @param pageOffset The position of the page from which the writing starts.
     * @param b The byte array from which the page's content will be get.
     * @param off The offset of <code>b</code> from which the page's content will be get.
     * @param length The number of bytes to be wrote.
     */
    public void write(int chunkNumber, long pageNumber, int pageOffset, 
		     byte[] b, int off, int length)
	throws PrimitiveStorageException, 
	       InvalidPageOffsetException, OutOfPageException,  
	       IOException
    {
	if (pageOffset<0 || pageOffset>=pageSize)
	    throw new InvalidPageOffsetException();
	if (pageOffset+length>pageSize)
	    throw new OutOfPageException();
	readInLastPage(chunkNumber,pageNumber);
	for (int i=0; i<length; i++)
	    lastPage[pageOffset+i]=b[i+off];
	lastUpdated=true;
    }

    /**
     * Reads fully a page.
     * @param chunkNumber The number of the chunk.
     * @param pageNumber The number of the page inside the chunk.
     * @param b The byte array which the content of the readed page will be stored.
     * @param off The offset of <code>b</code> from which the readed content will be stored.
     */
    public void readPage(int chunkNumber, long pageNumber, byte[] b, int off)
	throws PrimitiveStorageException, IOException
    {
	readInLastPage(chunkNumber,pageNumber);
	for (int i=0; i<pageSize; i++)
	    b[off+i]=lastPage[i];
    }

    /**
     * Writes partially a page starting from an offset and writing a specific length of bytes.
     * @param chunkNumber The number of the chunk.
     * @param pageNumber The number of the page inside the chunk.
     * @param b The byte array from which the page's content will be get.
     * @param off The offset of <code>b</code> from which the page's content will be get.
     */
    public void writePage(int chunkNumber, long pageNumber, byte[] b, int off)
	throws PrimitiveStorageException, IOException
    {
	readInLastPage(chunkNumber,pageNumber);
	for (int i=0; i<pageSize; i++)
	    lastPage[i]=b[i];
	lastUpdated=true;
    }

    /**
     * Reads a specific bit from a page.
     * @param chunkNumber The number of the chunk.
     * @param pageNumber The number of the page inside the chunk.
     * @param bitNumber The position of the bit inside the page. 
     *                  There are 8*pageSize available positions.
     * @return <code>true</code> If the bit value is <code>1</code>,
     *         <code>false</code> If the bit value is <code>0</code>.
     */
    public boolean readBit(int chunkNumber, long pageNumber, long bitNumber)
	throws PrimitiveStorageException, OutOfPageException, IOException
    {
	if (bitNumber<0 || bitNumber>=(long)pageSize*8)
	    throw new OutOfPageException();
	readInLastPage(chunkNumber,pageNumber);
	int bytePos=(int)(bitNumber/8);
	int bitPos=(int)(bitNumber%8);
	if ((lastPage[bytePos] & (byte)(0x1<<bitPos))==0)
	    return false;
	else
	    return true;
    }

    /**
     * Writes a specific bit to a page.
     * @param chunkNumber The number of the chunk.
     * @param pageNumber The number of the page inside the chunk.
     * @param bitNumber The position of the bit inside the page. 
     *                  There are 8*pageSize available positions.
     * @param value If value is <code>true</code> writes <code>1</code>,
     *              if it is <code>false</code> writes <code>0</code>.
     */
    public void writeBit(int chunkNumber, long pageNumber, long bitNumber, boolean value)
	throws PrimitiveStorageException, OutOfPageException, IOException
    {
	if (bitNumber<0 || bitNumber>=(long)pageSize*8)
	    throw new OutOfPageException();
	readInLastPage(chunkNumber,pageNumber);
	int bytePos=(int)(bitNumber/8);
	int bitPos=(int)(bitNumber%8);
	if (value)
	    lastPage[bytePos] |= (byte)(0x1<<bitPos);
	else
	    lastPage[bytePos] &= ~(byte)(0x1<<bitPos);
	lastUpdated=true;
    }

    /**
     * Reads a java 8-byte <code>long</code> from a page.
     * @param chunkNumber The number of the chunk.
     * @param pageNumber The number of the page inside the chunk.
     * @param pageOffset The position of the page from which the reading starts.
     * @return The readed 8 bytes as a <code>long</code>.
     */
    public long readLong(int chunkNumber, long pageNumber, int pageOffset)
	throws PrimitiveStorageException, 
	       InvalidPageOffsetException, OutOfPageException,  
	       IOException
    {
	if (pageOffset<0 || pageOffset>=pageSize)
	    throw new InvalidPageOffsetException();
	if (pageOffset+8>pageSize)
	    throw new OutOfPageException();
	readInLastPage(chunkNumber,pageNumber);
	long l=0x0l;
	for (int i=0; i<8; i++)
	    l |= (lastPage[pageOffset+i] << (8*i));
	return l;
    }

    /**
     * Write a java 8-byte <code>long</code> to a page.
     * @param chunkNumber The number of the chunk.
     * @param pageNumber The number of the page inside the chunk.
     * @param pageOffset The position of the page from which the writing starts.
     * @param number The readed 8 bytes as a <code>long</code> to be wrote in the page.
     */
    public void writeLong(int chunkNumber, long pageNumber, int pageOffset, long number)
	throws PrimitiveStorageException, 
	       InvalidPageOffsetException, OutOfPageException,  
	       IOException
    {
	if (pageOffset<0 || pageOffset>=pageSize)
	    throw new InvalidPageOffsetException();
	if (pageOffset+8>pageSize)
	    throw new OutOfPageException();
	readInLastPage(chunkNumber,pageNumber);
	for (int i=0; i<8; i++)
	    lastPage[pageOffset+i] = (byte)(0xFFl & (number >> (8*i)));
	lastUpdated=true;
    }
}











