package storage.primitiveStorage;

import java.io.*;
import java.util.*;

/**
 * This class is used to manage the operatios of a primitive storage component. 
 * The primitive storage component controls the access to secundary storage manager
 * (using chunks) and the intermediate memory (using a buffer). 
 * In the primitive storage component the page is the minimum storage unit.
 */
public final class PrimitiveStorageManager
{
    /**
     * The maximum size of a chunk.
     */
    static public final long MAX_CHUNK_SIZE=0x7FFFFFFFFFFFFFFFl;

    /**
     * The manimum size of a page.
     */
    static public final int MIN_PAGE_SIZE=8;

    /**
     * The maximum number which a chunk can be associated.
     */
    static public final long MAX_CHUNK_NUMBER=0x7FFFFFFF;
    
    /**
     * The default page size if other is not specified.
     */
    static public final int DEFAULT_PAGE_SIZE=4096;  

    private int pageSize;

    /**
     * A chunks is a part of secundary storage. 
     * Operative systems organize the secundary storage using files,
     * so, a chunk simply is a file or a part of a file. 
     * The primitive storage component can use
     * several chunks to organize its information. Chunks can help to overcome
     * some storage difficulties like the size file limits provide for some
     * operative systems, also chunks provides the use of several store devices for 
     * keeping a commun information. In database systems chunks are used to storage
     * huge databases. 
     */
    abstract public class Chunk
    {
	long maxSize;
	long offset;
	RandomAccessFile file;
	File fileName;

	Chunk()
	{
	}

	void init(File fileName, long offset, long maxSize)
	    throws FileNotFoundException
	{
	    this.offset=offset;
	    this.maxSize = maxSize - maxSize%pageSize;
	    this.fileName=fileName;
	    this.file=new RandomAccessFile(fileName,"rws");
	}

	/** 
	 * Reads a page from the chunk and copies the content in the specified byte array.
	 * @param pageNumber The number of the page to be readed.
	 * @param b The byte array when the page's content will be copied. 
	 */
	public void readPage(long pageNumber, byte[] b)
	    throws PageNumberOutOfRangeException, IOException
	{
	    if (pageNumber<0 || (pageNumber+1)*pageSize>maxSize)
		throw new PageNumberOutOfRangeException();
	    if (offset+(pageNumber+1)*pageSize>file.length())
		file.setLength(offset+(pageNumber+1)*pageSize);
	    file.seek(offset+pageNumber*pageSize);
	    file.readFully(b,0,pageSize);
	}

	/** 
	 * Writes a page in the chunk and taking the content from a specified byte array.
	 * @param pageNumber The number of the page to be wrote.
	 * @param b The byte array from which the page takes its content. 
	 */
	public void writePage(long pageNumber, byte[] b) 
	    throws PageNumberOutOfRangeException, IOException
	{
	    if (pageNumber<0 || (pageNumber+1)*pageSize>maxSize)
		throw new PageNumberOutOfRangeException();
	    file.seek(offset+pageNumber*pageSize);
	    file.write(b,0,pageSize);
	    file.getFD().sync();
	}

	/**
	 * Returns the maximum size (in bytes) of the chunk.
	 */
	public long getMaxSize()
	{
	    return maxSize;
	}

	/**
	 * Returns the file name used by the chunk.
	 */
	public String getFileName()
	{
	    return fileName.getAbsolutePath();
	}

	/**
	 * Returns the java <code>RandomAccessFile</code> used by the chunk.
	 */
	public RandomAccessFile getFile()
	{
	    return file;
	}

	/**
	 * Returns a string contains the chunk information.
	 */
	abstract public String getInfo() throws IOException;
    }


    /**
     * Flexible chunks are used to storage information in files which size can 
     * grow or decrease. This kind of chunks are used for storage information in
     * files with a high control of the operating system, in database terms this
     * represent the storage in cooked partitions.
     */
    public final class FlexibleChunk extends Chunk
    {
	FlexibleChunk(File fileName, long maxSize, boolean overwrite) 
	    throws InvalidChunkSizeException, FileNotFoundException, IOException
	{
	    if (maxSize<0 || maxSize>MAX_CHUNK_SIZE)
		throw new InvalidChunkSizeException();
	    init(fileName,0,maxSize);
	    if (overwrite)
		this.file.setLength(0);
	}

	FlexibleChunk(File fileName, boolean overwrite) 
	    throws FileNotFoundException, IOException
	{
	    init(fileName,0,MAX_CHUNK_SIZE);
	    if (overwrite)
		this.file.setLength(0);
	}

	FlexibleChunk(File fileName) 
	    throws FileNotFoundException
	{
	    init(fileName,0,MAX_CHUNK_SIZE);
	}

	/**
	 * Returns the real size (in bytes) of the chunk, i.e. the size of file used by the chunk.
	 */
	public long getSize() throws IOException
	{
	    return file.length();
	}

	/**
	 * Changes the size of the file used by the chunk.
	 * @param size The new size of the chunk, this size will be round according to page size.
	 */
	public void changeSize(long size) throws IOException
	{
	    long s=size-size%pageSize;
	    if (s<file.length())
		file.setLength(s);
	}

	public String getInfo() throws IOException
	{
	    String r="File:\""+getFileName()+"\" Size:"+getSize()+" Max-Size:"+getMaxSize()+"\n";
	   
	    file.seek(0);
	    for (long i=0; i<file.length(); i++)
		{
		    int b=0xFF&file.read();
		    r += (b<16?"0":"")+Integer.toHexString(b);
		    if (i%pageSize == pageSize-1)
			r+="\n";
		    else
			r+=" ";
		}
	    r+="\n";
	    return r;
	}
    }
    
    /**
     * Fixed chunks are used to storage information in files which size is considered
     * fixed. This kind of chunks are used for storage information in
     * files with a low control of the operating system, in database terms this
     * represent the storage in raw partitions. A fixed chunk has an offset that 
     * indicates where the chunks starts in the file, and also has a maximun size
     * that indicates what part of the file it occups.
     */
    public final class FixedChunk extends Chunk
    {	
	FixedChunk(File fileName,  long offset, long maxSize) 
	    throws InvalidChunkSizeException, InvalidChunkOffsetException,
		   UnaddressableChunkException, FileNotFoundException
	{
	    if (maxSize<0 || maxSize>MAX_CHUNK_SIZE)
		throw new InvalidChunkSizeException();
	    if (offset<0)
		throw new InvalidChunkOffsetException();
	    if (0x7FFFFFFFFFFFFFFFl-offset<maxSize)
		throw new UnaddressableChunkException();
	    init(fileName,offset,maxSize);
	}

	/**
	 * Returns the file offset when the chunk starts.
	 */
	public long getOffset()
	{
	    return offset;
	}

	public String getInfo() throws IOException
	{
	    String r="File:\""+getFileName()+"\" Offset:"+getOffset()+" Max-Size:"+getMaxSize()+"\n";
	    
	    file.seek(offset);
	    for (long i=0; i<getMaxSize(); i++)
		{
		    int b=0xFF&file.read();
		    r += (b<16?"0":"")+Integer.toHexString(b);
		    if (i%pageSize == pageSize-1)
			r+="\n";
		    else
			r+=" ";
		}
	    r+="\n";
	    return r;
	}
    }
    
    
    /** 
     * This class represents the chunk's table, i.e. a set
     * of number-chunk pairs.
     * This class uses a AA-Tree for implementing the table.
     */
    public class ChunksTable
    {
	/**
	 * Represents a number-chunk pair.
	 */
	public class Entry
	{
	    int chunkNumber;
	    Chunk chunk;
	    Entry left, right, parent, next, prev;
	    int level;
	    
	    Entry(int chunkNumber, Chunk chunk)
	    {
		this.chunkNumber=chunkNumber;
		this.chunk=chunk;
		left=right=parent=null;
		next=prev=null;
		level=1;
	    }
	    
	    /**
	     * Returns the chunk number.
	     */
	    public int getChunkNumber()
	    {
		return chunkNumber;
	    }
	    
	    /**
	     * Returns the chunk object.
	     */
	    public Chunk getChunk()
	    {
		return chunk;
	    }
	    
	    /**
	     * Returns the next number-chunk pair of the table in the
	     * insertion order.
	     */
	    public Entry getNext()
	    {
		return next;
	    }
	    
	    /**
	     * Returns the previous number-chunk pair of the table in the
	     * insertion order.
	     */
	    public Entry getPrevious()
	    {
		return prev;
	    }
	}
	
	private Entry root, head, tail;
	
	/**
	 * Constructs an empty chunk's table.
	 */
	public ChunksTable()
	{
	    root=head=tail=null;
	} 
	
	/**
	 * Checks if a chunk with a given number exists in the table.
	 * @param chunkNumber The chunk number to find in the table.
	 * @return <code>true</code> if a chunk with the given number is
	 * in the table, <code>false</code> otherwise.
	 */
	public boolean containsChunkNumber(int chunkNumber)
	{
	    Entry n=findNearestInTree(chunkNumber);
	    if (n!=null)
		return n.chunkNumber==chunkNumber;
	    else
		return false;
	}

	/**
	 * Returns the chunk with the specified number.
	 * @param chunkNumber The chunk number to find in the table.
	 * @return The chunk with the specified number, 
	 * <code>null</code> if the number does not exist.
	 */
	public Chunk get(int chunkNumber)
	{
	    Entry n=findNearestInTree(chunkNumber);
	    if (n!=null && n.chunkNumber==chunkNumber)
		return n.chunk;
	    else
		return null;
	}
	
	/**
	 * Returns the number-chunk pair with the specified number.
	 * @param chunkNumber The chunk number to find in the table.
	 * @return The number-chunk pair with the specified number, 
	 * <code>null</code> if the number does not exist.
	 */
	public Entry getEntry(int chunkNumber)
	{
	    Entry n=findNearestInTree(chunkNumber);
	    if (n!=null && n.chunkNumber==chunkNumber)
		return n;
	    else
		return null;
	}

	/**
	 * Puts a chunk with a specified number in the table.
	 * @param chunkNumber The chunk number to put.
	 * @param chunk The chunk to put.
	 */
	public void put(int chunkNumber, Chunk chunk)
	{
	    if (root==null)
		root=head=tail=new Entry(chunkNumber,chunk);
	    else
		{
		    Entry n=findNearestInTree(chunkNumber);
		    if (n.chunkNumber!=chunkNumber)
			{
			    Entry m=new Entry(chunkNumber,chunk);
			    if (chunkNumber<n.chunkNumber)
				n.left=m;
			    else
				n.right=m;
			    m.parent=n;
			    adjustTreeAfterInsertion(n);
			    insertInList(m);
			}
		    else
			n.chunk=chunk;
		}
	}
	
	/**
	 * Removes a chunk with a specified number from the table.
	 * @param chunkNumber The number of the chunk to remove.
	 */
	public void remove(int chunkNumber)
	{
	    Entry n=findNearestInTree(chunkNumber);
	    if (n!=null && n.chunkNumber==chunkNumber)
		{
		    removeFromTree(n);
		    removeFromList(n);
		}
	}
	
	/**
	 * Returns the first inserted number-chunk pair.
	 * @return The first inserted number-chunk pair, 
	 * <code>null</code> if there are not inserted chunks. 
	 */
	public Entry getFirst()
	{
	    return head;
	}
	
	/**
	 * Returns the last inserted number-chunk pair.
	 * @return The last inserted number-chunk pair, 
	 * <code>null</code> if there are not inserted chunks. 
	 */
	public Entry getLast()
	{
	    return tail;
	}

	private Entry findNearestInTree(int chunkNumber)
	{
	    Entry current=null;
	    Entry target=root;
	    while (target!=null)
		{
		    current=target;
		    if (chunkNumber<current.chunkNumber)
			target=current.left;
		    else if (chunkNumber>current.chunkNumber)
			target=current.right;
		    else
			return current;
		}
	    return current;
	}
	
	private Entry rotateWithLeftChild(Entry n)
	{
	    Entry m = n.left;
	    n.left = m.right;
	    if (m.right!=null)
		m.right.parent=n;
	    m.right = n;
	    if (n.parent==null)
		root=m;
	    else if (n.parent.left==n)
		n.parent.left=m;
	    else
		n.parent.right=m;
	    m.parent=n.parent;
	    n.parent=m;
	    return m;
	}
	
	private Entry rotateWithRightChild(Entry n)
	{
	    Entry m = n.right;
	    n.right = m.left;
	    if (m.left!=null)
		m.left.parent=n;
	    m.left = n;
	    if (n.parent==null)
		root=m;
	    else if (n.parent.left==n)
		n.parent.left=m;
	    else
		n.parent.right=m;
	    m.parent=n.parent;
	    n.parent=m;
	    return m;
	}
	
	private Entry skew(Entry t)
	{
	    if (t.left!=null && t.left.level==t.level)
		t=rotateWithLeftChild(t);
	    return t;
	}
	
	private Entry split(Entry t)
	{
	    if(t.right!=null && t.right.right!=null && t.right.right.level==t.level)
		{
		    t=rotateWithRightChild(t);
		    t.level++;
		}
	    return t;
	}
	
	
	private void adjustTreeAfterInsertion(Entry n)
	{
	    while (n!=null)
		{
		    n=skew(n);
		    n=split(n);
		    n=n.parent;
		}
	}
	
	private void reinsertEntryInTree(Entry isolated, Entry deleted)
	{
	    if (isolated!=deleted)
		{
		    isolated.level=deleted.level;
		    isolated.left=deleted.left;
		    isolated.right=deleted.right;
		    isolated.parent=deleted.parent;
		    
		    if (deleted.parent==null)
			root=isolated;
		    else if (deleted.parent.left==deleted)
			deleted.parent.left=isolated;
		    else
			deleted.parent.right=isolated;
		    
		    if (deleted.left!=null)
			deleted.left.parent=isolated;
		    if (deleted.right!=null)
			deleted.right.parent=isolated;
		}
	    deleted.level=0;
	    deleted.left=deleted.right=deleted.parent=null;
	}
	
	private void removeFromTree(Entry deleted)
	{
	    Entry victim;
	    Entry adjusted;
	    
	    victim=deleted;
	    Entry target=victim.right;
	    while (target!=null)
		{
		    victim=target;
		    target=victim.left;
		}
	    
	    if (victim.parent==deleted)
		adjusted=victim;
	    else
		adjusted=victim.parent;
	    
	    if (victim.parent==null)
		root=null;
	    else if (victim.parent.left==victim)
		{
		    if (victim.right!=null)
		       victim.right.parent=victim.parent;
		    victim.parent.left=victim.right;
		}
	    else
		{
		    if (victim.right!=null)
			victim.right.parent=victim.parent;
		    victim.parent.right=victim.right;
		}
	    reinsertEntryInTree(victim,deleted);
	    
	    while (adjusted!=null)
		{
		    if (((adjusted.left!=null && adjusted.left.level<adjusted.level-1) || 
			 (adjusted.left==null && 0<adjusted.level-1)) || 
			((adjusted.right!=null && adjusted.right.level<adjusted.level-1) || 
			 (adjusted.right==null && 0<adjusted.level-1)))
			{
			    adjusted.level--;
			    if (adjusted.right!=null && adjusted.right.level>adjusted.level)
				adjusted.right.level=adjusted.level;
			    adjusted=skew(adjusted);
			    if (adjusted.right!=null)
				{
				    adjusted.right=skew(adjusted.right);
				    if (adjusted.right.right!=null)
					adjusted.right.right=skew(adjusted.right.right);
				}
			    adjusted=split(adjusted);
			    if (adjusted.right!=null)
				adjusted.right=split(adjusted.right);
			}
		    adjusted=adjusted.parent;
		}
	}
	
	private void insertInList(Entry n)
	{
	    n.next=null;
	    n.prev=tail;
	    if (tail!=null)
		tail.next=n;
	    else
		head=n;
	    tail=n;
	}
	
	private void removeFromList(Entry n)
	{
	    if (n==head)
		head=n.next;
	    if (n==tail)
		tail=n.prev;
	    if (n.prev!=null)
		n.prev.next=n.next;
	    if (n.next!=null)
		n.next.prev=n.prev;
	    n.next=n.prev=null;
	}
    }

    private ChunksTable chunksTable;
    
    /* This class is used to manage input/output operatios to secundary storage, specifically 
       to de chunks. */
    private final class SecundaryDeviceManager
    {
	public void readPage(int chunkNumber, long pageNumber, byte[] b) 
	    throws NotExistingChunkNumberException, PageNumberOutOfRangeException, IOException
	{
	    if (!chunksTable.containsChunkNumber(chunkNumber))
		throw new NotExistingChunkNumberException();
	    else
		{
		    Chunk c=chunksTable.get(chunkNumber);
		    c.readPage(pageNumber,b);
		}
	}
	
	
	public void writePage(int chunkNumber, long pageNumber, byte[] b) 
	    throws NotExistingChunkNumberException, PageNumberOutOfRangeException, IOException
	{  
	    if (!chunksTable.containsChunkNumber(chunkNumber))
		throw new NotExistingChunkNumberException();
	    else
		{
		    Chunk c=chunksTable.get(chunkNumber);
		    c.writePage(pageNumber,b);
		}
	}
    }


    private SecundaryDeviceManager sdm;


    /** Represents the LRU (Least Recent Used) replace algorithm. */
    static public final int LRU_ALGORITHM=0;

    /** Represents the Clock or LRU with second chance replace algorithm. */
    static public final int CLOCK_ALGORITHM=1;

    /** Represents the FIFO (First In First Out) replace algorithm. */
    static public final int FIFO_ALGORITHM=2;

    /** Represents the MRU (Most Recent Used) replace algorithm. */
    static public final int MRU_ALGORITHM=3;

    /** Represents the LIFO (Last In First Out) replace algorithm. */
    static public final int LIFO_ALGORITHM=4;
    

    /*
     * This class represent the intermediate memory device, i.e. The buffer,
     * the buffer keeps pages according to a specific replace algorithm to 
     * increase the I/O proficiency. The buffer is implemented using a hash table
     * and a circular list.
     */
    private final class PageBuffer
    {
	// Stores the content of a page and information about if it has been updated or not.
	private final class PageDescriptor
	{
	    public byte[] page;
	    public boolean updated;

	    public PageDescriptor(byte[] page, boolean updated)
	    {
		this.page=new byte[pageSize];
		for (int i=0; i<pageSize; i++)
		    this.page[i]=page[i];
		this.updated=updated;
	    }

	    public void update(byte[] page)
	    {
		for (int i=0; i<pageSize; i++)
		    this.page[i]=page[i];
	    }

	    public void update(byte[] page, boolean updated)
	    {
		update(page);
		this.updated=updated;
	    }

	    public void copyPage(byte[] b)
	    {
		for (int i=0; i<pageSize; i++)
		    b[i]=this.page[i];
	    }
	}

	// Represents a page address, i.e.  page number and chunk number.
	private final class PageAddress
	{
	    public int chunkNumber;
	    public long pageNumber;   

	    public PageAddress(int chunkNumber, long pageNumber)
	    {
		this.chunkNumber=chunkNumber;
		this.pageNumber=pageNumber;
	    }

	    public void update(int chunkNumber, long pageNumber)
	    {
		this.chunkNumber=chunkNumber;
		this.pageNumber=pageNumber;
	    }

	    public int hashCode()
	    {
		return (int)(pageNumber+(chunkNumber<<10));
	    }

	    public int compareTo(PageAddress addr)
	    {
		if (chunkNumber>addr.chunkNumber)
		    return 1;
		else if (chunkNumber<addr.chunkNumber)
		    return -1;
		else
		    {
			if (pageNumber>addr.pageNumber)
			    return 1;
			else if (pageNumber<addr.pageNumber)
			    return -1;
			else 
			    return 0;
		    }
	    }

	    public boolean equals(PageAddress addr)
	    {
		return this.compareTo(addr)==0;
	    }

	    public boolean equals(Object o)
	    {
		return this.equals((PageAddress)o);
	    }

	    public int compareTo(Object o)
	    {
		return this.compareTo((PageAddress)o);
	    }
	}


	// Represents a bucket of the hash table and a node of the circular list.
	private final class Bucket
	{
	    public PageAddress pageAddress;
	    public PageDescriptor pageDescriptor;
	    public Bucket next, prev;
	    public Bucket overflowPrev, overflowNext;
	    public boolean referenced;

	    public Bucket(PageAddress pageAddress, PageDescriptor pageDescriptor)
	    {
		this.pageAddress=pageAddress;
		this.pageDescriptor=pageDescriptor;
		this.referenced=true;
		next=prev=overflowPrev=overflowNext=null;
	    }
	}

	private int capacity;
	private Bucket head;
	private Bucket[] table;
	private int size;
	private int replaceAlgorithm;

	public PageBuffer(int capacity, int replaceAlgorithm) throws InvalidBufferCapacityException
	{
	    if (capacity<0)
		throw new InvalidBufferCapacityException();
	    init(capacity,replaceAlgorithm);
	}

	public PageBuffer(int capacity) throws InvalidBufferCapacityException
	{
	    this(capacity,LRU_ALGORITHM);
	}

	public PageBuffer() 
	{
	    init(1,LRU_ALGORITHM);
	}

	private void init(int capacity, int replaceAlgorithm)
	{
	    this.capacity=capacity;
	    this.replaceAlgorithm=replaceAlgorithm;
	    table=new Bucket[capacity];
	    head=null;
	    size=0;
	}

	private void insert(Bucket b)
	{
	    if (head==null)
		{
		    b.next=b.prev=b;
		    head=b;
		}
	    else
		{
		    b.prev=head.prev;
		    b.next=head;
		    head.prev.next=b;
		    head.prev=b;
		}

	    int pos=b.pageAddress.hashCode()%capacity;
	    if (table[pos]==null)
		{
		    b.overflowPrev=b.overflowNext=null;
		    table[pos]=b;
		}
	    else
		{
		    Bucket c=table[pos];
		    c.overflowPrev=b;
		    b.overflowNext=c;
		    b.overflowPrev=null;
		    table[pos]=b;
		}

	    size++;    
	}

	private void remove(Bucket b)
	{
	    if (b==head)
		head=b.next;
	    if (b.prev!=b)
		b.prev.next=b.next;
	    else
		head=null;
	    if (b.next!=b)
		b.next.prev=b.prev;
	    b.next=b.prev=null;

	    int pos=b.pageAddress.hashCode()%capacity;
	    if (b.overflowPrev!=null)
		b.overflowPrev.overflowNext=b.overflowNext;
	    else
		table[pos]=b.overflowNext;
	    if (b.overflowNext!=null)
		b.overflowNext.overflowPrev=b.overflowPrev;
	    b.overflowPrev=b.overflowNext=null;

	    size--;
	}

	private void reinsert(Bucket b)
	{
	    if (replaceAlgorithm!=FIFO_ALGORITHM && replaceAlgorithm!=LIFO_ALGORITHM
		&& replaceAlgorithm!=CLOCK_ALGORITHM)
		{
		    if (b==head)
			head=b.next;
		    if (b.prev!=b)
			b.prev.next=b.next;
		    else
			head=null;
		    if (b.next!=b)
			b.next.prev=b.prev;
		    b.next=b.prev=null;
		    
		    if (head==null)
			{
			    b.next=b.prev=b;
			    head=b;
			}
		    else
			{
			    b.prev=head.prev;
			    b.next=head;
			    head.prev.next=b;
			    head.prev=b;
			}
		}
	    else if (replaceAlgorithm==CLOCK_ALGORITHM)
		b.referenced=true;
	}
	
	private void replace(Bucket b, int chunkNumber, long pageNumber, byte[] page, boolean updated)
	    throws IOException
	{
	    int pos=b.pageAddress.hashCode()%capacity;
	    if (b.overflowPrev!=null)
		b.overflowPrev.overflowNext=b.overflowNext;
	    else
		table[pos]=b.overflowNext;
	    if (b.overflowNext!=null)
		b.overflowNext.overflowPrev=b.overflowPrev;
	    b.overflowPrev=b.overflowNext=null;
	
	    b.pageAddress.update(chunkNumber,pageNumber);
	    b.pageDescriptor.update(page,updated);

	    pos=b.pageAddress.hashCode()%capacity;
	    if (table[pos]==null)
		{
		    b.overflowPrev=b.overflowNext=null;
		    table[pos]=b;
		}
	    else
		{
		    Bucket c=table[pos];
		    c.overflowPrev=b;
		    b.overflowNext=c;
		    b.overflowPrev=null;
		    table[pos]=b;
		}
	}
	

	private void replace(int chunkNumber, long pageNumber, byte[] page, boolean updated)
	    throws IOException
	{
	    if (replaceAlgorithm==CLOCK_ALGORITHM)
		{
		    while (head.referenced)
			{
			    head.referenced=false;
			    head=head.next;
			}			   
		    transfer(head);
		    replace(head,chunkNumber,pageNumber,page,updated);
		    head.referenced=true;
		    head=head.next;		    
		}
	    else if (replaceAlgorithm==LIFO_ALGORITHM || replaceAlgorithm==MRU_ALGORITHM)
		{
		    transfer(head.prev);
		    replace(head.prev,chunkNumber,pageNumber,page,updated);
		} 
	    else
		{
		    transfer(head);
		    replace(head,chunkNumber,pageNumber,page,updated);
		    head=head.next;
		}
	}

	private void transfer(Bucket b) throws IOException
	{
	    if (b.pageDescriptor.updated)
		{
		     try {
			 sdm.writePage(b.pageAddress.chunkNumber, b.pageAddress.pageNumber, b.pageDescriptor.page);
			 b.pageDescriptor.updated=false;
		     } catch (NotExistingChunkNumberException nec) {
		     } catch (PageNumberOutOfRangeException pnor) {
		     }
		}
	}

	private PageAddress addressAux=new PageAddress(-1,-1);

	private Bucket find(int chunkNumber, long pageNumber)
	{
	    addressAux.chunkNumber=chunkNumber;
	    addressAux.pageNumber=pageNumber;
	    
	    int pos=addressAux.hashCode()%capacity;
	    if (table[pos]!=null)
		{
		    Bucket c=table[pos];
		    while (c!=null && !c.pageAddress.equals(addressAux))
			c=c.overflowNext;
		    return c;
		}
	    else
		return null;
	}
	
	
	public void putPage(int chunkNumber, long pageNumber, byte[] page, boolean updated)
	    throws IOException
	{
	    if (capacity>0)
		{
		    Bucket b=find(chunkNumber,pageNumber);
		    		    
		    if (b==null)
			{
			    if (buffer.size>=capacity)
				replace(chunkNumber,pageNumber,page,updated);
			    else
				insert(new Bucket(new PageAddress(chunkNumber,pageNumber),
						  new PageDescriptor(page,updated)));
				    
			}
		    else
			reinsert(b);
		}
	    
	}

	public boolean readPage(int chunkNumber, long pageNumber, byte[] b)
	{
	    Bucket bk=find(chunkNumber,pageNumber);
	    if (bk!=null)
		{
		    bk.pageDescriptor.copyPage(b);
		    reinsert(bk);
		    return true;
		}
	    else
		return false;
	}

	public boolean writePage(int chunkNumber, long pageNumber, byte[] b)
	{
	    Bucket bk=find(chunkNumber,pageNumber);
	    if (bk!=null)
		{
		    bk.pageDescriptor.update(b,true);
		    reinsert(bk);
		    return true;
		}
	    else
		return false;
	}

	public void transferAndClear()
	    throws IOException
	{
	    while (head!=null)
		{
		    transfer(head);
		    remove(head);
		}
	}

	public void clear()
	    throws IOException
	{
	    while (head!=null)
		remove(head);
	}
	
	public void transfer()
	    throws IOException
	{
	    Bucket b=head;
	    if (b!=null)
		do
		    {
			transfer(b);
			b=b.next;
		    }
		while (b!=head);
	}
    }

    
    private PageBuffer buffer;
    private boolean activedBuffer;
    
    /**
     * Constructs a primitive storage manager with a specific page size (in bytes), 
     * a specific buffer size (in bytes)
     * and a specific replace algorithm.
     */
    public PrimitiveStorageManager(int pageSize, int bufferSize, int replaceAlgorithm)
	throws InvalidPageSizeException, InvalidBufferCapacityException
    {
	if (pageSize<MIN_PAGE_SIZE)
	    throw new InvalidPageSizeException();
	init(pageSize,new PageBuffer(bufferSize/pageSize, replaceAlgorithm),true);
    }

    /**
     * Constructs a primitive storage manager with a specific page size (in bytes),
     * a specific buffer size (in bytes)
     * and the LRU replace algorithm.
     */
    public PrimitiveStorageManager(int pageSize, int bufferSize)
	throws InvalidPageSizeException, InvalidBufferCapacityException
    {
	this(pageSize,bufferSize,LRU_ALGORITHM);
    }

    /**
     * Constructs a primitive storage manager with a specific page size (in bytes) 
     * and with no actived buffer.
     */
    public PrimitiveStorageManager(int pageSize)
	throws InvalidPageSizeException
    {
	if (pageSize<MIN_PAGE_SIZE)
	    throw new InvalidPageSizeException();
	init(pageSize,new PageBuffer(), false);
    }

    /**
     * Constructs a primitive storage manager with the default page size and with no actived buffer.
     */
    public PrimitiveStorageManager()
    {
	init(DEFAULT_PAGE_SIZE,new PageBuffer(), false);
    }

    private void init(int pageSize, PageBuffer buffer, boolean activedBuffer)
    {
	this.pageSize=pageSize;
	this.chunksTable=new ChunksTable();
	this.buffer=buffer;
	this.sdm=new SecundaryDeviceManager();
	this.activedBuffer=activedBuffer;
    }

    /**
     * Actives the buffer.
     */
    public void activeBuffer()
    {
	if (!activedBuffer)
	    activedBuffer=true;
    }

    /**
     * Deactives the buffer.
     */
    public void deactiveBuffer()
	throws IOException
    {
	if (activedBuffer)
	    {
		buffer.transferAndClear();
		activedBuffer=false;
	    }
    }

    
    private void addChunk(int chunkNumber, Chunk chunk)
	throws InvalidChunkNumberException, AlreadyExistingChunkNumberException
    {
	if (chunkNumber<0 || chunkNumber>=MAX_CHUNK_NUMBER)
	    throw new InvalidChunkNumberException();
	if (chunksTable.containsChunkNumber(chunkNumber))
	    throw new AlreadyExistingChunkNumberException();
	else
	    chunksTable.put(chunkNumber,chunk);
    }

    /**
     * Load a fixed chunk with a specified chunk number.
     * @param chunkNumber The number which the chunk will be identified.
     * @param fileName The file to be used by the chunk.
     * @param offset The offset in the file when the chunk starts.
     * @param maxSize The maximun size that the chunk can occup.
     */
    public void loadFixedChunk(int chunkNumber, File fileName, long offset, long maxSize)
	throws InvalidChunkNumberException, AlreadyExistingChunkNumberException,
	       InvalidChunkSizeException, InvalidChunkOffsetException, 
	       UnaddressableChunkException, FileNotFoundException
    {
	addChunk(chunkNumber,new FixedChunk(fileName,offset,maxSize));
    }

    /**
     * Load a flexible chunk with a specified chunk number.
     * @param chunkNumber The number which the chunk will be identified.
     * @param fileName The file to be used by the chunk.
     * @param maxSize The maximun size that the chunk can grow.
     * @param overwrite If it is <code>false</code> the files's content will not be changed, 
     *        if it is <code>true</code> the files's content will be erased.
     */
    public void loadFlexibleChunk(int chunkNumber, File fileName, long maxSize, boolean overwrite)
	throws InvalidChunkNumberException, AlreadyExistingChunkNumberException,
	       FileNotFoundException, InvalidChunkSizeException, IOException
    {
	addChunk(chunkNumber,new FlexibleChunk(fileName,maxSize,overwrite));
    }

    /**
     * Load a flexible chunk with a specified chunk number and the default maximum size.
     * @param chunkNumber The number which the chunk will be identified.
     * @param fileName The file to be used by the chunk.
     * @param overwrite If it is <code>false</code> the files's content will not be changed, 
     *        if it is <code>true</code> the files's content will be erased.
     */
    public void loadFlexibleChunk(int chunkNumber, File fileName, boolean overwrite)
	throws InvalidChunkNumberException, AlreadyExistingChunkNumberException,
	       FileNotFoundException, IOException
    {
	addChunk(chunkNumber,new FlexibleChunk(fileName,overwrite));
    }

    /**
     * Load a flexible chunk with a specified chunk number, the default maximum size 
     * and keeping the previous file's content.
     * @param chunkNumber The number which the chunk will be identified.
     * @param fileName The file to be used by the chunk.
     */
    public void loadFlexibleChunk(int chunkNumber, File fileName)
	throws InvalidChunkNumberException, AlreadyExistingChunkNumberException,
	       FileNotFoundException
    {
	addChunk(chunkNumber,new FlexibleChunk(fileName));
    }

    /**
     * Removes the chunk with the specified number.
     * @param chunkNumber The number of the chunk to be removed.
     */
    public void removeChunk(int chunkNumber)
	throws NotExistingChunkNumberException
    {
	if (!chunksTable.containsChunkNumber(chunkNumber))
	    throw new NotExistingChunkNumberException();
	else
	    chunksTable.remove(chunkNumber);
    }

    /**
     * Returns the chunk with the specified number.
     * @param chunkNumber The number of the chunk to be retrieved.
     * @return The <code>Chunk</code> object requested.
     */
    public Chunk getChunk(int chunkNumber)
    {
	if (chunksTable.containsChunkNumber(chunkNumber))
	    return chunksTable.get(chunkNumber);
	else
	    return null;
    }

    
    /**
     * Copies the buffer's content to the secundary storage.
     */
    public void transferBuffer()
	throws IOException
    {
	if (activedBuffer)
	    buffer.transfer();
    }

    /**
     * Copies the buffer's content to the secundary storage and clean the buffer.
     */
    public void transferAndClearBuffer()
	throws IOException
    {
	if (activedBuffer)
	    buffer.transferAndClear();
    }

    /**
     * Read a specified page in a specified chunk and copies its content in an array.
     * @param chunkNumber The chunk number in which the page is.
     * @param pageNumber The number of the page inside the chunk.
     * @param b The byte array in which the page's content will be copied.
     */
    public void readPage(int chunkNumber, long pageNumber, byte[] b)
	throws NotExistingChunkNumberException, PageNumberOutOfRangeException, 
	       IOException
    {
	if (!activedBuffer)
	    sdm.readPage(chunkNumber,pageNumber,b);
	else if (!buffer.readPage(chunkNumber,pageNumber,b))
	    {
		sdm.readPage(chunkNumber,pageNumber,b);
		buffer.putPage(chunkNumber,pageNumber,b,false);
	    }
    }

    /**
     * Write a specified page in a specified chunk and copies its content from an array.
     * @param chunkNumber The chunk number in which the page is.
     * @param pageNumber The number of the page inside the chunk.
     * @param b The byte array from which the page's content will be copied.
     */
    public void writePage(int chunkNumber, long pageNumber, byte[] b)
	throws NotExistingChunkNumberException, PageNumberOutOfRangeException, 
	       IOException
    {
	if (!activedBuffer)
	    sdm.writePage(chunkNumber,pageNumber,b);
	else if (!buffer.writePage(chunkNumber,pageNumber,b))
	    buffer.putPage(chunkNumber,pageNumber,b,true);
    }

    /**
     * Return the used page size (in bytes).
     */
    public int getPageSize()
    {
	return pageSize;
    }

    /**
     * This class is used to retrieve the used chunk numbers. 
     * Chunks numbers are retrived in a sequence with the insertion order.
     */
    public class ChunkNumberIterator
    {
	private ChunksTable.Entry current;
	
	ChunkNumberIterator()
	{
	    current=chunksTable.getFirst();
	}
	
	/**
	 * Indicates if there are more chunk numbers in the sequence.
	 */
	public boolean hasNext()
	{
	    return current!=null;
	}
	
	/**
	 * Returns the next chunk number in the sequence.
	 */
	public int next()
	{
	    int r=current.getChunkNumber();
	    current=current.getNext();
	    return r;
	}
    }
    
    /**
     * Returns the sequence of used chunk numbers.
     */
    public ChunkNumberIterator getChunkNumbers()
    {
	return new ChunkNumberIterator();
    }

    /**
     * Closes all files used in this component.
     */
    public void closeAll()
    {
	ChunkNumberIterator i=getChunkNumbers();
	while (i.hasNext())
	    {
		int n=i.next();
		try {
		    getChunk(n).getFile().close();
		} catch (IOException e) {}
	    }
    }
    
    /**
     * Shows information about this component.
     */
    public void show() throws IOException
    {
	ChunkNumberIterator i=getChunkNumbers();
	while (i.hasNext())
	    {
		int n=i.next();
		System.out.println("Chunk "+n+".");
		System.out.println(getChunk(n).getInfo());
	    }
    }
}
