package storage.tools;

import java.io.*;
import storage.primitiveStorage.*;
import storage.intermediateStorage.*;
import storage.advancedStorage.*;
import storage.*;

/**
 * This class provides methods for setting and tuning the store components.
 */
public final class StorageToolsManager
{
    static private final String ENCODING="ISO-8859-1";

    private AdvancedStorageManager asm;

    /**
     * Returns the underlying primitive storage manager.
     */
    public PrimitiveStorageManager getPrimitiveStorageManager()
    {
	return asm.getPrimitiveStorageManager();
    }

    /**
     * Returns the underlying intermediate storage manager.
     */
    public IntermediateStorageManager getIntermediateStorageManager()
    {
	return asm.getIntermediateStorageManager();
    }
    
    /**
     * Returns the underlying advanced storage manager.
     */
    public AdvancedStorageManager getAdvancedStorageManager()
    {
	return asm;
    }


    static private class Tokenizer
    {
	private String line;
	private int currentPos;
	private String lastToken;
	private int lineNumber;

	public Tokenizer(String line, int lineNumber)
	{
	    load(line, lineNumber);
	}

	public Tokenizer()
	{
	    load(null,0);
	}

	public int getLineNumber()
	{
	    return lineNumber;
	}

	public void load(String line,int lineNumber)
	{
	    this.line=line;
	    this.currentPos=0;
	    this.lastToken=null;
	    this.lineNumber=lineNumber;
	    shift();
	}

	private void shift()
	{
	    if (line!=null)
		{
		    while (currentPos<line.length() && line.charAt(currentPos)!='%' &&  
			   Character.isSpaceChar(line.charAt(currentPos)))
			currentPos++;
		    
		    if (currentPos>=line.length() || line.charAt(currentPos)=='%')
			lastToken=null;
		    else
			{
			    int tokenPosition;
			    if (line.charAt(currentPos)=='\"')
				{
				    tokenPosition=currentPos+1;
				    currentPos++;
				    while (currentPos<line.length() && line.charAt(currentPos)!='\"')
					currentPos++;
				    lastToken=line.substring(tokenPosition,currentPos);
				    currentPos++;
				}
			    else
				{
				    tokenPosition=currentPos;
				    while (currentPos<line.length() && !Character.isSpaceChar(line.charAt(currentPos)))
					currentPos++;
				    lastToken=line.substring(tokenPosition,currentPos);
				}
			}
		}
	}

	public boolean hasNextToken()
	{
	    return lastToken!=null;
	}

	public String nextToken()
	{
	    String token=lastToken;
	    shift();
	    return token;
	}
    }


    static private final String loadErrorMessage="Bad format in line: ";

    private void loadFixedChunk(Tokenizer t)
	throws PrimitiveStorageException, IOException, BadFileFormatException
    {
	int chunkNumber;
	if (t.hasNextToken())
	    {
		try {
		    chunkNumber=Integer.parseInt(t.nextToken());
		} catch (NumberFormatException e) {
		    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());
		}
	    }
	else
	    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());

	String chunkFile;
	if (t.hasNextToken())
	    chunkFile=t.nextToken();
	else
	    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());	
    
	long chunkOffset;
	if (t.hasNextToken())
	    {
		try {
		    chunkOffset=Long.parseLong(t.nextToken());
		} catch (NumberFormatException e) {
		    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());
		}
	    }
	else
	    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());

	long chunkMaxSize;
	if (t.hasNextToken())
	    {
		try {
		    chunkMaxSize=Long.parseLong(t.nextToken());
		} catch (NumberFormatException e) {
		    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());
		}
	    }
	else
	    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());

	getPrimitiveStorageManager().loadFixedChunk(chunkNumber, new File(chunkFile), chunkOffset, chunkMaxSize);
    }


    private void loadFlexibleChunk(Tokenizer t)
	throws PrimitiveStorageException, IOException, BadFileFormatException
    {
	int chunkNumber;
	if (t.hasNextToken())
	    {
		try {
		    chunkNumber=Integer.parseInt(t.nextToken());
		} catch (NumberFormatException e) {
		    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());
		}
	    }
	else
	    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());

	String chunkFile;
	if (t.hasNextToken())
	    chunkFile=t.nextToken();
	else
	    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());	
    
	if (t.hasNextToken())
	    {
		try {
		    long chunkMaxSize;
		    chunkMaxSize=Long.parseLong(t.nextToken());
		    	getPrimitiveStorageManager().loadFlexibleChunk(chunkNumber, new File(chunkFile), 
								       chunkMaxSize, false);
		} catch (NumberFormatException e) {
		    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());
		}
	    }
	else
	    getPrimitiveStorageManager().loadFlexibleChunk(chunkNumber, new File(chunkFile), false);
    }

    private void loadChunk(Tokenizer t)
	throws PrimitiveStorageException, IOException, BadFileFormatException
    {
	String chunkType;
	if (t.hasNextToken())
	    chunkType=t.nextToken();
	else
	    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());

	String tok=chunkType.trim().toLowerCase();
	if (tok.equals("fixed"))
	    loadFixedChunk(t);
	else if (tok.equals("flexible"))
	    loadFlexibleChunk(t);
	else
	    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());
    }

    
    private void loadSpace(Tokenizer t)
	throws PrimitiveStorageException, IntermediateStorageException, AdvancedStorageException, 
	       IOException, BadFileFormatException
    {
	int segmentNumber;
	if (t.hasNextToken())
	    {
		try {
		    segmentNumber=Integer.parseInt(t.nextToken());
		} catch (NumberFormatException e) {
		    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());
		}
	    }
	else
	    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());

	int spaceNumber;
	if (t.hasNextToken())
	    {
		try {
		    spaceNumber=Integer.parseInt(t.nextToken());
		} catch (NumberFormatException e) {
		    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());
		}
	    }
	else
	    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());

	int chunkNumber;
	if (t.hasNextToken())
	    {
		try {
		    chunkNumber=Integer.parseInt(t.nextToken());
		} catch (NumberFormatException e) {
		    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());
		}
	    }
	else
	    throw new BadFileFormatException(loadErrorMessage+t.getLineNumber());

	getAdvancedStorageManager().getSegment(segmentNumber).loadSpace(spaceNumber,chunkNumber,false);
    }


    /** Loads the storage settings (chunk and space organization) from a file. */
    public void loadSettingsFromFile(File file)
	throws BadFileFormatException, IOException,
	       PrimitiveStorageException, IntermediateStorageException, AdvancedStorageException
    {
	Tokenizer t=new Tokenizer();
	BufferedReader in=new BufferedReader(new FileReader(file));
	String line=in.readLine();
	int lineNumber=1;
	while (line!=null)
	    {
		t.load(line,lineNumber);
		if (t.hasNextToken())
		    {
			String tok=t.nextToken().trim().toLowerCase();
			if (tok.equals("chunk"))
			    loadChunk(t);
			else if (tok.equals("space"))
			    loadSpace(t);
		    }
		line=in.readLine();
		lineNumber++;
	    }
    }

    /** Saves the current storage settings (chunk and space organization) to a file. */
    public void saveSettingsToFile(File file)
	throws IOException
    {
	FileOutputStream out=new FileOutputStream(file);
	PrimitiveStorageManager.ChunkNumberIterator i=getPrimitiveStorageManager().getChunkNumbers();
	while (i.hasNext())
	    {
		String s=null;
		int chunkNumber=i.next();
		PrimitiveStorageManager.Chunk chunk=getPrimitiveStorageManager().getChunk(chunkNumber);
		if (chunk instanceof PrimitiveStorageManager.FixedChunk)
		    s="CHUNK FIXED "+chunkNumber+" \""+chunk.getFileName()+"\" "+
			((PrimitiveStorageManager.FixedChunk)chunk).getOffset()+" "+chunk.getMaxSize()+"\n";
		else if (chunk instanceof PrimitiveStorageManager.FlexibleChunk)
		    s="CHUNK FLEXIBLE "+chunkNumber+" \""+chunk.getFileName()+"\" "+
			chunk.getMaxSize()+"\n";
		if (s!=null)
		    out.write(s.getBytes(ENCODING));
	    }
	
	int segmentsNumber=getAdvancedStorageManager().getNumberOfSegments();
	for (int segmentNumber=0; segmentNumber<segmentsNumber; segmentNumber++)
	    {
		AdvancedStorageManager.Segment.SpaceNumberIterator k
		    =getAdvancedStorageManager().getSegment(segmentNumber).getSpaceNumbers();
		while (k.hasNext())
		    {
			String s=null;
			int spaceNumber=k.next();
			int chunkNumber
			    =getAdvancedStorageManager().getSegment(segmentNumber)
			    .getSpace(spaceNumber).getChunkNumber();
			s="SPACE "+segmentNumber+" "+spaceNumber+" "+chunkNumber+"\n";
			out.write(s.getBytes(ENCODING));
		    }
	    }
    }

    /** Show the information of storage organization. */
    public void showInfo()
	throws PrimitiveStorageException, IntermediateStorageException, IOException
    {
	System.out.println("STORAGE INFORMATION.\n\nCHUNKS:\n");
	PrimitiveStorageManager.ChunkNumberIterator i=getPrimitiveStorageManager().getChunkNumbers();
	while (i.hasNext())
	    {
		String s=null;
		int chunkNumber=i.next();
		PrimitiveStorageManager.Chunk chunk=getPrimitiveStorageManager().getChunk(chunkNumber);
		if (chunk instanceof PrimitiveStorageManager.FixedChunk)
		    s="number:"+chunkNumber+" type:fixed file:\""+chunk.getFileName()+"\" offset:"+
			((PrimitiveStorageManager.FixedChunk)chunk).getOffset()+" max-size:"+
			chunk.getMaxSize();
		else if (chunk instanceof PrimitiveStorageManager.FlexibleChunk)
		    s="number:"+chunkNumber+" type:flexible file:\""+chunk.getFileName()+
			"\" max-size:"+chunk.getMaxSize()+
			" real-size:"+((PrimitiveStorageManager.FlexibleChunk)chunk).getSize();
		if (s!=null)
		    System.out.println(s);
	    }
	
	System.out.println("\nSPACES:\n");
	int segmentsNumber=getAdvancedStorageManager().getNumberOfSegments();
	for (int segmentNumber=0; segmentNumber<segmentsNumber; segmentNumber++)
	    {
		AdvancedStorageManager.Segment.SpaceNumberIterator k
		    =getAdvancedStorageManager().getSegment(segmentNumber).getSpaceNumbers();
		while (k.hasNext())
		    {
			String s=null;
			int spaceNumber=k.next();
			AdvancedStorageManager.Segment.Space space
			    =getAdvancedStorageManager().getSegment(segmentNumber).getSpace(spaceNumber);
			s="segment:"+segmentNumber+" space:"+spaceNumber+" used-size:"+space.getUsedSize()+
			    " chunk:"+space.getChunkNumber();
			System.out.println(s);
		    }
	    }
    }

    /** 
     * Reduces the logic size of spaces and if it is posible also reduces the real size of 
     * the associated chunk.
     */ 
    public void trimSpace(int segmentNumber, int spaceNumber)
	throws IOException, PrimitiveStorageException, IntermediateStorageException
    {
	AdvancedStorageManager.Segment.Space space
	    =getAdvancedStorageManager().getSegment(segmentNumber).getSpace(spaceNumber);
	space.reduce();
	PrimitiveStorageManager.Chunk chunk=getPrimitiveStorageManager().getChunk(space.getChunkNumber());
	if (chunk instanceof PrimitiveStorageManager.FlexibleChunk)
	    ((PrimitiveStorageManager.FlexibleChunk)chunk).changeSize(space.getUsedSize());
    }

    /**
     * Execute a trim for all spaces of the storage organization. 
     */
    public void trimAllSpaces()
	throws IOException, PrimitiveStorageException, IntermediateStorageException
    {
	int segmentsNumber=getAdvancedStorageManager().getNumberOfSegments();
	for (int segmentNumber=0; segmentNumber<segmentsNumber; segmentNumber++)
	    {
		AdvancedStorageManager.Segment.SpaceNumberIterator k
		    =getAdvancedStorageManager().getSegment(segmentNumber).getSpaceNumbers();
		while (k.hasNext())
		    trimSpace(segmentNumber,k.next());
	    }
    }

    /**
     * Creates and loads a fixed chunk in the storage organization.
     */
    public void createFixedChunk(int chunkNumber, File file, long maxSize, long offset)
	throws PrimitiveStorageException, IOException
    {
	getPrimitiveStorageManager().loadFixedChunk(chunkNumber, file, offset, maxSize);
    }

    /**
     * Creates and loads a flexible chunk with a specified maximun size
     * in the storage organization, the previous
     * content of the file is destoyed.
     */
    public void createFlexibleChunk(int chunkNumber, File file, long maxSize)
	throws PrimitiveStorageException, IOException
    {
	getPrimitiveStorageManager().loadFlexibleChunk(chunkNumber, file, maxSize, true);
    }

    /**
     * Creates and loads a flexible chunk in the storage organization, 
     * the previous content of the file is destoyed.
     */
    public void createFlexibleChunk(int chunkNumber, File file)
	throws PrimitiveStorageException, IOException
    {
	getPrimitiveStorageManager().loadFlexibleChunk(chunkNumber, file, true);
    }

    /**
     * Creates a new space with a associated chunk, the previous content of the file
     * is replaced. 
     */
    public void createSpace(int segmentNumber, int spaceNumber, int chunkNumber)
	throws PrimitiveStorageException, IntermediateStorageException, AdvancedStorageException, IOException
    {
	getAdvancedStorageManager().getSegment(segmentNumber).loadSpace(spaceNumber,chunkNumber,true);
    }

    /**
     * Drop or removes a space from the storage organization.
     */
    public void dropSpace(int segmentNumber, int spaceNumber)
	throws UndroppableSpaceException, AdvancedStorageException
    {
	AdvancedStorageManager.Segment segment=getAdvancedStorageManager().getSegment(segmentNumber);
	if (segment.getSpace(spaceNumber).isEmpty())
	    segment.removeSpace(spaceNumber);
	else
	    throw new UndroppableSpaceException();
    }

    private boolean isChunkInAnySpace(int chunkNumber)
    {
	int segmentsNumber=getAdvancedStorageManager().getNumberOfSegments();
	for (int segmentNumber=0; segmentNumber<segmentsNumber; segmentNumber++)
	    {
		AdvancedStorageManager.Segment segment=getAdvancedStorageManager().getSegment(segmentNumber);
		AdvancedStorageManager.Segment.SpaceNumberIterator k=segment.getSpaceNumbers();
		while (k.hasNext())
		    if (segment.getSpace(k.next()).getChunkNumber()==chunkNumber)
			return true;
	    }
	return false;
    }

    /**
     * Drop or removes a chunk from the storage organization.
     */
    public void dropChunk(int chunkNumber)
	throws UndroppableChunkException, PrimitiveStorageException
    {
	if (!isChunkInAnySpace(chunkNumber))
	    getPrimitiveStorageManager().removeChunk(chunkNumber);
	else
	    throw new UndroppableChunkException();
    }

    /**
     * Creates a storage tools' magager from an advanced storage manager.
     */
    public StorageToolsManager(AdvancedStorageManager asm)
    {
	this.asm=asm;
    }
}
